/***********************************************************************
* Copyright (c) 1995-2003 by peter fuerst
*
*	This program is free software; you can redistribute it and/or modify
*	it - as explained in the accompaning file COPYING - under the terms
*	of the GNU General Public License as published by the Free Software
*	Foundation; either version 2, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	Author: Peter Frst, post@pfrst.de
*
*	pfm.c	10/25/95
*	conf.c
*
***********************************************//*$preferred tabstop=4*/

#include <unistd.h>
#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <ctype.h>
#include <dirent.h>
#include <time.h>
#include <sys/param.h>
#include <grp.h>
#include <pwd.h>
#include <signal.h>
#include <errno.h>


/***** define & typedef ********************************************/

#include "pfm.h"
#include "pfmstat.h"

#define BOLD "\007" /* use BELl to emphasize the next character */

#ifndef const_cast
 #define const_cast(typ,val)	((typ)val)
#endif


/***** forward declaration *****************************************/

struct cmdrecord;
__DC( static const char *_pfmConfDocdir,(struct cmdrecord*) );
__DC( static const char *_pfmConfPagerPrg,(struct cmdrecord*) );
__DC( static const char *_pfmConfEditorCmd,(struct cmdrecord*) );
__DC( static const char *_pfmConfViewerCmd,(struct cmdrecord*) );
__DC( static const char *_pfmConfMailerCmd,(struct cmdrecord*) );
__DC( static const char *_pfmConfPrintCmd,(struct cmdrecord*) );
__DC( static const char *_pfmConfCpdirCmd,(struct cmdrecord*) );
#ifdef SUPPORT_INET_URL
__DC( static const char *_pfmConfUrlHdlrCmd,(struct cmdrecord*) );
__DC( static const char *_pfmConfWwwCmd,(struct cmdrecord*) );
__DC( static const char *_pfmConfFtpCmd,(struct cmdrecord*) );
__DC( static const char *_pfmConfTelnetCmd,(struct cmdrecord*) );
#endif
__DC( static bool pfmConfScanboolean,(const char*,bool,const char**) );
__DC( static char *pfmConfSkipws,(char*) );


/***** global ******************************************************/

const char  *pfm_Shellprompt = "$";
#ifdef DEUTSCH
 const char  pfm_C_hit_any_key[] = "Weiter mit beliebiger Taste... ";
 const char  pfm_C_hit_enter[] = "<enter> ... ";
 const char  pfm_C_dir_pathname[] = "Verzeichnispfad:"; /*"Pfad zum Verzeichnis:";*/
 const char  pfm_C_pathname[] = "Pfadname:";
 const char  pfm_C_file_name[] = "Dateiname:";
 const char  pfm_C_new_name[] = "Neuer Name:";
 const char  pfm_C_destination[] = "Ziel:";
 const char  pfm_C_date_and_time[] = "Datum/Zeit:";
 /* Hmmm, since more than two centuries we hardly can use the logically correct
  * 2nd person (Seid Ihr sicher ?), but have to use 3rd person (Sind Sie...) */
 const char  pfm_C_confirm[] = "Sind Sie sicher [J/N] ?";
 const char  pfm_C_no_nein_non[] = "Nein";
 const char  pfm_C_yes_ja_oui[] = "Ja";
#else
 const char  pfm_C_hit_any_key[] = "Hit any key to continue ... ";
 const char  pfm_C_hit_enter[] = "Hit <enter> ... ";
 const char  pfm_C_dir_pathname[] = "Directory Pathname:";
 const char  pfm_C_pathname[] = "Pathname:";
 const char  pfm_C_file_name[] = "Filename:";
 const char  pfm_C_wild_name[] = "Wild filename:";
 const char  pfm_C_new_name[] = "New name:";
 const char  pfm_C_destination[] = "Destination:";
 const char  pfm_C_date_and_time[] = "Date/Time:";
 const char  pfm_C_confirm[] = "Are you sure [Y/N] ?";
 const char  pfm_C_no_nein_non[] = "No";
 const char  pfm_C_yes_ja_oui[] = "Yes";
#endif
/* strerror(ENOMEM) mostly would not fit too well */
const char  pfm_C_out_of_stack[] = "Out of stack space !";


/***** local *******************************************************/

#ifdef DEUTSCH
 static const char show_dotfiles_txt[] = "'.'-Dateien anzeigen [J/N] ?";
 static const char more_errors_txt[] = "mehr Fehler...";
 static const char save_conf_txt[] = "Konfiguration speichern [J/N] ?";
 static const char not_before_txt[] = "geaendert nicht vor:";
 static const char not_after_txt[] = "geaendert nicht nach:";
 #ifdef NCURSES_MOUSE_VERSION
 static const char use_mouse_txt[] = "Maus aktivieren [J/N] ?";
 #endif
#else
 static const char show_dotfiles_txt[] = "Show dot-files [Y/N] ?";
 static const char more_errors_txt[] = "more errors...";
 static const char save_conf_txt[] = "Save configuration [Y/N] ?";
 static const char not_before_txt[] = "mTime not before:";
 static const char not_after_txt[] = "mTime not after:";
 #ifdef NCURSES_MOUSE_VERSION
 static const char use_mouse_txt[] = "Activate mouse [J/N] ?";
 #endif
#endif

static const char enterbd_cmt_txt[] =
"\n"
"! pfm.Enter: pattern, modemask, notbefore, notafter, K|C, key|command\n"
#if 0
#ifdef DEUTSCH
 "! ',', '\\', Leer- oder nicht-druckbare Zeichen (oder nachfolgende Ziffern)\n"
 "! in <pattern> muessen als oktale Escape-Sequenz kodiert werden: \\ddd\n"
 "! <modemask> ist in der Form von ls' Modusanzeige anzugeben. Abschliessende\n",
 "! Folgen von '?' koennen durch '*' abgekuerzt werden.\n"
 "! Zeiten koennen als yyyy mm|monthname dd hh:mm:ss oder als Sekunden\n"
 "! seit \"the epoch\" angegeben werden.\n"
 "! <key> ist ein einzelnes druckbares Nicht-Leer-Zeichen oder eine\n"
 "! Oktalzahl ddd (mit mindestens zwei Ziffern).\n"
 "! Beispiele:\n!\n"
 "! pfm.Enter: , dr?x???????, 1970 Jan 1 01:00:00, 0x7fffffff, K, S\n"
 "! pfm.Enter: *.html, -r?????????, 0, 2030 12 31 23:59:00, C, lynx '%F'\n"
#else
 "! ',', '\\', spaces or nonprintable chars (or digits following any of\n"
 "! these) in <pattern> must be coded as octal escape sequence: \\ddd\n"
 "! <modemask> is like ls's mode output. Trailing sequences of '?'\n"
 "! may be abbreviated with a '*'\n"
 "! Time values can be given as yyyy mm|monthname dd hh:mm:ss or as\n"
 "! seconds since \"the epoch\".\n"
 "! <key> is either a single printable non-space character or octal\n"
 "! number ddd (with at least two digits).\n"
 "! examples:\n!\n"
 "! pfm.Enter: , dr?x???????, 1970 Jan 1 01:00:00, 0x7fffffff, K, S\n"
 "! pfm.Enter: *.html, -r?????????, 0, 2030 12 31 23:59:00, C, lynx '%F'\n"
#endif
#else
#ifdef DEUTSCH
 "! eine detailierte Beschreibung findet sich in der Hilfe/Dokumentation.\n"
#else
 "! for detailed description see help/documentation.\n"
#endif
#endif
;


/********************************************************************
*	constant default values...
********************************************************************/

#define PFMCONFHOME "~/.pfm" /* instead of a comment ;-) */
static const char conf_home_def[] = PFMCONFHOME;
static const char conf_file_def[] = PFMCONFHOME "/rc";
#undef PFMCONFHOME
#ifdef PFM_BUILD_DIR
 static const char doc_dir_def[] = PFM_BUILD_DIR "/dok/e";
 #define doc_dir_def doc_dir_def /* mark as available to compiler */
#endif
/* assert( pfm_fwildc_start == '%' && pfm_Fname_fwildc == 'F' ); */

static const char cmdLpr[] =
	#ifdef __COHERENT__
		"lpr -B '%F'";
	#else
		"lpr -h '%F'";
	#endif

static const char cmdCpdir[] =
	/* wrappers for "cp -R" */
	#if defined(__FreeBSD__) || defined(__NetBSD__)
		BUILTIN_CMD_PFX " -RHpv '%D' '%A'";
	#elif defined(__linux__) /* GNU "cp -R" */
		BUILTIN_CMD_PFX " -Rdpv '%D' '%A'";
	#elif defined(__AIX__)
		BUILTIN_CMD_PFX " -Rhp '%D' '%A'";
	#elif defined(__IRIX__)
		BUILTIN_CMD_PFX " -Rp '%D' '%A'";
	/* native "cpdir" commands */
	#elif defined(__COHERENT__)
		"cpdir -dv '%D' '%A'";
	#elif defined(__XENIX__) || defined(__SCO__) /* "old" SCO ;-) */
		"copy -rmov '%D' '%A'";
	#else /* wrapper for unknown cp command */
		"mycpdir -? '%D' '%A'";
	#endif

static const char cmdVi[]  = "vi '%F'";
static const char *argvVi[] = {"vi", "%F", 0};

static const char cmdMore[] =
	#ifdef __linux__
		"less -dimnu";
	#else
		"more -lu";
	#endif
static const char *argvMore[] =
	#ifdef __linux__
		{"less", "-dimnu", 0};
	#else
		{"more", "-lu", 0};
	#endif
static const char cmdMail[]    = "mail '%F'";
#ifdef SUPPORT_INET_URL
#define init_url_cmd_str(cmd) cmd
#else
#define init_url_cmd_str(cmd) ""
#endif
static const char cmdLynx[]   = init_url_cmd_str("lynx '%U%F'");
static const char cmdFtp[]    = init_url_cmd_str("ftp '%F'");
static const char cmdTelnet[] = init_url_cmd_str("telnet '%F'");
#undef init_url_cmd_str

static const char true_txt[]  = "True";
static const char false_txt[] = "False";


/*
 *	test for static command entry in preparation to reconfigure it.
 */
__DF(
static bool pfmConfIsstaticstr,(cmd), const char *cmd )
{
	return (
		!cmd || !*cmd /* empty commands are not allocated */
		|| cmdVi == cmd
		|| cmdMore == cmd
		|| cmdMail == cmd
		|| cmdLpr == cmd
		|| cmdCpdir == cmd
		|| cmdLynx == cmd
		|| cmdFtp == cmd
		|| cmdTelnet == cmd
		|| conf_home_def == cmd
		#ifdef doc_dir_def
		|| doc_dir_def == cmd
		#endif
	);
}


/********************************************************************
*	configurable parameters...
********************************************************************/

/*
 *	boolean flags: table of configuration file entries ...
 */
typedef struct
{
	bool        r;   /* redefined (from file/interactive) ? */
	bool        b;   /* flag */
	const char  *id; /* identfication tag (for file...) */
	const char  *t;  /* describing text */
}
x_flgrecord;

/*
 *	show dot-files in directory window ?
 */
static x_flgrecord _S_show_dot_files =
{	0, 1, "DotFiles", show_dotfiles_txt
	};

#ifdef NCURSES_MOUSE_VERSION
/*
 *	enable ncurses' interface to mouse-events ?
 */
static x_flgrecord _S_enable_mouse =
{	0, 0, "Mouse", use_mouse_txt
	};
#endif

/*
 *	some commands in execv-style...
 */
static const char	**pfm_pager_argv   = argvMore;
static const char	**pfm_editor_argv  = argvVi;

/*
 *	commands: table of configuration file entries ...
 */
typedef struct cmdrecord x_cmdrecord;
struct cmdrecord
{
	bool        r;      /* redefined (from file/interactive) ? */
	const char  *id;    /* identfication tag (for file...) */
	char        *c;     /* command string */
	const char* __DC((*q),(x_cmdrecord*)); /* query environment etc. for command */
	const char  *t;     /* describing text */
};

static x_cmdrecord cmdtbl[] =
{
	{0, "DocDir",   0, _pfmConfDocdir,     "Where to find pfm help/doc:"},
	#define PFM_CMD_DOC_IDX	0
	{0, "PagerPrg", 0, _pfmConfPagerPrg,   "Pager for builtin viewer:"},
	#define PFM_CMD_PGR_IDX	1
	{0, "Editor",   0, _pfmConfEditorCmd,  "Editor command:"},
	#define PFM_CMD_EDT_IDX	2
	{0, "Viewer",   0, _pfmConfViewerCmd,  "Command to override builtin viewer:"},
	#define PFM_CMD_VIE_IDX	3
	{0, "Mailer",   0, _pfmConfMailerCmd,  "Mailer command:"},
	#define PFM_CMD_MAI_IDX	4
	{0, "Printer",  0, _pfmConfPrintCmd,   "Print command:"},
	#define PFM_CMD_PRN_IDX	5
	{0, "Cpdir",    0, _pfmConfCpdirCmd,   "Cpdir command:"},
	#define PFM_CMD_CPD_IDX	6
#ifdef SUPPORT_INET_URL
	{0, "UrlHdlr",  0, _pfmConfUrlHdlrCmd, "default URL-handler command:"},
	#define PFM_CMD_URL_IDX	(1+PFM_CMD_CPD_IDX)
	{0, "Www",      0, _pfmConfWwwCmd,     "http-browser command:"},
	#define PFM_CMD_WWW_IDX	(1+PFM_CMD_URL_IDX)
	{0, "Ftp",      0, _pfmConfFtpCmd,     "Ftp-client command:"},
	#define PFM_CMD_FTP_IDX	(1+PFM_CMD_WWW_IDX)
	{0, "Telnet",   0, _pfmConfTelnetCmd,  "Telnet-client command:"},
	#define PFM_CMD_TEL_IDX	(1+PFM_CMD_FTP_IDX)
#endif
};

/*
 *	Pager command, currently not really a configurable parameter,
 *	will be built from pfm_pager_cmdPrg instead.
 */
static char *pfm_pager_cmd = 0;


/*
 *	table of enter-key bindings ...
 */
typedef struct enterbinding
{
	struct enterbinding *next;
	char	*pat;
	char	att[12];
	time_t	beg;
	time_t	end;
	char	typ;
	union {
	int 	key;
	char	*cmd;
	}		act;
}
x_enterbinding;

static x_enterbinding *p_enter_tbl = 0;
/*
static x_enterbinding fixed_enter_tbl[] =
{
	{0, 0, "d??????????", 0L, 0x7fffffffL, 'K', {'S'}},
	{0, "*.html", "-*", 0L, 0x7fffffffL, 'C', {"lynx '%F'"}},
	{0, "*.html", "-*", 0L, 0x7fffffffL, 'C', {"netscape './%F' &"}},
};
*/


/********************************************************************
*	satisfy (external) queries for configuration parameters
********************************************************************/

/*
 *	most recently read or written configuration-file
 */
__DF(
static const char *pfmConffile,( fname ), const char *fname )
{
	static char	*conffile = 0;

	if ((!fname || !*fname) && !conffile) /* enforce initilization */
		fname = conf_file_def;

	if (fname && *fname && (!conffile || strcmp(conffile,fname)))
	{	/*
		 *	eventually expand "~" i.e. $HOME
		 */
		int k = 2+pfmExpandsize(fname);

		if (conf_file_def == conffile) conffile = 0;
		conffile = (char*) realloc( conffile, k );
		if ( conffile )
			pfmExpandargs( strcpy(conffile,fname), k );
	}
	if ( !conffile )
		conffile = const_cast( char*, conf_file_def );
	return conffile;
}

/*
 *	show dot-files in directory window ?
 */
bool pfmConfShowdotfiles ()
{
	return _S_show_dot_files.b;
}

#ifdef NCURSES_MOUSE_VERSION
/*
 *	enable ncurses' interface to mouse-events ?
 */
bool pfmConfMouseinput ()
{
	return _S_enable_mouse.b;
}
#endif


/*
 *	directory that contains help and documentation
 */
__DF(
static const char *_pfmConfDocdir,(cmd), x_cmdrecord *cmd )
{
	int k;

	if ( !cmd->c )
	{
		#ifdef doc_dir_def
			/*	is the build directory still available at runtime ? */
			struct stat	sbuf;
			if (!stat(doc_dir_def,&sbuf) && S_ISDIR(sbuf.st_mode))
				cmd->c = const_cast( char*, doc_dir_def );
			else
		#endif
				cmd->c = const_cast( char*, conf_home_def );
	}
	k = pfmVExpandsize( cmd->c, "", (char*)0, (char*)0 );
	if (k > strlen(cmd->c))
	{	/*
		 *	eventually expand "~" i.e. $HOME
		 */
		bool iss = pfmConfIsstaticstr(cmd->c);
		char *p = (char*) (iss ? malloc(k+2) : realloc(cmd->c,k+2));
		if (p)
		{	if (iss) strcpy( p, cmd->c );
			pfmVExpandargs( cmd->c = p, k+2, "", (char*)0, (char*)0 );
		}
	}
	return cmd->c;
}

const char *pfmConfDocdir ()
{
	return _pfmConfDocdir( cmdtbl+PFM_CMD_DOC_IDX );
}


/*
 *	various commands...
 */
 
/*
 *	Pager program used by the builtin file viewer for variabel lines.
 *	Must read its data from stdin. (Program and options only, no
 *	placeholder-variables)
 */
__DF(
static const char *_pfmConfPagerPrg,(cmd), x_cmdrecord *cmd )
{
	if ( !cmd->c )
	{	if ( (cmd->c = getenv("PAGER")) )
			cmd->c = *cmd->c ? strdup(cmd->c):0;
		if ( !cmd->c )
			cmd->c = const_cast( char*, cmdMore );
	}
	return cmd->c;
}

const char *pfmConfPagerPrg ()
{
	return _pfmConfPagerPrg( cmdtbl+PFM_CMD_PGR_IDX );
}

/*
 *	Pager command (currently used only by the help function ?)
 *	Reads its data from a file given on its commandline.
 */
const char *pfmConfPagerCmd ()
{
	if ( !pfm_pager_cmd )
	{	const char *cp = pfmConfPagerPrg();
		pfm_pager_cmd = (char*)malloc(strlen(cp)+6);
		if ( !pfm_pager_cmd )
		{	static char cmd[sizeof(cmdMore)+6];
			pfm_pager_cmd = cmd;
			cp = cmdMore;
		}
		pfm_pager_cmd = strcat( strcpy(pfm_pager_cmd,cp), " '%F'" );
	}
	return pfm_pager_cmd;
}

/*
 *	Editor command (currently used by the edit-file command)
 */
__DF(
static const char *_pfmConfEditorCmd,(cmd), x_cmdrecord *cmd )
{
	if ( !cmd->c )
	{	char *cp = getenv("EDITOR");
		if (!cp || !*cp) cp = getenv("VISUAL");

		if (cp && *cp)
		{	cmd->c = (char*)malloc(strlen(cp)+6);
			if ( cmd->c )
				strcat( strcpy(cmd->c,cp), " '%F'" );
		}
		if ( !cmd->c )
			cmd->c = const_cast( char*, cmdVi );
	}
	return cmd->c;
}

const char *pfmConfEditorCmd ()
{
	return _pfmConfEditorCmd( cmdtbl+PFM_CMD_EDT_IDX );
}

/*
 *	File viewer command, may override the builtin viewer.
 */
__DF(
static const char *_pfmConfViewerCmd,(cmd), x_cmdrecord *cmd )
{
	/* NULL or empty means use builtin viewer */
	if ( !cmd->c ) cmd->c = "";
	return cmd->c;
}

const char *pfmConfViewerCmd ()
{
	return _pfmConfViewerCmd( cmdtbl+PFM_CMD_VIE_IDX );
}

/*
 *	Mail user agent command.
 */
__DF(
static const char *_pfmConfMailerCmd,(cmd), x_cmdrecord *cmd )
{
	if ( !cmd->c )
	{	char *cp = getenv("MAILER");
		if (cp && *cp)
		{	cmd->c = (char*)malloc(strlen(cp)+6);
			if ( cmd->c )
				strcat( strcpy(cmd->c,cp), " '%F'" );
		}
		if ( !cmd->c )
			cmd->c = const_cast( char*, cmdMail );
	}
	return cmd->c;
}

const char *pfmConfMailerCmd ()
{
	return _pfmConfMailerCmd( cmdtbl+PFM_CMD_MAI_IDX );
}

/*
 *	Command to print a selected file.
 */
__DF(
static const char *_pfmConfPrintCmd,(cmd), x_cmdrecord *cmd )
{
	return (cmd->c ? cmd->c:cmdLpr);
}

const char *pfmConfPrintCmd ()
{
	return _pfmConfPrintCmd( cmdtbl+PFM_CMD_PRN_IDX );
}

/*
 *	Program to copy (recursively) a directory to (if possible onto)
 *	another.
 *	Note: here %D and %A stand for from- and to-directory respcetive !
 */
__DF(
static const char *_pfmConfCpdirCmd,(cmd), x_cmdrecord *cmd )
{
	return (cmd->c ? cmd->c:cmdCpdir);
}

const char *pfmConfCpdirCmd ()
{
	return _pfmConfCpdirCmd( cmdtbl+PFM_CMD_CPD_IDX );
}

/*
static const char *pfm_pc_shell = 0;

const char *pfmConfShell ()
{
	if (!pfm_pc_shell)
	{	if ( (pfm_pc_shell = getenv("SHELL")) )
			pfm_pc_shell = *pfm_pc_shell ? strdup(pfm_pc_shell):0;
		if (!pfm_pc_shell) pfm_pc_shell = "/bin/sh";
	}
	return pfm_pc_shell;
}
*/

#ifdef SUPPORT_INET_URL

__DF(
static const char *_pfmConfUrlHdlrCmd,(cmd), x_cmdrecord *cmd )
{
	if ( !cmd->c )
	{	char *cp = getenv("BROWSER");

		if (cp && *cp)
		{	cmd->c = (char*)malloc(strlen(cp)+8);
			if ( cmd->c )
				strcat( strcpy(cmd->c,cp), " '%U%F'" );
		}
		if ( !cmd->c )
			cmd->c = const_cast( char*, cmdLynx );
	}
	return cmd->c;
}

const char *pfmConfUrlHdlrCmd ()
{
	return _pfmConfUrlHdlrCmd( cmdtbl+PFM_CMD_URL_IDX );
}

/*
 *	commands to support urls...
 */
__DF(
static const char *_pfmConfWwwCmd,(cmd), x_cmdrecord *cmd )
{
	return (cmd->c ? cmd->c:_pfmConfUrlHdlrCmd(cmd));
}

const char *pfmConfWwwCmd ()
{
	return _pfmConfWwwCmd( cmdtbl+PFM_CMD_WWW_IDX );
}

__DF(
static const char *_pfmConfFtpCmd,(cmd), x_cmdrecord *cmd )
{
	return (cmd->c ? cmd->c:cmdFtp);
}

const char *pfmConfFtpCmd ()
{
	return _pfmConfFtpCmd( cmdtbl+PFM_CMD_FTP_IDX );
}

__DF(
static const char *_pfmConfTelnetCmd,(cmd), x_cmdrecord *cmd )
{
	return (cmd->c ? cmd->c:cmdTelnet);
}

const char *pfmConfTelnetCmd ()
{
	return _pfmConfTelnetCmd( cmdtbl+PFM_CMD_TEL_IDX );
}
#endif


/********************************************************************
*	handle enter-key-bindings.
********************************************************************/

/*
 *	perform enter-key bindings
 */
__DF(
int pfmMnuEnterkey,(menu), x_menubar *menu )
{
	x_filedat *cf = pfmCwdCurrentfile();
	if (cf)
	{	x_enterbinding *ep;

		for (ep = p_enter_tbl; ep; ep = ep->next)
		{	int k;
			if (ep->pat
				&& !pfmMatchname(ep->pat,cf->nm))
				continue;

			pfmFileEncodemode( ep->att, &k, cf->st.mode );
			k &= ~S_ISVTX;
			if ((cf->st.mode & ~S_ISVTX) != k) continue;

			if (cf->st.mtime < ep->beg ||
			    ep->end < cf->st.mtime) continue;

			switch ( ep->typ )
			{	case 'C':
					if ( ep->act.cmd )
					{	char *p;
						k = 2+pfmVExpandsize( ep->act.cmd, cf->nm,
						                      (char*)0, (char*)0 );
						p = (char*)alloca(k);
						strcpy( p, ep->act.cmd );
						pfmVExpandargs( p, k, cf->nm, (char*)0, (char*)0 );
						pfmExecshell( p, EXECSH_WAIT );
					}
					break;
				case 'K':
					if ( isprint(ep->act.key) )
					{	x_menubar	*pm;
						short	u;
						k = ep->act.key;
						u = k & 0xff80 ? k:toupper(k);
						for (pm = menu; pm->key && pm->fct; pm++)
							if (k == pm->key || u == pm->key)
							{	if (pm->fct != pfmMnuEnterkey) /* ! */
									(void)(*pm->fct)(menu);
								break;
							}
					}
			}
			break;
		}
	}
	return 0;
}


/********************************************************************
*	update enter-key-bindings.
********************************************************************/
/*
 *	release table of enter-key-bindings (to rebuild from conf-file).
 */
__DF(
static void pfmConfReleasentertable,(ptbl), x_enterbinding **ptbl )
{
	x_enterbinding *t1 = *ptbl;
	while ( t1 )
	{	x_enterbinding *t2 = t1->next;
		#if 0 /* t1, pat, act.cmd allocated in one chunk ! */
		if (t1->pat && *t1->pat)
			free(t1->pat);
		if ('C' == t1->typ && t1->act.cmd && *t1->act.cmd)
			free(t1->act.cmd);
		#endif
		free(t1);
		t1 = t2;
	}
	*ptbl = 0;
}


/*
 *	allocate memory for and save enter-key-binding
 *	(for configuration by menue and from file).
 */
__DF(
static int pfmConfAppendenterbinding,( val ), const x_enterbinding *val )
{
	x_enterbinding *entry;
	/*
	 *	allocating the entry in one chunk gives at most one failure
	 *	and hence at most one "return ENOMEM" without the need
	 *	to keep track of and eventually free() other parts ;-)
	 */
	int i = sizeof(*val);
	if ( val->pat ) i += strlen(val->pat)+1;
	if (val->typ != 'K') i += strlen(val->act.cmd)+1;

	if ( (entry = (x_enterbinding*)malloc(i)) )
	{
		char *s = (char*)(entry+1);
		*entry = *val;
		entry->next = 0;

		if (val->typ != 'K')
		{	strcpy( entry->act.cmd = s, val->act.cmd );
			s = strchr(s,0) + 1;
		}
		if ( val->pat )
			strcpy( entry->pat = s, val->pat );

		if ( !p_enter_tbl )
			p_enter_tbl = entry;
		else
		{	x_enterbinding *x = p_enter_tbl;
			while ( x->next ) x = x->next;
			x->next = entry;
		}
		return 0;
	}
	return (-ENOMEM);
}


/********************************************************************
*	read configuration from file
********************************************************************/
/*
 *	find 1st non-white-space character.
 */
__DF(
static char *pfmConfSkipws,(s), char *s )
{
	while ( isspace(*s) ) ++s;
	return s;
}

/*
 *	trim trailing white-space.
 */
__DF(
static char *pfmConfClipwstail,(s), char *s )
{
	char *x = strchr(s,0)-1;
	while (s <= x && isspace(*x))
		*x-- = 0;
	return s;
}

/*
 *	evaluate boolean flag from a configuration-file-entry.
 */
__DF(
static bool pfmConfScanboolean,( tag, def, end ),
const char *tag __j__
bool       def __j__
const char **end )
{
	static struct
	{	const char *tag;
		bool       val;
	}
	on_off_tbl[] =
	{	{true_txt, 1}, {false_txt, 0},
		{"t",    1}, {"f",     0},
		{"1",    1}, {"0",     0},
	/*	{"true", 1}, {"false", 0}, */
		{"on",   1}, {"off",   0},
		{"an",   1}, {"aus",   0},
	};
	int  i;
	for (i = 0; i < sizeof(on_off_tbl)/sizeof(*on_off_tbl); ++i)
	{	const char *ref = on_off_tbl[i].tag;
		if (pfmStrcasematch(tag,ref) && !isalnum(tag[strlen(ref)]))
		{	if ( end ) *end = tag+strlen(ref);
			return on_off_tbl[i].val;
		}
	}
	if ( end ) *end = tag;
	return def;
}

/*
 *	error-message for one configuration-file-line.
 */
__DF(
static void pfmConfInvalidlinemessage,( nerr, nline, s ),
int         nerr __j__
int         nline __j__
const char *s )
{
	if (10 == nerr)
		pfmTermErrormessage( more_errors_txt, (char*)0 );
	else if (nerr < 10)
	{	char buf[80];
		int  i;
		sprintf( buf, "read conf: invalid line %d: ", nline );
		i = strlen(buf);
		for ( ; i < sizeof(buf)-1 && *s; ++i, ++s)
			buf[i] = isprint(*s) ? *s:'?';
		buf[i] = 0;
		pfmTermErrormessage( buf, (char*)0 );
	}
}

/*
 *	evaluate dotfile-mode from a configuration-file-entry.
 */
__DF(
static int pfmConfScandotfilemode,( id, val ),
const char *id __j__
char       *val  )
{
	if ( !strcasecmp(id,_S_show_dot_files.id) )
	{	int  i;
		bool b = _S_show_dot_files.b;
		_S_show_dot_files.b = pfmConfScanboolean( val, 1, (const char**)0 );
		_S_show_dot_files.r = 1;
		if (b != _S_show_dot_files.b) /* apply, if changed */
			pfmCwdReread();
		return 1;
	}
	return 0;
}

#ifdef NCURSES_MOUSE_VERSION
__DF(
static int pfmConfScanmousemode,( id, val ),
const char *id __j__
char       *val  )
{
	if ( !strcasecmp(id,_S_enable_mouse.id) )
	{	extern int pfmTermFixmouse();
		int  i;
		bool b = _S_enable_mouse.b;
		_S_enable_mouse.b = pfmConfScanboolean( val, 0, (const char**)0 );
		_S_enable_mouse.r = 1;
		if (b != _S_enable_mouse.b) /* apply, if changed */
			pfmTermFixmouse();
		return 1;
	}
	return 0;
}
#else
#define pfmConfScanmousemode(id,val) (!strcasecmp(id,"mouse"))
#endif

/*
 *	evaluate a command from a configuration-file-entry.
 */
__DF(
static int pfmConfScaninsertcmdtblentry,( id, val ),
const char *id __j__
char       *val  )
{
	int i;

	for (i = 0; i < sizeof(cmdtbl)/sizeof(*cmdtbl); ++i)
		if ( !strcasecmp(id,cmdtbl[i].id) )
		{	bool isstatic = pfmConfIsstaticstr(cmdtbl[i].c);
			if (val && *val)
			{	char *p = cmdtbl[i].c;
				if (!p || strcmp(p,val))
					p = isstatic ? strdup(val):pfmStrdup(val,p);
				if (p)
				{	cmdtbl[i].c = p;
					cmdtbl[i].r = 1;
					return 1;
				}
				return (-ENOMEM);
			}
			else /* undefined or empty: reset to default */
			{	if (cmdtbl[i].c && !isstatic)
					free(cmdtbl[i].c);
				cmdtbl[i].c = 0;
				cmdtbl[i].r = 1;
				return 1;
			}
		}
	return 0;
}


/*
 *	evaluate a enter-key-binding from a configuration-file-entry.
 */
__DF(
static time_t pfmConfScandate,(val), char *val )
{
	char   *r;
	time_t t = strtol( val, &r, 0 );
	if ( *pfmConfSkipws(r) )
	{	struct tm s;
		if ( pfmScandateandtime(val,&s,&t) )
			return (time_t)-1;
	}
	return t;
}

__DF(
static int pfmConfScaninsertenterbinding,( id, val ),
const char *id __j__
char       *val  )
{
	x_enterbinding tmp;
	char  *p, *s;
	int   i;

	if ( strcasecmp(id,"enter") ) return 0;
	/*
	 *	wildcard-pattern
	 */
	if ( !(s = strchr(val,',')) ) return (-EINVAL);
	*s = 0;

	for (p = pfmConfClipwstail(val); *p; ++p)
		if ('\\' == *p)
		{	char *r;
			char c = strtol( p+1, &r, 8 );
			if (!c || p+1 == r) goto invalid;
			*p = c;
			strcpy( p+1, r );
		}
	memset( &tmp, 0, sizeof(tmp) );
	if (*val) tmp.pat = val;
	/*
	 *	file mode
	 */
	p = pfmConfSkipws(s+1);
	if ( !(s = strchr(p,',')) ) goto invalid;
	*s = 0;
	pfmConfClipwstail(p);
	memcpy( tmp.att, p, sizeof(tmp.att)-1 );
	tmp.att[sizeof(tmp.att)-1] = 0;
	if ( pfmFileEncodemode(tmp.att,&i,0) ) goto invalid;
	/*
	 *	modification time limits
	 */
	p = s+1;
	if ( !(s = strchr(p,',')) ) goto invalid;
	*s = 0;
	if ((tmp.beg = pfmConfScandate(p)) == (time_t)-1) goto invalid;

	p = s+1;
	if ( !(s = strchr(p,',')) ) goto invalid;
	*s = 0;
	if ((tmp.end = pfmConfScandate(p)) == (time_t)-1) goto invalid;
	/*
	 *	action type
	 */
	p = pfmConfSkipws(s+1);
	if ( !(s = strchr(p,',')) ) goto invalid;
	*s = 0;
	tmp.typ = *pfmConfClipwstail(p);
	if (p[1] || tmp.typ != 'C' && tmp.typ != 'K') goto invalid;
	/*
	 *	action key/command
	 */
	pfmConfClipwstail( p = pfmConfSkipws(s+1) );
	if ( !*p ) goto invalid;

	if (tmp.typ != 'K')
		tmp.act.cmd = p;
	else
		if ( p[1] )
		{	char *r;
			tmp.act.key = strtol( p, &r, 8 );
			if (!tmp.act.key || *r) goto invalid;
		}
		else
		{	tmp.act.key = *p;
			if ( !isprint(tmp.act.key) ) goto invalid;
		}
	/*
	 *	allocate the entry
	 */
	i = pfmConfAppendenterbinding(&tmp);
	return (i ? i:1);

	invalid:
	strcpy( val, p ); /* prepare for error-message */
	return (-EINVAL);
}


/*
 *	read configuration file.
 */
__DF(
static int pfmConfReadconffile,( fname ), const char *fname )
{
	FILE *fp = fopen( fname, "r" );

	if ( fp )
	{	char buf[128];
		int  errs = 0, nl = 0;

		pfmConfReleasentertable(&p_enter_tbl);

		while ( fgets(buf,sizeof(buf),fp) )
		{	char *p, *s;
			int  i;
			++nl;
			strtok( buf, "\n\r" );
			switch ( *(p = pfmConfSkipws(buf)) )
			{	case 0: case '!': case '#': case '/':
					continue;
			}
			if (!(s = strtok(p,".*")) || strcasecmp(s,"pfm"))
			{	pfmConfInvalidlinemessage( errs, nl, buf );
				++errs;
				continue;
			}
			p = strtok( (char*)0, ":" ); /* identifier */
			if (p && *p) pfmConfClipwstail(p);

			if (!p || !*p)
			{	pfmConfInvalidlinemessage( errs, nl, buf );
				++errs;
				continue;
			}
			if ( (s = strtok((char*)0,"")) ) /* value */
				if ( *(s = pfmConfSkipws(s)) ) pfmConfClipwstail(s);
			/*
			 *	at this point identifier and value contain
			 *	no more leading or trailing whitespace.
			 */
			i = pfmConfScaninsertcmdtblentry( p, s );
			if (!i) i = pfmConfScaninsertenterbinding( p, s );
			if (!i) i = pfmConfScandotfilemode( p, s );
			if (!i) i = pfmConfScanmousemode( p, s );
			if (i <= 0)
			{	if (-ENOMEM == i)
					pfmConfInvalidlinemessage( errs, nl, strerror(-i) );
				else
				{	strcat( strcat( strcpy(buf,p), ": " ), s );
					pfmConfInvalidlinemessage( errs, nl, buf );
				}
				++errs;
			}
		}
		errs |= ferror(fp);
		fclose(fp);
		return errs;
	}
	return 1;
}

__DF(
int pfmConfRead,( fname ), const char *fname )
{
	return pfmConfReadconffile( fname = pfmConffile(fname) );
}


/********************************************************************
*	write configuration file
********************************************************************/
/*
 *	write enterkey-binding entries.
 */
__DF(
static int pfmConfWritenterbindings,( fp ), FILE *fp )
{
	x_enterbinding *ep;

	fputs( enterbd_cmt_txt, fp );

	for (ep = p_enter_tbl; ep; ep = ep->next)
	{	char *s;
		fputs( "pfm.Enter: ", fp );
		if ( ep->pat )
		{	int  octal = 0;
			for (s = ep->pat; *s; ++s)
			{	if (',' == *s || '\\' == *s || !isgraph(*s))
					octal = 8;
				else /* if ( octal ) */
					if (*s < '0' || '7' < *s)
						octal = 0;
				if ( octal )
					fprintf( fp, "\\0%o", (int)*s );
				else
					fputc( *s, fp );
			}
		}
		fprintf( fp, ", %s, 0x%lx, 0x%lx, %c, ",
			ep->att, (long)ep->beg, (long)ep->end, ep->typ );
		if (ep->typ != 'C')
			fprintf( fp,
				isprint(ep->act.key) ? "%c\n":"0%o\n",
				ep->act.key );
		else
			fprintf( fp, "%s\n", ep->act.cmd );
	}
	return ferror(fp);
}

/*
 *	write full configuration file.
 */
__DF(
static int __pfmConfWrite,( fp ), FILE *fp )
{
	time_t t = time(&t);
	int i;

	fprintf( fp, "! %s\n", ctime(&t) );

	if ( !_S_show_dot_files.r ) fputc( '!', fp );
	fprintf( fp,
		"pfm.%s: %s\n\n",
		_S_show_dot_files.id,
		pfmConfShowdotfiles() ? true_txt:false_txt );
	#ifdef NCURSES_MOUSE_VERSION
		if ( !_S_enable_mouse.r ) fputc( '!', fp );
		fprintf( fp, "pfm.%s: %s\n\n",
			_S_enable_mouse.id,
			pfmConfMouseinput() ? true_txt:false_txt );
	#endif

	for (i = 0; i < sizeof(cmdtbl)/sizeof(*cmdtbl); ++i)
	{	char *cmd = cmdtbl[i].c;
		if ( !cmd )
			cmd = const_cast( char*, (*cmdtbl[i].q)(cmdtbl+i) );
		/*
		 *	We print values, that were not explicitely changed (from file,
		 *	by menu), only as comment.  Otherwise default-values or values
		 *	derived from a maybe coincidental environment will be taken as
		 *	intentional settings and silently override any new environment
		 *	(EDITOR,PAGER,...) when the file is read again.
		 */
		if ( !cmdtbl[i].r ) fputc( '!', fp );
		fprintf( fp, "pfm.%s: %s\n", cmdtbl[i].id, cmd );
	}
	pfmConfWritenterbindings(fp);
	return ferror(fp);
}

__DF(
int pfmConfWrite,( fname ), const char *fname )
{
	FILE *fp;

	fname = pfmConffile(fname);

	if ( strchr(fname,'/') )
	{	char *p = (char*) alloca(strlen(fname)+1);
		if (p)
		{	*strrchr( strcpy(p,fname), '/' ) = 0;
			pfmFileMkdir(p);
		}
	}
	if ( (fp = fopen(fname,"w")) )
	{	int i = __pfmConfWrite(fp);
		fclose(fp);
		return i;
	}
	return 1;
}


/********************************************************************
*	configuration by menu
********************************************************************/

static int changed_by_menu_entry = 0;

__DF(
static int pfmMnuConfHelp,(m), x_menubar *m )
{
	(void)m;
	return pfmHelp("configuration");
}

/*
 *	read configuration from user defined file
 */
__DF(
static int pfmMnuConfReadconfig,(m), x_menubar *m )
{
	char	dest[CWD_MAX_LNG];

	if ( !pfmTermGetstrl(pfm_C_file_name,sizeof(dest),dest,
	                     pfmConffile((char*)0),(char*)0) )
		pfmConfRead(dest);
	(void)m;
	return 0;
}

/*
 *	save configuration to user defined file
 */
__DF(
static int pfmMnuConfSaveconfig,(m), x_menubar *m )
{
	char	dest[CWD_MAX_LNG];

	if ( !pfmTermGetstrl(pfm_C_destination,sizeof(dest),dest,
	                     pfmConffile((char*)0),(char*)0) )
		if ( pfmConfWrite(dest) )
			pfmTermErrormessage( strerror(errno), (char*)0 );
	(void)m;
	return 0;
}


/*
 *	ask user for dotfile-mode
 */
__DF(
static int pfmMnuConfDotfilemode,(m), x_menubar *m )
{
	bool b = _S_show_dot_files.b;
	_S_show_dot_files.b = pfmTermYesNo( show_dotfiles_txt, b );
	if (b != _S_show_dot_files.b)
	{	_S_show_dot_files.r = 1;
		++changed_by_menu_entry;
		/* and apply this change */
		pfmCwdReread();
	}
	(void)m;
	return 0;
}

#ifdef NCURSES_MOUSE_VERSION
/*
 *	ask user for ncurses' mouse-mode
 */
__DF(
static int pfmMnuConfMousemode,(m), x_menubar *m )
{
	extern int pfmTermFixmouse();
	bool b = _S_enable_mouse.b;
	_S_enable_mouse.b = pfmTermYesNo( use_mouse_txt, b );
	if (b != _S_enable_mouse.b)
	{	_S_enable_mouse.r = 1;
		++changed_by_menu_entry;
		/* and apply this change */
		pfmTermFixmouse();
	}
	(void)m;
	return 0;
}
#endif


/********************************************************************
*	defining enterkey-bindings by menu
********************************************************************/

/*
 *	let the user define enterkey-bindings interactively.
 */
#define USE_ENTERKEY_SUB_MENU 0
/*
 *	display all the enterkey-bindings
 */
__DF(
static int pfmMnuConfListenterbinding,(m), x_menubar *m )
{
	FILE *fp;

	pfmPreExecshell(0);
	
	if ( (fp = popen(pfmConfPagerPrg(),"w")) )
	{
		#if USE_ENTERKEY_SUB_MENU
			pfmConfWritenterbindings(fp);
		#else
			__pfmConfWrite(fp);
		#endif
		pclose(fp);
	}
	pfmPostExecshell(EXECSH_WAIT|EXECSH_NORRD);
	clearok( stdscr, TRUE );
	(void)m;
	return 0;
}

/*
 *	change existing enterkey-binding(s)
 */
__DF(
static int pfmMnuConfEditenterbinding,(m), x_menubar *m )
{
	FILE *fp;
	char *fn = tmpnam((char*)0);

	if ( !(fp = fopen(fn,"w")) )
		pfmTermErrormessage( strerror(errno), (char*)0 );
	else
	{	int i = pfmConfWritenterbindings(fp);
		fclose(fp);
		if (i)
			pfmTermErrormessage( strerror(errno), (char*)0 );
		else
		{	const char *p1 = pfmConfEditorCmd();
			char *p2 = (char*) alloca(strlen(p1)+1);
			p1 = strcpy( p2, p1 );
			if ( (p2 = strchr(p1,'&')) ) *p2 = 0;
			pfmFileEditfile( p1, 0, fn );
			if ( pfmConfReadconffile(fn) )
				pfmTermErrormessage( pfm_C_hit_enter, (char*)0 );
		}
		unlink(fn);
	}
	(void)m;
	return 0;
}

#if USE_ENTERKEY_SUB_MENU
#if 0
/*
 *	ask for enterkey-binding components
 */
__DF( static int pfmConfEnterbindquery,( ent ), x_enterbinding *ent )
{
	const char *p = pfmFileGetWildnamepattern( (char*)0, ent->pat );
	if ( !p ) return 1;

	ent->pat = const_cast( char*, p );
	
	p = pfmFileGetModepattern( (char*)0, *ent->att ? ent->att:(char*)0 );
	if ( !p ) return 1;

	strncpy( ent->att, p, sizeof(ent->att)-1 );

	ent->beg = pfmFileGetDateandtime( not_before_txt, (struct tm*)0 );
	if ((time_t)-1 == ent->beg) return 1;

	ent->end = pfmFileGetDateandtime( not_after_txt, (struct tm*)0 );
	if ((time_t)-1 == ent->end) return 1;

	/* ask for key or command ... */
	return 0;
}
__DF( static int pfmMnuConfNewenterbinding,(m), x_menubar *m )
{
	x_enterbinding entry;
	memset( &entry, 0, sizeof(entry) );
	if ( !pfmConfEnterbindquery(&entry) )
	{	pfmConfAppendenterbinding(&entry);
		++changed_by_menu_entry;
	}
	(void)m;
	return 0;
}
#else
#define pfmMnuConfNewenterbinding pfmMnuConfEditenterbinding
#endif
#define pfmMnuConfDelenterbinding pfmMnuConfEditenterbinding

/*
 *	Enterkey-bindings menue
 */
__DF(
static int pfmConfEnterbindconfmnu,(m), x_menubar *m )
{
	static x_menubar	entermenu[] =
	{
		{ 'A', "b"BOLD"Ack ", pfmMnuCancelmenu },
		/*
		{ 'Q', "", pfmMnuCancelmenu },
		{ 'X', "", pfmMnuCancelmenu },
		*/
		{ 'H', BOLD"Help ", pfmMnuConfHelp },
		{ '?', "", pfmMnuConfHelp },
		{ 'L', BOLD"List ", pfmMnuConfListenterbinding },
		{ 'N', BOLD"New ", pfmMnuConfNewenterbinding },
		{ 'D', BOLD"Delete ", pfmMnuConfDelenterbinding  },
		{ 'E', BOLD"Edit ", pfmMnuConfEditenterbinding   },
		{ 0, "", pfmMnuNop }
	};
	static char	reentry = 0;

	if ( !reentry )
	{	reentry = 1;
		pfmMnuRunmenu( entermenu, (char*)0 );
		reentry = 0;
	}
	(void)m;
	return 0;
}
#endif /* USE_ENTERKEY_SUB_MENU */


/********************************************************************
*	defining commands by menu
********************************************************************/
/*
 *	Protect command variables from expansion by pfmExpandargs().
 *	Command patterns are stored literally, so their variables must be
 *	protected, when they are presented for interactive edition and
 *	following user-input expansion.
 *	(protect: \ -> \\, % -> \%, ~ -> \~ 
 *	expand:  \\ -> \, \% -> %, \~ -> ~)
 */
__DF(
static char  *pfmConfEscargs,( expr, size ),
char	expr[/*size*/] __j__
int 	size )
{
	if (size > 0)
	{
		char *buf;
		int  n;

		for (n = 0, buf = expr; *buf; ++n, ++buf)
			switch (*buf)
			{	case '~':
				case '\\':
				case pfm_fwildc_start:
					++n;
			}
		if ( (buf = (char*)alloca(n+2)) )
		{	char *p = buf, *q = expr;
			do
			{	switch (*q)
				{	case '~':
					case '\\':
					case pfm_fwildc_start:
						*p++ = '\\';
				}
				*p = *q++;
			}
			while (*p++);
			strncpy( expr, buf, size );
			expr[size-1] = 0;
		}
	}
	return expr;
}

/*
 *	let the user define commands (editor,... URL-handlers, ... )
 */
__DF(
static int pfmConfCommandquery,(cmd), x_cmdrecord *cmd )
{
	char buf[256];
	/* e.g.
	 * rc-file/memory: cmd -f ~/.rc '%F\~'
	 * presented as:   cmd -f \~/.rc '\%F\\\~'
	 */
	/* (*cmd->q)(cmd), not cmd->c ! */
	strncpy( buf, (*cmd->q)(cmd), sizeof(buf) ) [sizeof(buf)-1] = 0;
	pfmConfEscargs( buf, sizeof(buf) );

	if ( !pfmTermGetstrl(cmd->t,sizeof(buf),buf,buf,(char*)0) )
	{	/* e.g.
		 * user input:      %D/%F -f ~/.rc '\%F\\\~'
		 * rc-file/memory:  /usr/bin/cmd -f /home/me/.rc '%F\~'
		 * user input:      %D/%F -f \~/.rc '\%F\\\~'
		 * rc-file/memory:  /usr/bin/cmd -f ~/.rc '%F\~'
		 * both invoked as: /usr/bin/cmd -f /home/me/.rc 'myfile.c~'
		 */
		pfmExpandargs( buf, sizeof(buf) );
		pfmConfScaninsertcmdtblentry( cmd->id, buf );
		++changed_by_menu_entry;
		return 0;
	}
	return (-1);
}

__DF(
static int _pfmMnuConfDocdirquery,(m), x_menubar *m )
{
	pfmConfCommandquery(cmdtbl+PFM_CMD_DOC_IDX);
	(void)m;
	return 0;
}

__DF(
static int _pfmMnuConfPagerquery,(m), x_menubar *m )
{
	pfmConfCommandquery(cmdtbl+PFM_CMD_PGR_IDX);
	(void)m;
	return 0;
}

__DF(
static int _pfmMnuConfEditorcmdquery,(m), x_menubar *m )
{
	pfmConfCommandquery(cmdtbl+PFM_CMD_EDT_IDX);
	(void)m;
	return 0;
}

__DF(
static int _pfmMnuConfViewercmdquery,(m), x_menubar *m )
{
	pfmConfCommandquery(cmdtbl+PFM_CMD_VIE_IDX);
	(void)m;
	return 0;
}

__DF(
static int _pfmMnuConfMailhandlerquery,(m), x_menubar *m )
{
	pfmConfCommandquery(cmdtbl+PFM_CMD_MAI_IDX);
	(void)m;
	return 0;
}

__DF(
static int _pfmMnuConfPrintcmdquery,(m), x_menubar *m )
{
	pfmConfCommandquery(cmdtbl+PFM_CMD_PRN_IDX);
	(void)m;
	return 0;
}

__DF(
static int _pfmMnuConfCpdirquery,(m), x_menubar *m )
{
	pfmConfCommandquery(cmdtbl+PFM_CMD_CPD_IDX);
	(void)m;
	return 0;
}
#ifdef SUPPORT_INET_URL

__DF(
static int _pfmMnuConfWwwhandlerquery,(m), x_menubar *m )
{
	pfmConfCommandquery(cmdtbl+PFM_CMD_WWW_IDX);
	(void)m;
	return (-1); /* back to main conf-menue */
}

__DF(
static int _pfmMnuConfFtphandlerquery,(m), x_menubar *m )
{
	pfmConfCommandquery(cmdtbl+PFM_CMD_FTP_IDX);
	(void)m;
	return (-1); /* back to main conf-menue */
}

__DF(
static int _pfmMnuConfTelnethandlerquery,(m), x_menubar *m )
{
	pfmConfCommandquery(cmdtbl+PFM_CMD_TEL_IDX);
	(void)m;
	return (-1); /* back to main conf-menue */
}

__DF(
static int _pfmMnuConfDefaulthandlerquery,(m), x_menubar *m )
{
	pfmConfCommandquery(cmdtbl+PFM_CMD_URL_IDX);
	(void)m;
	return (-1); /* back to main conf-menue */
}


/*
 *	URL-handler menue
 */
__DF(
static int pfmConfUrlhdlrconfmnu,(m), x_menubar *m )
{
	static x_menubar	confmenu[] =
	{
		{ 'A', "b"BOLD"Ack ", pfmMnuCancelmenu },
		/*
		{ 'Q', "", pfmMnuCancelmenu },
		{ 'X', "", pfmMnuCancelmenu },
		*/
		{ 'H', BOLD"Help ", pfmMnuConfHelp },
		{ '?', "", pfmMnuConfHelp },
		{ 'W', BOLD"Www ", _pfmMnuConfWwwhandlerquery },
		{ 'F', BOLD"Ftp ", _pfmMnuConfFtphandlerquery },
		{ 'T', BOLD"Telnet ", _pfmMnuConfTelnethandlerquery },
		{ 'U', "other "BOLD"Url ", _pfmMnuConfDefaulthandlerquery },
		{ 0, "", pfmMnuNop }
	};
	static char	reentry = 0;

	if ( !reentry )
	{	reentry = 1;
		pfmMnuRunmenu( confmenu, (char*)0 );
		reentry = 0;
	}
	(void)m;
	return 0;
}
#endif


/********************************************************************
*	Main configuration menue
********************************************************************/
__DF(
int pfmMnuConfigure,(m), x_menubar *m )
{
	static x_menubar	confmenu[] =
	{
		{ 'A', "b"BOLD"Ack ", pfmMnuCancelmenu },
		/*
		{ 'Q', "", pfmMnuCancelmenu },
		{ 'X', "", pfmMnuCancelmenu },
		*/
		{ 'H', BOLD"Help ", pfmMnuConfHelp },
		{ '?', "", pfmMnuConfHelp },
		{ 'R', BOLD"Read ", pfmMnuConfReadconfig },
		{ 'W', BOLD"Write ", pfmMnuConfSaveconfig },
		#if USE_ENTERKEY_SUB_MENU
		{ 'B', BOLD"Bind ", pfmConfEnterbindconfmnu },
		#else
		{ 'L', BOLD"List ", pfmMnuConfListenterbinding },
		{ 'B', BOLD"Bind ", pfmMnuConfEditenterbinding   },
		#endif
		{ 'O', "d"BOLD"Ot ", pfmMnuConfDotfilemode },
		#ifdef NCURSES_MOUSE_VERSION
		{ 'S', "mou"BOLD"Se ", pfmMnuConfMousemode },
		#endif
		{ 'D', BOLD"Doc ", _pfmMnuConfDocdirquery },
		{ 'P', BOLD"Pager ", _pfmMnuConfPagerquery },
		{ 'E', BOLD"Editor ", _pfmMnuConfEditorcmdquery },
		{ 'V', BOLD"Viewer ", _pfmMnuConfViewercmdquery },
		{ 'M', BOLD"Mailer ", _pfmMnuConfMailhandlerquery },
		{ 'N', "pri"BOLD"Nt ", _pfmMnuConfPrintcmdquery },
		{ 'C', BOLD"Cpdir ", _pfmMnuConfCpdirquery },
		#ifdef SUPPORT_INET_URL
		{ 'U', BOLD"Url ", pfmConfUrlhdlrconfmnu },
		#endif
		{ 0, "", pfmMnuNop }
	};
	static char	reentry = 0;

	if ( !reentry )
	{	reentry = 1;
		changed_by_menu_entry = 0;
		pfmMnuRunmenu( confmenu, (char*)0 );
		if (changed_by_menu_entry > 0)
			if ( pfmTermYesNo(save_conf_txt,FALSE) )
				pfmMnuConfSaveconfig(confmenu);
		reentry = 0;
	}
	(void)m;
	return (-1); /* back up to main menue */
}


/***** eof *********************************************************/
