/***********************************************************************
* Copyright (c) 1995-2003 by peter fuerst
*
*	This program is free software; you can redistribute it and/or modify
*	it - as explained in the accompaning file COPYING - under the terms
*	of the GNU General Public License as published by the Free Software
*	Foundation; either version 2, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	Author: Peter Frst, post@pfrst.de
*
*	@(#)pfmcwd.c	10/25/95
*
***********************************************//*$preferred tabstop=4*/

#include <stdio.h>
#include <stddef.h>
#include <dirent.h>
#include <limits.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <ctype.h>
#include <sys/param.h>

#include "pfm.h"
#include "pfmstat.h"
#ifdef __COHERENT__
 __DC( int strcasecmp,(const char*,const char*) );
#endif

#ifdef __COHERENT_286__ /* || defined(?) || ... */
 #define FIXED_FILE_TBL
#else
 #undef FIXED_FILE_TBL
#endif


/***** define & typedef ********************************************/

#define _s_islnk(data)	S_ISLNK((data).st.lmode)

#define BOLD "\007" /* use BELl to emphasize the next character */


/***** forward declaration *****************************************/

__DC( static int  __pfmCwdChangedir,(const char*,unsigned,unsigned) );
__DC( static int  pfmexecsort,(bool) );
__DC( static int  pfmdrawfileline,( int, x_filedat*, int ) );
__DC( static void pfmdrawfilesrate,(bool) );
__DC( static void pfmdrawmarkedcount,(bool) );
__DC( static void pfmdrawtimes,(bool) );

/***** local data **************************************************/

static char	fileswinheader[] =
    "  filename            size     modify-date/time   attribute ";
/*	 ^0^2                  ^22       ^32       ^42     ^50      ^59*/

static char	files_txt[]   = "files";
static char	blocks_txt[]  = "blocks";
static char	bytes_txt[]   = "bytes";
static char	used_txt[]    = "used";
static char	marked_txt[]  = "marked";
static char	_10U_fmt[]    = "%10u";
static char	_10LU_fmt[]   = "%10lu";
static char	_10S_S_fmt[]  = "%10s %s";
static char	_10LU_S_fmt[] = "%10lu %s";
static char	_U_U_filesrd_fmt[] = "Only %u files of %u read";
static char	no_filetable_txt[] = "Cannot allocate minimal filetable";
static char	cannot_chdir_txt[] = "Cannot chdir()";
#define	resfilesrate	1000	/* Resolution is promille	*/
#define	cursorcolumn	1

/*
 *	On small systems it is often better to set aside some static memory
 *	for direntry.filetable with a reasonable number of entries.
 *	Dynamically allocating filetable on demand there may sometimes lead
 *	to fragmentation of the data-segment, which may not leave enough
 *	consecutive memory for a direntry.filedata request.  While a
 *	fixed table might allow to read in all files (although it takes
 *	away more memory a priori and imposes a fixed limit on the number
 *	of files).
 */
#ifdef FIXED_FILE_TBL
 #define FIXED_FILE_TBL_LIMIT\
	((sizeof(int*)/sizeof(short)) * (sizeof(int*)/sizeof(short)) * 750)

 static x_filedat *fixed_file_table[FIXED_FILE_TBL_LIMIT+2];
#endif

static struct
{
	x_filedat	**filetable;	/* table of file-data-block adresses	*/
	unsigned	tblimit;		/* number of slots in filetable	*/
	x_filedat	*filedata;		/* array of these blocks and ...	*/
	size_t		datsize;		/* ... its size in bytes	*/
	unsigned	datcnt;			/* number of files in filedata (<= tfiles)	*/
	unsigned	start;			/* display filetable[start..]	*/
	unsigned	current;		/* [current] currently under the cursor	*/
	struct
	{	unsigned	tfiles;		/* number of files in CWD	*/
		unsigned	dotscnt;	/* number of 'dot-files'	*/
		fsize_t		ubytes;		/* sum of all file sizes in CWD	*/
		fsize_t		tbytes;		/* total number of bytes occupied in CWD	*/
		long		blocks;		/* total number of blocks occupied in CWD	*/
	}			num;
	struct
	{	unsigned	files;		/* number of marked files	*/
		fsize_t		ubytes;		/* sum of sizes of marked files	*/
		fsize_t		tbytes;		/* total number of bytes occupied by mark..	*/
		long		blocks;		/* number of blocks occupied by mark..	*/
	}			marked;
	struct stat	stat;			/* filled with stat(".")	*/
	unsigned	bsize;			/* stat.st_blksize or BSIZE	*/
	unsigned	flag;
	unsigned	sort;			/* type of current sorting 'A',...'U'	*/
	int 		filesrate;		/* current file's relative position */
}
direntry =
{
	#ifdef FIXED_FILE_TBL
		fixed_file_table, FIXED_FILE_TBL_LIMIT,
	#else
		(x_filedat**)0, 0,
	#endif
	(Void*)0, 0, 0, 0, 0, {0,}, {0,}, {0,}, 0, 0, 0, 0,
};

static char __cwd_root[] = "/";
static char __cwd_dot[] = ".";


/**** functions ****************************************************/

#if 0 /* not needed for now... */
__DF(
unsigned pfmCwdSetCurrentfile,( idx, rfrsh ), unsigned idx __j__ bool rfrsh )
{
	unsigned curr = direntry.current;

	if (idx < direntry.datcnt && idx != curr)
	{	direntry.current = idx;
		if (rfrsh) pfmCwdWinDrawfileswin(TRUE);
	}
	return curr;
}
#endif
/*
 *	return data of the file at the cursor position
 */
x_filedat  *pfmCwdCurrentfile ()
{
	if ( direntry.filetable )
		if (direntry.current < direntry.datcnt)
			return direntry.filetable[direntry.current];
	return 0;
}

/*
 *	return data of the "idx"th file-table-entry (sorted by current method)
 */
__DF(
x_filedat  *pfmCwdDirtblentry,( idx ), unsigned idx )
{
	if ( direntry.filetable )
		if (idx < direntry.datcnt)
			return direntry.filetable[idx];
	return 0;
}

#if 0 /* not needed for now... */
/*
 *	return data of the "idx"th directory-entry (in order as received by readdir)
 *	(can't use filetable for this, since it may be sorted in a different order)
 */
__DF(
x_filedat  *pfmCwdDirentry,( idx ), unsigned idx )
{
	x_filedat	*p = 0;
	if (idx < direntry.datcnt)
	{	unsigned	n = 0;
		while (n < idx)
		{	unsigned i = x_filedat_align(strlen(p->nm) + sizeof(x_filedat));
			p = (x_filedat*) ((char*)p + i);
			n++;
		}
	}
	return p;
}
#endif

/*
 *	mark a file
 */
__DF(
static void addmarkedsize,( p ), x_filedat *p )
{
	unsigned b = ((direntry.bsize-1)+p->st.size) / direntry.bsize;
	direntry.marked.blocks += b;
	direntry.marked.tbytes += direntry.bsize*b;
	direntry.marked.ubytes += p->st.size;
	direntry.marked.files++;
}

__DF(
static void submarkedsize,( p ), x_filedat *p )
{
	unsigned b = ((direntry.bsize-1)+p->st.size) / direntry.bsize;
	direntry.marked.blocks -= b;
	direntry.marked.tbytes -= direntry.bsize*b;
	direntry.marked.ubytes -= p->st.size;
	direntry.marked.files--;
}


__DF(
static __inline__ int ismarkable,( p ), x_filedat *p )
{
	if ( S_ISREG(p->st.mode) )
		return 1;
	if ( S_ISDIR(p->st.mode) )
	{	register char *s = p->nm;
		if (*s != '.' || (*++s && (*s != '.' || *++s)))
			return 1;
	}
	if ( S_ISLNK(p->st.mode) )
		return 1; /* symbolic link, pointing nowhere */
	return 0;
}

__DF(
int pfmCwdMarkon,( p, msk ), x_filedat *p __j__ unsigned msk )
{
	if (p)
	{	if (ismarkable(p))
		{	/* msk &= (X_FILEDAT_MARK_FLAG | X_FILEDAT_RMND_FLAG); */
			if ((msk & X_FILEDAT_MARK_FLAG)
				&& !(p->flag & X_FILEDAT_MARK_FLAG))
					addmarkedsize(p);
			p->flag |= msk;
		}
		return p->flag;
	}
	return 0;
}

__DF(
int pfmCwdMarkoff,( p, msk ), x_filedat *p __j__ unsigned msk )
{
	if (p)
	{	/* msk &= (X_FILEDAT_MARK_FLAG | X_FILEDAT_RMND_FLAG); */
		if ((msk & X_FILEDAT_MARK_FLAG)
			&& (p->flag & X_FILEDAT_MARK_FLAG))
			if (direntry.marked.files)
				submarkedsize(p);
		return (p->flag &= ~msk);
	}
	return 0;
}

/*
 *	mark the file at the cursor position
 */
x_filedat *pfmCwdMarkcurrentfile ()
{
	x_filedat	*p = 0;

	if ( direntry.filetable )
	{	p = direntry.filetable[direntry.current];
		if (p)
		#if 0
		if (ismarkable(p))
		#else
		if (*p->nm != '.' || (p->nm[1] && (p->nm[1] != '.' || p->nm[2])))
		#endif
		{	unsigned oldmrk = p->flag & X_FILEDAT_MARK_FLAG;
			p->flag ^= X_FILEDAT_MARK_FLAG;
			if (oldmrk)
				submarkedsize(p);
			else
				addmarkedsize(p);
			pfmdrawmarkedcount(TRUE);
			pfmdrawfileline( direntry.current - direntry.start + 1,
							p, A_STANDOUT );
			pfmTermSubwinRefresh( pfm_FilesWindow, TRUE );
		}
	}
	return p;
}

/********************************************************************
*	read contents of the current working directory
********************************************************************/

__DF(
static void  pfmalignlines,( numlines ), int numlines )
{	/*
	 *	Always start at the begin, if all entries fit into the window
	 */
	if (direntry.datcnt < numlines) direntry.start = 0;
	/*
	 *	Adjust 'direntry.current' to the range of table[0..(datcnt-1)]
	 */
	if (direntry.datcnt <= direntry.current)
		if ((direntry.current = direntry.datcnt) > 0)
			direntry.current--;
	/*
	 *	If necessary, move the window to cover 'direntry.current'
	 *	(line 0: header, lines 1..(numlines-1): file-entries)
	 */
	if (direntry.current < direntry.start)
		direntry.start = direntry.current;

	if (direntry.current > direntry.start+numlines-2)
		direntry.start = direntry.current - numlines + 2;
}


__DF(
static x_filedat  *pfmfindnode,( node ), ino_t node )
{
	if ( direntry.filetable )
	{
		unsigned	i;

		for (i = 0; i < direntry.datcnt; i++)
			if (direntry.filetable[i] &&
				node == direntry.filetable[i]->st.ino)
				return direntry.filetable[i];
	}
	return 0;
}


static void  deletedirentry ()
{
	x_filedat **tbl = direntry.filetable;
	if (direntry.filedata)
		 free(direntry.filedata);
	if (tbl)
	#ifdef FIXED_FILE_TBL
		memset( tbl, 0, sizeof(x_filedat*)*FIXED_FILE_TBL_LIMIT );
	#else
		free(tbl);
	#endif
	memset( &direntry, 0, sizeof(direntry) );
	#ifdef FIXED_FILE_TBL
		/* don't forget to restore fixed entries ! */
		if ( (direntry.filetable = tbl) )
			direntry.tblimit = FIXED_FILE_TBL_LIMIT;
	#endif
}


__DF(
static x_filedat* reallocdirentry,( name, flag, pstat ),
const char *name __j__
short	   flag __j__
const struct _pfmstat
           *pstat )
{
	if (direntry.datcnt < direntry.tblimit)
	{
		/* ensure proper alignment of direntry.filedata and all entries */
		size_t incr = x_filedat_align(strlen(name) + sizeof(x_filedat));
		x_filedat *pd = realloc(direntry.filedata, direntry.datsize+incr);

		if ( pd )
		{	if (pd != direntry.filedata)
			{	int	i, k = (char*)pd - (char*)direntry.filedata;
				for (i = 0; i < direntry.datcnt; i++)
					direntry.filetable[i] =
						(x_filedat*) ((char*)direntry.filetable[i] + k);
			}
			direntry.filedata = pd;
			pd = (x_filedat*) ((char*)pd + direntry.datsize);
			direntry.datsize += incr;
			direntry.filetable[direntry.datcnt] = pd;
			direntry.datcnt++;
			strcpy( pd->nm, name );
			pd->flag = flag;
			pd->st   = *pstat;
			return pd;
		}
	}
	return 0;
}


int pfmCwdReread ()
{
	__DC(extern DIR *opendir,(const char*));
	__DC(extern int closedir,(DIR*));
	__DC(extern struct dirent *readdir,(DIR*));
	#define ALLOCNT(num) ((num)+2)

	struct _pfmstat	ptmp;
	unsigned	start = direntry.start;
	unsigned	current = direntry.current;
	char    	sort = direntry.sort;
	bool        nodotfiles = !pfmConfShowdotfiles();

	/* release allocated memory and zero out filetable */
	deletedirentry();

	if ( !pfmFileStat(__cwd_dot,&direntry.stat,&ptmp) )
	{
		DIR	*cwd;

		direntry.bsize = BLKSIZE(direntry.stat);
		if (!direntry.bsize) direntry.bsize = 512;
		direntry.num.ubytes = direntry.stat.st_size;
		direntry.num.blocks = ((direntry.bsize-1)+direntry.stat.st_size)
							/ direntry.bsize;
		direntry.num.tbytes = direntry.bsize * direntry.num.blocks;

		if ( (cwd = opendir(__cwd_dot)) )
		{
			#ifndef FIXED_FILE_TBL
			/*
			 *	The idea to allocate memory for FILE_TBL_MAX entries and
			 *	to shrink this buffer later, is not as good as it looks
			 *	at first sight. The freed memory usually will not (with
			 *	paged memory enabled) or can not (due to other allocations
			 *	beyond its end, when no paging available) be released to
			 *	the system. This means that our process would keep a lot
			 *	of unused memory, in worst case as long as it lives. So we
			 *	only allocate what we need (or not much more) !
			 */
			for (direntry.tblimit = 1;  readdir(cwd); )
				direntry.tblimit++;
			rewinddir(cwd);

			#define FILE_TBL_MAX \
				(sizeof(int*) < 4 ? 2000 : (sizeof(size_t) < 4 ? 16000:64000))

			if (direntry.tblimit > FILE_TBL_MAX)
				direntry.tblimit = FILE_TBL_MAX;
			#undef FILE_TBL_MAX

			direntry.filetable =
				(x_filedat**) calloc( ALLOCNT(direntry.tblimit),
				                      sizeof(x_filedat*) );

			if ( !direntry.filetable )
				direntry.tblimit = 0;
			else
			#endif /* FIXED_FILE_TBL */
			{
				struct dirent	*q;

				while ( (q = readdir(cwd)) )
				{
					struct stat	stmp;
					short	flag = 0;

					if (!q->d_ino) continue; 	/* invalid/deleted entry	*/

					if ( !pfmFileStat(q->d_name,&stmp,&ptmp) )
					{
						if ( pfmFileMultilinked(stmp.st_mode,stmp.st_nlink) )
							flag |= X_FILEDAT_MLNK_FLAG;

						if ( !pfmfindnode(stmp.st_ino) )
						{	unsigned bsize, blocks;
							#ifdef HAS_SYMLINK_STAT
								if ( S_ISLNK(ptmp.lmode) )
									lstat( q->d_name, &stmp );
							#endif
							#if 1
								if ( !(bsize = BLKSIZE(stmp)) )
							#endif
									bsize = direntry.bsize;
							blocks = ((bsize-1)+stmp.st_size) / bsize;
							direntry.num.ubytes += stmp.st_size;
							direntry.num.blocks += blocks;
							direntry.num.tbytes += bsize * blocks;
						}
					}
					direntry.num.tfiles++;

					if (nodotfiles && '.' == *q->d_name)
						#if 0
						if ( q->d_name[1] ) /* not . */
							if (q->d_name[1] != '.' || q->d_name[2]) /* not .. */
						#endif
							{	direntry.num.dotscnt++;
								continue;
							}
					/* limit already reached ?	*/
					if (direntry.datcnt < direntry.tblimit)
						reallocdirentry( q->d_name, flag, &ptmp );

				}
				#ifndef FIXED_FILE_TBL
				/*
				 *	now shrink the filetable, if appropriate:
				 */
				if (direntry.datcnt < direntry.tblimit)
				{	/* do not call realloc( ptr, 0 ) ! */
					direntry.tblimit = direntry.datcnt;
					direntry.filetable = (x_filedat**)
								realloc(direntry.filetable,
								        ALLOCNT(direntry.tblimit)
								        *sizeof(x_filedat*));
				}
				#endif
				direntry.sort = sort ? sort:'N';
				pfmexecsort(FALSE);
				/*
				 *	for savety adjust 'current' as soon as possible !
				 *	(only the paranoid survive...  ;-)
				 */
				direntry.current =
					current < direntry.datcnt
					? current
					: (direntry.datcnt ? direntry.datcnt-1:0);

				direntry.start = start > current ? current:start;

				#if DEBUG
					printf( "%u entries (of %u files, %u hidden) in table\n",
							direntry.datcnt, direntry.num.tfiles,
							direntry.num.dotscnt );
					printf( "stored %u bytes of attributes from %u files\n",
							direntry.datsize, direntry.datcnt );
					printf( " %s\n", pfm_C_hit_any_key );
					while (pfmTermGetch() < 1);
				#endif
			}
			closedir(cwd);
		}
	}
	#ifndef FIXED_FILE_TBL
	if ( !direntry.filetable )
	{	direntry.tblimit = 1;
		direntry.filetable =
			(x_filedat**) calloc( ALLOCNT(direntry.tblimit),
			                      sizeof(x_filedat*) );
		if ( !direntry.filetable )
			pfmExit( 1, "pfmCwdReread(): ", no_filetable_txt, (char*)0 );
	}
	#endif
	if (direntry.datcnt < 1)
	{	static char nm[] = ".";
		reallocdirentry( nm, 0, &ptmp );
	}
	pfmTermRedraw();	/* ... and display it	*/

	if (direntry.datcnt+direntry.num.dotscnt < direntry.num.tfiles)
	{	char buf[40];
		sprintf( buf, _U_U_filesrd_fmt,
				direntry.datcnt, direntry.num.tfiles );
		pfmTermErrormessage( strerror(ENOMEM), buf, (char*)0 );
	}
	#undef ALLOCNT
	return 0;
}


int pfmCwdRereadChanged ()
{
	struct stat	stmp;

	if ( !stat(__cwd_dot,&stmp) )
		if ( memcmp(&direntry.stat, &stmp, sizeof(stmp)) )
		{	pfmCwdReread();
			return 1;
		}
	pfmdrawtimes(TRUE);
	pfmCwdWinDrawfileswin(TRUE);
	return 0;
}


/********************************************************************
*	draw the current working directory's window
********************************************************************/

static int pfmcalcfilesrate ()
{
	if (direntry.num.tfiles)
	{	long	i = resfilesrate;
		return (i*(direntry.current+1)) / direntry.num.tfiles;
	}
	return 0;
}


__DF(
static void pfmdrawfileswinheader,( hdr ), const char *hdr )
{
	int	ww = wwidth(pfm_FilesWindow);
	int	i;

	wstandout(pfm_FilesWindow);
	wmove( pfm_FilesWindow, 0, 0 );
	for (i = 0; i < ww && *hdr; i++, hdr++)
		waddch( pfm_FilesWindow, *hdr );
	for (; i < ww; i++)
		waddch( pfm_FilesWindow, ' ' );
	wstandend(pfm_FilesWindow);
}


#ifdef __MAY_HAVE_FSIZE_64__
__DF(
static int fsize_t_long,( sz, lg ), fsize_t sz __j__ long *lg )
{
	int ret = 0;

	if (sizeof(sz) > sizeof(long))	/* trigger optimization... */
		if ((sz & 0xffffffffL) != sz)
		{	sz >>= 10;
			ret = 1;
		}
	*lg = (long)sz;
	return ret;
}
#else	/* avoid overhead alltogether */
 #define fsize_t_long(sz, lg) (*(lg)=(sz), 0)
#endif

__DF(
static void pfmbuildfileline,( data, buf, endp ),
x_filedat *data __j__
char	*buf __j__
char	**endp )
{
	static char	mmm[][4] = { "Jan", "Feb",
	#ifdef DEUTSCH
		"Mae", "Apr", "Mai", "Jun", "Jul", "Aug", "Sep", "Okt", "Nov", "Dez" };
	#else
		"Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	#endif
	char	*bp;
	struct tm	*ts = localtime(&data->st.mtime);
	int	i = data->flag & (X_FILEDAT_MARK_FLAG | X_FILEDAT_RMND_FLAG), n;
	long sz;
	/*
	.123456789.123456789.123456789.123456789.123456789.123456789.
	0.........1.........2.........3.........4.........5.........6
	*~NameNameNameNa>123456789   Mon-dd-yy  hh:mm:ss  drwxrwxrwx
	*/
	buf[0] = i ? (i & X_FILEDAT_MARK_FLAG ? '*':'.') : ' ';
	i = data->flag & X_FILEDAT_MLNK_FLAG;
	buf[1] = _s_islnk(*data) ? '~':(i ? '`':' ');

	for (i = 0, bp = buf+2; i < 14; i++, bp++)
	{	unsigned c = data->nm[i];
		if (c && (!isascii(c) || !isprint(c))) c = '?';
		if ( !(*bp = c) ) break;
	}
	*endp = bp;
	if (data->nm[i])	/* strlen(data->nm) > 14	*/
		*bp++ = '>';
	else while (i++ < 15)
		*bp++ = ' ';

	i = fsize_t_long( data->st.size, &sz );
	if ( S_ISDIR(data->st.mode) )
	{	bp = " <%*lu%.1s>%-8s";
		n = 1;
	}
	else
	{	bp = "%*lu%.1s%-3s";
		n = i ? 8:9;
	}
	sprintf( buf+17, bp, n, sz, i ? "K":"", "" );
	sprintf( buf+29, "%.3s-%02u-%02u  %02u:%02u:%02u  ",
			mmm[ts->tm_mon], ts->tm_mday, ts->tm_year % 100,
			ts->tm_hour, ts->tm_min, ts->tm_sec );
	pfmFileDecodemode( data->st.mode, buf+50 );
}


__DF(
static int pfmdrawfileline,( line, data, attr ),
int			line  __j__
x_filedat	*data __j__
int			attr )
{
	if (line < 0 || wheight(pfm_FilesWindow) <= line) return 0;

	wmove( pfm_FilesWindow, line, 0 );

	if (data)
	{	char	buf[64], *bp;
		int	ww = wwidth(pfm_FilesWindow);
		int	i;

		pfmbuildfileline( data, buf, &bp );
		if (ww < sizeof(buf)) buf[ww<3 ? 3:ww] = 0;

		waddch( pfm_FilesWindow, buf[0] );
		waddch( pfm_FilesWindow, buf[1] );
		if (attr)
		{	wstandout(pfm_FilesWindow);
			i = *bp;
			*bp = 0;
		}
		waddstr( pfm_FilesWindow, buf+2 );
		if (attr)
		{	const char *cp;
			wstandend(pfm_FilesWindow);
			*bp = i;
			if (bp < strchr(buf,0))
				waddstr( pfm_FilesWindow, bp );
			cp = strlen(data->nm) <= 14
				? fileswinheader : pfmStrtail(data->nm,ww);
			pfmdrawfileswinheader(cp);
		}
	}
	else
	{	waddstr( pfm_FilesWindow, "  ?" );
		wclrtoeol(pfm_FilesWindow);
	}
	wmove( pfm_FilesWindow, line, cursorcolumn );
	pfmTermAdjParentCursor(pfm_FilesWindow);
	return 0;
}


__DF(
int pfmCwdWinDrawfileswin,( rfrsh ), bool rfrsh )
{
	unsigned	i, line;

	pfmalignlines(wheight(pfm_FilesWindow));

	direntry.filesrate = pfmcalcfilesrate();
	pfmdrawfilesrate(TRUE);

	leaveok( pfm_FilesWindow, TRUE );
	werase(pfm_FilesWindow);
	pfmdrawfileswinheader(fileswinheader);

	for (i = direntry.start, line = 1;
		i < direntry.datcnt && line < wheight(pfm_FilesWindow);
		i++, line++)
	{	unsigned	j = direntry.current - i;
		int			attr = j ? A_NORMAL : A_STANDOUT;

		pfmdrawfileline( line, direntry.filetable[i], attr );
	}
	leaveok( pfm_FilesWindow, FALSE );
	pfmCwdWinCursorhome();
	if (rfrsh) pfmTermSubwinRefresh( pfm_FilesWindow, TRUE );

	return 0;
}


/********************************************************************
*	draw notes' window
********************************************************************/
/*
 *   0	[----------------]
 *   2	.......... files
 *   3	.......... blocks
 *   4	.......... bytes
 *   5	.......... used
 *   7	    marked files
 *   8	.......... files
 *   9	.......... bytes
 * h-7	euser:egroup
 * h-6	[user:group]
 * h-5	host.domain
 * h-3	Day Mon DD YYYY
 * h-2	HH:MM:SS
 */
static int noteswindowstartline = 2;
#define NOTESWINDOWMAXLINES	18

__DF(
static void pfmdrawfilesrate,( rfrsh ), bool rfrsh )
{
	char buf[64];
	int	ww = wwidth(pfm_NotesWindow);
	int	k = direntry.filesrate;

	if (sizeof(buf) <= ww) ww = sizeof(buf)-1;
	memset( buf, '-', ww );
	buf[ww-1] = ']';
	buf[k ? 1+(k-1L)*(ww-2)/resfilesrate:1] = '|';
	*buf = '[';
	buf[ww] = 0;
	mvwaddstr( pfm_NotesWindow, 0, 0, buf );
	#if 0
		wclrtoeol(pfm_NotesWindow);
	#endif
	if (rfrsh) wrefresh(pfm_NotesWindow);
}


__DF(
static void pfmdrawtimes,( rfrsh ), bool rfrsh )
{	/*
	..123456789.123456789.12345.
	"Wed Jun 30 21:49:08 1993\n"
	*/
	int	ww = wwidth(pfm_NotesWindow);
	int	l = wheight(pfm_NotesWindow);
	l -= (l < NOTESWINDOWMAXLINES-1 ? 1:2);

	if (ww > 2 && l > 0)
	{	char buf[32];
		time_t	t = time(&t);
		char	*p = ctime(&t);
		ww -= 2;
		if (sizeof(buf) <= ww) ww = sizeof(buf)-1;

		sprintf( buf, "%8.8s", p+11 );
		buf[ww] = 0;
		mvwaddstr( pfm_NotesWindow, l, 2, buf );
		if (--l > 0)
		{	sprintf( buf, "%11.11s %4.4s", p, p+20 );
			buf[ww] = 0;
			mvwaddstr( pfm_NotesWindow, l, 2, buf );
		}
		if (rfrsh) wrefresh(pfm_NotesWindow);
	}
}

static void pfmdrawusrgrphost ()
{
	int	ww = wwidth(pfm_NotesWindow);
	int	l = wheight(pfm_NotesWindow);
	l -= (l < NOTESWINDOWMAXLINES-1 ? 4:5);

	if (ww > 2 && l > 0)
	{	char buf[40];
		ww -= 2;
		if (sizeof(buf) <= ww) ww = sizeof(buf)-1;

		strcat( strcat( strcpy(buf,pfm_Hostname), "." ), pfm_Domainname );
		buf[ww] = 0;
		mvwaddstr( pfm_NotesWindow, l, 2, buf );

		if (--l > 0)
		{	int	egid = getegid();
			int	gid = getgid();
			int	euid = geteuid();
			int	uid = getuid();

			if (uid != euid || gid != egid)
			{	char *p;
				if ( *pfm_Usrname )
					sprintf( buf, "(%s:", pfm_Usrname );
				else
					sprintf( buf, "(%u:", uid );
				p = strchr(buf,0);
				if ( *pfm_Grpname )
					sprintf( p, "%s)", pfm_Grpname );
				else
					sprintf( p, "%u)", gid );
				buf[ww] = 0;
				mvwaddstr( pfm_NotesWindow, l, 2, buf );
			}
			if (--l > 0)
			{	if ( *pfm_eUsrname )
					sprintf( buf, "%s:", pfm_eUsrname );
				else
					sprintf( buf, "%u:", euid );
				if ( *pfm_eGrpname )
					strcat( buf, pfm_eGrpname );
				else
					sprintf( strchr(buf,0), "%u", egid );
				buf[ww] = 0;
				mvwaddstr( pfm_NotesWindow, l, 2, buf );
			}
		}
	}
}

__DF(
static void pfmdrawmarkedcount,( rfrsh ), bool rfrsh )
{
	int	h = wheight(pfm_NotesWindow);
	int l = noteswindowstartline+5;
	if (l < h)
	{	char buf[40];
		int	ww = wwidth(pfm_NotesWindow);
		if (sizeof(buf) <= ww) ww = sizeof(buf)-1;

		sprintf( buf, _10S_S_fmt, marked_txt, files_txt );
		buf[ww] = 0;
		mvwaddstr( pfm_NotesWindow, l, 0, buf );
		if (++l < h)
		{	sprintf( buf, _10LU_S_fmt, (long)direntry.marked.files, files_txt );
			buf[ww] = 0;
			mvwaddstr( pfm_NotesWindow, l, 0, buf );
		}
		if (++l < h)
		{	long sz;
			int i = fsize_t_long( direntry.marked.tbytes, &sz );
			sprintf( buf, i ? "%9luK %s":_10LU_S_fmt, sz, bytes_txt );
			buf[ww] = 0;
			mvwaddstr( pfm_NotesWindow, l, 0, buf );
		}
		if (rfrsh) wrefresh(pfm_NotesWindow);
	}
}


__DF(
int pfmCwdWinDrawnoteswin,( rfrsh ), bool rfrsh )
{
	char buf[40];
	int	ww = wwidth(pfm_NotesWindow);
	int	h = wheight(pfm_NotesWindow);
	int l;

	if (h < NOTESWINDOWMAXLINES) noteswindowstartline = 1;
	if (sizeof(buf) <= ww) ww = sizeof(buf)-1;

	leaveok( pfm_NotesWindow, TRUE );
	werase(pfm_NotesWindow);

	pfmdrawfilesrate(FALSE);
	pfmdrawusrgrphost();
	pfmdrawtimes(FALSE);

	if ((l = noteswindowstartline) < h)
	{	sprintf( buf, _10LU_S_fmt, (long)direntry.num.tfiles, files_txt );
		buf[ww] = 0;
		mvwaddstr( pfm_NotesWindow, l, 0, buf );
	}
	if (++l < h)
	{	sprintf( buf, _10LU_S_fmt, (long)direntry.num.blocks, blocks_txt );
		buf[ww] = 0;
		mvwaddstr( pfm_NotesWindow, l, 0, buf );
	}
	if (++l < h)
	{	long sz;
		int i = fsize_t_long( direntry.num.tbytes, &sz );
		sprintf( buf, i ? "%9luK %s":_10LU_S_fmt, sz, bytes_txt );
		buf[ww] = 0;
		mvwaddstr( pfm_NotesWindow, l, 0, buf );
	}
	if (++l < h)
	{	long sz;
		int i = fsize_t_long( direntry.num.ubytes, &sz );
		sprintf( buf, i ? "%9luK %s":_10LU_S_fmt, sz, used_txt );
		buf[ww] = 0;
		mvwaddstr( pfm_NotesWindow, l, 0, buf );
	}
	pfmdrawmarkedcount(FALSE);

	leaveok( pfm_NotesWindow, FALSE );
	pfmCwdWinCursorhome();

	if (rfrsh) pfmTermSubwinRefresh( pfm_NotesWindow, TRUE );

	return 0;
}


/********************************************************************
*	move in the current working directory's window
********************************************************************/

/*
 *	scroll files window: move cursor one line up
 */
int pfmCwdWin1lineUp ()
{
	#define SCROLL_SUBWIN_OK 0
	pfmalignlines(wheight(pfm_FilesWindow));

	if (0 < direntry.current)
	{
		pfmdrawfileline( direntry.current-direntry.start+1,
						direntry.filetable[direntry.current], A_NORMAL );

		if (--direntry.current < direntry.start)
		{	direntry.start = direntry.current;
			#if SCROLL_SUBWIN_OK /* scrolling is incompatible with ...	*/
				wmove( pfm_FilesWindow, 1, 0 );
				winsertln(pfm_FilesWindow);
			#else /* ... refreshing the parent window !	*/
				pfmCwdWinDrawfileswin(TRUE);
				return 0;
			#endif
		}
		pfmdrawfileline( direntry.current-direntry.start+1,
						direntry.filetable[direntry.current],
						A_STANDOUT );
		#if SCROLL_SUBWIN_OK /* scrolling is incompatible with ...	*/
			/* can't use pfmTermSubwinRefresh() here anyway */
			wrefresh(pfm_FilesWindow);
		#else
			/* pfmTermSubwinRefresh( pfm_FilesWindow, TRUE ); */
			wrefresh(pfm_FilesWindow); /* that will do here */
		#endif
		direntry.filesrate = pfmcalcfilesrate();
		pfmdrawfilesrate(TRUE);
	}
	return 0;
}

/*
 *	scroll files window: move cursor one line down
 */
int pfmCwdWin1lineDown ()
{
	pfmalignlines(wheight(pfm_FilesWindow));

	if (direntry.num.tfiles && direntry.current < direntry.datcnt-1)
	{
		unsigned  m = wheight(pfm_FilesWindow) - 2;
		pfmdrawfileline( direntry.current-direntry.start+1,
						direntry.filetable[direntry.current], A_NORMAL );

		if (direntry.start + m < ++direntry.current)
		{	direntry.start = direntry.current - m;
			#if SCROLL_SUBWIN_OK /* scrolling is incompatible with ...	*/
				wmove( pfm_FilesWindow, 1, 0 );
				wdeleteln(pfm_FilesWindow);
			#else /* ... refreshing the parent window !	*/
				pfmCwdWinDrawfileswin(TRUE);
				return 0;
			#endif
		}
		pfmdrawfileline( direntry.current-direntry.start+1,
						direntry.filetable[direntry.current],
						A_STANDOUT );
		#if SCROLL_SUBWIN_OK /* scrolling is incompatible with ...	*/
			/* can't use pfmTermSubwinRefresh() here anyway */
			wrefresh(pfm_FilesWindow);
		#else
			/* pfmTermSubwinRefresh( pfm_FilesWindow, TRUE ); */
			wrefresh(pfm_FilesWindow); /* that will do here */
		#endif
		direntry.filesrate = pfmcalcfilesrate();
		pfmdrawfilesrate(TRUE);
	}
	return 0;
	#undef SCROLL_SUBWIN_OK
}

/*
 *	scroll files window: move cursor one page up
 */
int pfmCwdWinPageUp ()
{
	unsigned  n = wheight(pfm_FilesWindow) - 2;

	if (direntry.start > n)
	{	direntry.current -= n;
		direntry.start -= n;
	}
	else
	{	direntry.current -= direntry.start;
		direntry.start = 0;
	}
	pfmCwdWinDrawfileswin(TRUE);
	return 0;
}

/*
 *	scroll files window: move cursor one page down
 */
int pfmCwdWinPageDown ()
{
	unsigned  n = wheight(pfm_FilesWindow) - 2;

	if (direntry.start + n < direntry.datcnt)
	{	direntry.start += n;
		direntry.current += n;
		if (direntry.current >= direntry.datcnt)
			direntry.current = direntry.datcnt - 1;
	}
	pfmCwdWinDrawfileswin(TRUE);
	return 0;
}

/*
 *	start files window at the first entry
 */
int pfmCwdWinTopscroll ()
{
	unsigned  n = wheight(pfm_FilesWindow) - 2;

	direntry.current = n < direntry.current ? n:0;
	direntry.start = 0;
	pfmCwdWinDrawfileswin(TRUE);
	return 0;
}

/*
 *	start files window at the tail of entries
 */
int pfmCwdWinEndscroll ()
{
	unsigned m = wheight(pfm_FilesWindow);

	direntry.start = direntry.datcnt < m-1 ? 0:direntry.datcnt+1-m;

	direntry.current =
	#if 0
		direntry.current < direntry.start ? direntry.start :
	#endif
		direntry.datcnt-1;

	pfmCwdWinDrawfileswin(TRUE);
	return 0;
}

#ifdef NCURSES_MOUSE_VERSION
/*
 *	Directly jump to a given line in the files window
 */
__DF(
void pfmCwdWinJumptoline,( x, y ), int x __j__ int y )
{
	int n = wheight(pfm_FilesWindow);
	if (0 < y && y < n) /* 'start' is at line 1 */
	{
	#if 1
		/* Do it the least expensive or... */
		pfmalignlines(n); /* only the paranoid survive */
		pfmdrawfileline( direntry.current-direntry.start+1,
						direntry.filetable[direntry.current],
						A_NORMAL );
		direntry.current = direntry.start + y - 1;
		pfmalignlines(n);
		pfmdrawfileline( direntry.current-direntry.start+1,
						direntry.filetable[direntry.current],
						A_STANDOUT );
		wrefresh(pfm_FilesWindow);
		direntry.filesrate = pfmcalcfilesrate();
		pfmdrawfilesrate(TRUE);
	#else
		/* ...the simplest way ? */
		direntry.current = direntry.start + y - 1;
		pfmCwdWinDrawfileswin(TRUE);
	#endif
	}
}
#endif


/********************************************************************
*	change the current working directory
********************************************************************/

static char *__cwd_cwdir = __cwd_dot;
static char *__cwd_altdir = __cwd_root;
#define IS_ALLOC_DIR(dir) ((dir) && (dir)!=__cwd_dot && (dir)!=__cwd_root)

const char *pfmCwdCWD ()
{
	return __cwd_cwdir;
}

const char *pfmCwdAltDir ()
{
	return __cwd_altdir;
}

__DF(
static int  __pfmCwdChangedir,( dir, start, idx ),
const char	*dir __j__
unsigned  	start __j__
unsigned  	idx )
{
	if ( dir )
	{
		if (strcmp(__cwd_dot,dir) && strcmp(__cwd_cwdir,dir))
			if ( chdir(dir) )
			{	pfmTermErrormessage( cannot_chdir_txt, dir, (char*)0 );
				return 0;
			}
			else
			{	char *p, buffer[CWD_MAX_LNG];
				p = pfmStrdup( getcwd(buffer,sizeof(buffer)),
								IS_ALLOC_DIR(__cwd_cwdir) ? __cwd_cwdir:(char*)0 );
				__cwd_cwdir = p ? p:__cwd_dot;
				direntry.current = idx;
				direntry.start = start;
			}
		pfmCwdReread(); /* reread also on chdir(CWD) ...	*/
		return 1;
	}
	return 0;
}

int pfmCwdChangedir ()
{
	char	dest[CWD_MAX_LNG];

	if ( !pfmTermGetstrl(pfm_C_dir_pathname,sizeof(dest),dest,__cwd_altdir,(char*)0) )
	{
		char	*oldcwd = __cwd_cwdir;

		pfmExpandargs( dest, sizeof(dest) );

		if (*dest && pfmFileGotoURL(dest,FALSE) > 0)
		{	/* successfully back from URL */
			char *p = strdup(dest);
			if ( p )
			{	if ( IS_ALLOC_DIR(__cwd_altdir) )
					free(__cwd_altdir);
				__cwd_altdir = p;
			}
		}
		else /* not an URL or an error occurred */
		{
			static unsigned __cwd_altidx = 0, __cwd_altbas = 0;
			unsigned current = direntry.current;
			unsigned start = direntry.start;
			unsigned idx, bas;

			if ( strcmp(dest,__cwd_altdir) )
				idx = bas = 0;
			else
			{	idx = __cwd_altidx;
				bas = __cwd_altbas;
			}
			__cwd_cwdir = __cwd_dot; /* protect from 'free(__cwd_cwdir)' */

			if (__pfmCwdChangedir(*dest ? dest:__cwd_root, bas, idx) <= 0)
				__cwd_cwdir = oldcwd;
			else
			{	if ( IS_ALLOC_DIR(__cwd_altdir) )
					free(__cwd_altdir);
				__cwd_altdir = oldcwd;
				__cwd_altidx = current;
				__cwd_altbas = start;
			}
		}
	}
	return 0;
}

/*
 *	change to the directory at cursor position and back
 */
static unsigned __cwd_prvidx = 0;
static unsigned __cwd_prvbas = 0;

__DF(
int pfmCwdChangenamedir,( name ), const char *name )
{
	if ( !strcmp("..",name) ) return pfmCwdPreviousdir();

	__cwd_prvidx = direntry.current;
	__cwd_prvbas = direntry.start;
	__pfmCwdChangedir( name, 0, 0 );
	return 0;
}

int pfmCwdPreviousdir ()
{
	int r = 0;

	if ( __cwd_cwdir[1] )
	{	r = __pfmCwdChangedir( "..", __cwd_prvbas, __cwd_prvidx );
		if (r > 0) __cwd_prvidx = __cwd_prvbas = 0;
	}
	return r;
}


/********************************************************************
*	sort current directory by name, time, suffix, ...
********************************************************************/

static struct
{
	__DC(int (*cmp),(const char*,const char*) );
	char prm[20];
}
sortcompare = { strcmp, BOLD"+ignorecase " };

__DF(
static int  pfmignorecasesort,(m), x_menubar *m )
{
	if (strcmp == sortcompare.cmp)
	{	sortcompare.cmp = strcasecmp;
		sortcompare.prm[1] = '-';
		sortcompare.prm[2] = 'I';
		return -1; /* force redrawing */
	}
	(void)m;
	return 0;
}

__DF(
static int  pfmcasesensesort,(m), x_menubar *m )
{
	if (sortcompare.cmp != strcmp)
	{	sortcompare.cmp = strcmp;
		sortcompare.prm[1] = '+';
		sortcompare.prm[2] = 'i';
		return -1; /* force redrawing */
	}
	(void)m;
	return 0;
}

/*
 *	compare x_filedat entries
 */
__DF(
static int __pfmfiledatcompare,( p1, p2 ), x_filedat **p1 __j__ x_filedat **p2 )
{
	if (!*p1) return (*p2 ? 1:0); /* invalid *p1: *p1 >= *p2	*/

	if (!*p2) return -1;	/* invalid *p2: *p1 < *p2	*/

	switch ( direntry.sort )
	{
		char	*s1, *s2;
		long	dt;
		int		i;

		case 'A':
			dt = (*p2)->st.atime - (*p1)->st.atime;
			goto comparelong;
		case 'C':
			dt = (*p2)->st.ctime - (*p1)->st.ctime;
			goto comparelong;
		case 'M':
			dt = (*p2)->st.mtime - (*p1)->st.mtime;
			goto comparelong;
		case 'S':
			dt = (*p2)->st.size - (*p1)->st.size;
		comparelong:
			return dt ? (dt < 0 ? -1:1) : 0;
		case 'X':
			s1 = strrchr( (*p1)->nm, '.' );
			s2 = strrchr( (*p2)->nm, '.' );
			if (!s1) s1 = "";
			if (!s2) s2 = "";
			if ( (i = (*sortcompare.cmp)(s1,s2)) ) return i;
		case 'N':
			return (*sortcompare.cmp)( (*p1)->nm, (*p2)->nm );
		default:
			if (*p1 == *p2) return 0;
			return (*p1 < *p2 ? -1:1);
	}
}

/*
 *	sort the table of x_filedat entries
 */
__DF(
static int  pfmexecsort,(display), bool display )
{
	#if 0
		direntry.current = direntry.start = 0;
	#endif
	shellsort( direntry.filetable, direntry.datcnt, sizeof(x_filedat*),
				(__DC(int(*),(const void*,const void*)))__pfmfiledatcompare );
	if (display) pfmCwdWinDrawfileswin(TRUE);
	return 0;
}

/*
 *	Menue functions to choose the way of sorting CWD entries
 */
__DF(
static int pfmunsort,(m), x_menubar *m )
{
	direntry.sort = 0;
	pfmexecsort(TRUE);
	(void)m;
	return -1;
}

__DF(
static int  pfmatimesort,(m), x_menubar *m )
{
	direntry.sort = 'A';
	pfmexecsort(TRUE);
	(void)m;
	return -1;
}

__DF(
static int  pfmctimesort,(m), x_menubar *m )
{
	direntry.sort = 'C';
	pfmexecsort(TRUE);
	(void)m;
	return -1;
}

__DF(
static int  pfmmtimesort,(m), x_menubar *m )
{
	direntry.sort = 'M';
	pfmexecsort(TRUE);
	(void)m;
	return -1;
}

__DF(
static int  pfmnamesort,(m), x_menubar *m )
{
	direntry.sort = 'N';
	pfmexecsort(TRUE);
	(void)m;
	return -1;
}

__DF(
static int  pfmsizesort,(m), x_menubar *m )
{
	direntry.sort = 'S';
	pfmexecsort(TRUE);
	(void)m;
	return -1;
}

__DF(
static int  pfmsuffixsort,(m), x_menubar *m )
{
	direntry.sort = 'X';
	pfmexecsort(TRUE);
	(void)m;
	return -1;
}

__DF(
static int pfmMnuCwdHelp,(m), x_menubar *m )
{
	(void)m;
	return pfmHelp("sort");
}

int pfmCwdSortdir ()
{
	static x_menubar  sortdirmenu[] =
	{
		{ A_ESC, BOLD"Help, or sort by ", pfmMnuNop },
		{ 'H', "", pfmMnuCwdHelp },
		{ '?', "", pfmMnuCwdHelp },
		{ 'N', BOLD"Name, ", pfmnamesort },
		{ 'X', "suffi"BOLD"X, ", pfmsuffixsort },
		{ 'S', BOLD"Size, ", pfmsizesort },
		{ 'A', BOLD"Acc/", pfmatimesort },
		{ 'C', BOLD"Chg/", pfmctimesort },
		{ 'M', BOLD"Mod-"BOLD"time, ", pfmmtimesort },
		{ 'T', "", pfmmtimesort },
		{ 'D', BOLD"Don't sort, ", pfmunsort },
		{ '+', sortcompare.prm, pfmignorecasesort },
		{ '-', "", pfmcasesensesort },
		{ 0, "", pfmMnuNop }
	};
	static char	reentry = 0;

	if ( !reentry )
	{	__DC(int (*cmp),(const char*,const char*) );
		reentry = 1;
		do
		{	cmp = sortcompare.cmp;
			pfmMnuRunmenu( sortdirmenu, (char*)0 );
		} while (cmp != sortcompare.cmp);
		reentry = 0;
	}
	return 0;
}


/********************************************************************
*	move the physical cursor to the current line
********************************************************************/

void  pfmCwdWinCursorhome ()
{
	wmove( pfm_FilesWindow,
			1+direntry.current-direntry.start, cursorcolumn );
	pfmTermAdjParentCursor(pfm_FilesWindow);
}


/***** eof *********************************************************/
