/***********************************************************************
* Copyright (c) 1995-2003 by peter fuerst
*
*	This program is free software; you can redistribute it and/or modify
*	it - as explained in the accompaning file COPYING - under the terms
*	of the GNU General Public License as published by the Free Software
*	Foundation; either version 2, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	Author: Peter Frst, post@pfrst.de
*
*	@(#)files.c	10/25/95
*	@(#)dirtree.c
*
***********************************************//*$preferred tabstop=4*/

#include <stdio.h>
#include <stddef.h>
#include <dirent.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/param.h>
#include "termuxx.h"
#include "spawn.h"

#include "pfm.h"
#include "pfmstat.h"
#include "nulltrap.h"

/***** define & typedef ********************************************/

#define _s_islnk(data)	S_ISLNK((data).st.lmode)
#define _s_isdir(data)	S_ISDIR((data).st.mode)


/***** forward declaration *****************************************/

__DC( extern DIR *opendir,(const char*) );
__DC( extern int closedir,(DIR*) );
__DC( extern struct dirent *readdir,(DIR*) );

#if defined(__COHERENT__) /* || defined(__XENIX__) || defined(__SCO__) */
 /* We know at compile-time, that *BUILTIN* is not used. See conf.c. */
 #define HAS_NATIVE_CPDIR_CMD 1
#else
 static int
 __DC( pfmFileBuiltincpdir,(const char*, const char*, const char*) );
#endif


/***** global ******************************************************/


/***** local *******************************************************/

#ifdef DEUTSCH
 static const char rmdir_txt[] = "Verzeichnis mit Inhalt loeschen ?";
 static const char *rmdir1YN[] = {"Einzeln bestaetigen", "Ja", "Nein"};
#else
 static const char rmdir_txt[] = "Remove directory and contents ?";
 static const char *rmdir1YN[] = {"Confirm each", "Yes", "No"};
#endif


/********************************************************************
*	directory-tree operations
********************************************************************/

/*
 *	(Recursively) create a directory
 */
__DF(
int pfmFileMkdir,( dir ), const char *dir )
{
	if (dir && *dir)
	{	char	*p, *s;

		p = (char*) xalloca(strlen(dir)+1);

		for (dir = strcpy(p,dir); p; p = s)
		{	s = strchr(p,'/');
			if (s) *s = 0;
			if (*p)
			{	struct stat sbuf;
				if ( stat(dir,&sbuf) )
				{	int mode = S_IREAD | S_IWRITE | S_IEXEC |
					           S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
					if ( mkdir(dir,mode) )
					{	pfmTermErrormessage(strerror(errno), dir, (char*)0);
						return (-1);
					}
					pfmFileSGetLastdest(dir);
				}
				else if ( !S_ISDIR(sbuf.st_mode) )
				{	pfmTermErrormessage(strerror(ENOTDIR), dir, (char*)0);
					return (-1);
				}
			}
			if (s) *s++ = '/';
		}
	}
	return 0;
}

/*
 *	(Recursively) create a named directory
 */
__DF(
int pfmMnuFileMkdir,(m), x_menubar *m )
{
	char	dest[CWD_MAX_LNG];

	if ( !pfmTermGetstrl(pfm_C_dir_pathname,sizeof(dest),dest,
							pfmFileSGetLastdest((char*)0),(char*)0) )
	{	pfmExpandargs( dest, sizeof(dest) );
		pfmFileMkdir(dest);
		pfmFileSGetLastdest(dest);
		pfmCwdRereadChanged();
	}
	(void)m;
	return (-1);
}


/*
 *	Recursively erase a directory
 *
 *	It's not worthwile to fill this file with a replacement for the
 *	"rm -r" command. rm's syntax is identical on all *U*X, performance
 *	isn't an issue here and it should work best, so just invoke it !
 */
__DF(
static int _pfmFileRmdirectory,( data ), x_filedat *data )
{
	if (data && strcmp(data->nm,".") && strcmp(data->nm,".."))
		if (!_s_isdir(*data))
			pfmTermErrormessage( strerror(ENOTDIR), data->nm, (char*)0 );
		else
		{	char ret[20];
			char *buf = (char*) xalloca(24+strlen(rmdir_txt));
			strcpy( buf, pfmStrtail(data->nm,20) );
			strcat( strcat(buf,": "), rmdir_txt );
			if ( !pfmTermGetstrl(buf,sizeof(ret),ret,rmdir1YN[0],
 			                     rmdir1YN[1],rmdir1YN[2],
 			                     (char*)0) )
			{	char c = toupper(*ret);
				if (c == *rmdir1YN[0] || c == *rmdir1YN[1])
				{	char *rm[4];
					rm[0] = "rm";
					rm[1] = strcasecmp(rmdir1YN[1],ret) ? "-ri":"-r";
					rm[2] = data->nm;
					rm[3] = 0;
					if ( pfmExecargv(rm,0) )
					{	pfmTermErrormessage( strerror(errno), (char*)0 );
						return (-1);
					}
				}
			}
		}
	return 0;
}

/*
 *	Recursively erase the directory at the cursor position
 */
__DF(
int pfmMnuFileRmdir,(m), x_menubar *m )
{
	x_filedat	*info = pfmCwdCurrentfile();

	if (info && !_pfmFileRmdirectory(info))
		pfmCwdRereadChanged();
	(void)m;
	return (-1);
}


/*
 *	Print one entry in the directory-tree
 */
__DF(
static int __pfmFilePrinttreenode,( out, name, islnk, depth ),
FILE       *out __j__
const char *name __j__
bool       islnk __j__
int        depth )
{
	const char *p;

	if (depth > 0)
	{	while (--depth > 0) fputs( "|  ", out );
		fputs( "|--", out );
	}
	for (p = name; *p; ++p)
		fputc( isprint(*p) ? *p:'?', out );

	#if HAS_SYMLINK
		if ( islnk )
		{	char *p;
			int i;
			fputs( " -> ", out );
			p = (char*) xalloca(512);
			i = readlink( name, p, 512 );
			if (i > 0)
			{	p[i] = 0;
				fputs( p, out );
			}
		}
	#else
		(void)islnk;
	#endif
	fputc( '\n', out );
	return 0;
}

/*
 *	Print current-working-directory's tree
 */
__DF(
static int pfmFileCwdtree,( out, depth ), FILE *out __j__ int depth )
{
	DIR	*dir;

	if ( (dir = opendir(".")) )
	{
		struct dirent	*ent;
		int n = 0;

		while ( (ent = readdir(dir)) )
			if (strcmp(ent->d_name,".") && strcmp(ent->d_name,".."))
			{
				struct stat sbuf;
				x_filedat   info;
	
				pfmFileStat( ent->d_name, &sbuf, &info.st );
	
				if ( _s_isdir(info) )
				{	__pfmFilePrinttreenode( out, ent->d_name,
					                        _s_islnk(info), depth );
					++n;

					if ( !_s_islnk(info) )
					{	/*
						static ino_t ind[64];
						static int   pos = 0;
	
						if (pos < sizeof(ind)/sizeof(*ind))
						{	int	i;
							for (i = 0; i < pos; ++i)
								if (ent->d_ino == ind[i]) break;
	
							if (i == pos)
							{	ind[pos++] = ent->d_ino;
						*/		if ( !chdir(ent->d_name) )
								{	n += pfmFileCwdtree( out, depth+1 );
									chdir("..");
								}
						/*		pos--;
							}
						}
						*/
					}
				}
			}
		closedir(dir);
		return n;
	}
	return (-1);
}

/*
 *	Traverse and print directory-tree
 */
__DF(
static int pfmFileTree,( name, info ),
const char	*name __j__
x_filedat	*info )
{
	if (_s_isdir(*info) && !_s_islnk(*info))
	{	if ( !chdir(name) )
		{	FILE *fp;

			pfmPreExecshell(0);

			if ( (fp = popen(pfmConfPagerPrg(),"w")) )
			{	const char *cp = pfmCwdCWD();
				fputs( cp, fp );
				fprintf( fp, cp[1] ? "/%s\n":"%s\n", name );
				pfmFileCwdtree( fp, 1 );
				pclose(fp);
			}
			pfmPostExecshell(EXECSH_WAIT|EXECSH_NORRD);
			clearok( stdscr, TRUE );
		}
		chdir( pfmCwdCWD() );
	}
	return 0;
}

/*
 *	Print directory-tree of a directory (at the cursor position).
 */
__DF(
int pfmMnuFileTree,(m), x_menubar *m )
{
	static const char pre[] = {pfm_fwildc_start, pfm_Fname_fwildc, 0};
	char	dest[CWD_MAX_LNG];

	if ( !pfmTermGetstrl(pfm_C_dir_pathname,sizeof(dest),dest,pre,(char*)0) )
	{	struct stat	sbuf;
		x_filedat	info;
		pfmExpandargs( dest, sizeof(dest) );
		pfmFileStat( dest, &sbuf, &info.st );
		pfmFileTree( dest, &info );
	}
	(void)m;
	return (-1);
}


/*
 *	(Recursively) copy a directory hierarchy
 */
__DC(static int pfmFileCopydirexe,(const char*,const char*,const char*));
__DC(static int pfmFileBuiltincpiterate,(const char*,const char*,const char*));

/*
 *	The commands entry point: expand source and destination,
 *	check source.
 */
__DF(
int pfmFileCopydir,( quelle, ziel ),
const char *quelle __j__
const char *ziel )
{
	struct stat	sbuf;
	/*
	 *	Expand "quelle" with global (default) variables
	 */
	int  lq = pfmExpandsize(quelle) + 2;
	int  lz = pfmExpandsize(ziel) + pfmExpandsize(pfmFileBasename(quelle)) + 2;
	char *qp = (char*) xalloca(lq+lz+2);
	pfmExpandargs( strcpy(qp,quelle), lq );
	/*
	 *	source must be an existing directory
	 */
	if (stat(qp,&sbuf) || !S_ISDIR(sbuf.st_mode))
		pfmTermErrormessage( strerror(ENOTDIR), qp, (char*)0 );
	else
	{	const char *cpdir = pfmConfCpdirCmd();
		/*
		 *	Expand "ziel" with global (default) variables
		 */
		char *zp = strchr(qp,0) + 1;
		pfmExpandargs( strcpy(zp,ziel), lz );
		/*
		 *	Finally expand trailing "/[*]" in "ziel"
		 */
		pfmSetCopyWildname( qp, zp, lz );
		#if !HAS_NATIVE_CPDIR_CMD
		if ( pfmStrmatch(cpdir,BUILTIN_CMD_PFX) )
			return pfmFileBuiltincpdir( cpdir, qp, zp );
		else
		#endif
			return pfmFileCopydirexe( cpdir, qp, zp );
	}
	return (-1);
}

/*
 *	Last stage of pfmFileCopydir, check destination,
 *	expand and execute command.
 */
__DF(
static int pfmFileCopydirexe,( cpdir, quelle, ziel ),
const char *cpdir __j__
const char *quelle __j__
const char *ziel )
{
	struct stat	sbuf;
	/*
	 *	destination must not exist or else be a directory
	 */
	if (!stat(ziel,&sbuf) && !S_ISDIR(sbuf.st_mode))
		pfmTermErrormessage( strerror(ENOTDIR), ziel, (char*)0 );
	else
	{	/*
		 *	now we can expand %D, %A to source, destination respective
		 *	(%F,%N,%X are not valid wildcards in this command pattern).
		 */
		int  l = pfmVExpandsize( cpdir, "-?", quelle, ziel );
		char *c = (char*) xalloca(l+=20);
		pfmVExpandargs( strcpy(c,cpdir), l, "-?", quelle, ziel );
		/*
		 *	the user may want to change parameters to "cpdir"
		 */
		if ( !pfmTermGetstrl("",l,c,c,(char*)0) )
			return pfmExecshell( c, EXECSH_WAIT );
	}
	return (-1);
}


#if !HAS_NATIVE_CPDIR_CMD
/*
 *	Wrapper for call(s) to recursive cp (cp -R ...)
 */
__DF(
static int pfmFileBuiltincpdir,( cpdir, quelle, ziel ),
const char *cpdir __j__
const char *quelle __j__
const char *ziel )
{
	static const char cpr[] = "cp -R";
	struct stat	sbuf;
	char cmd[40];
	/*
	 *	Check, if we can get along with a single cp invocation.
	 */
	bool single = 0;
	if ( stat(ziel,&sbuf) )
		single = 1;
	else
	{	const char *b = pfmFileBasename(ziel);
		if ( !strcmp(pfmFileBasename(quelle),b) )
		{	/*
			 *	Alright, we can simply cp to destination's parent directory.
			 */
			int n = b - ziel;
			switch (n)
			{	case 0:
					ziel = ".";
					break;
				case 1:
					ziel = "/";
					break;
				default:
				{	char *p = (char*) xalloca(n);
					strncpy( p, ziel, n )[n-1] = 0;
					ziel = p;
				}
			}
			single = 1;
		}
	}
	while (*cpdir && !isspace(*cpdir)) ++cpdir; /* skip argv[0] */

	strncat( strcpy(cmd,cpr), cpdir, sizeof(cmd)-sizeof(cpr) );
	cmd[sizeof(cmd)-1] = 0;

	if ( single )
		return pfmFileCopydirexe( cmd, quelle, ziel );
	else
		return pfmFileBuiltincpiterate( cmd, quelle, ziel );
}

/*
 *	For cp'ing (cp -R ...) each entry in source to destination.
 */
__DF(
static int pfmFileBuiltincpiterate,( cppat, quelle, ziel ),
const char *cppat __j__
const char *quelle __j__
const char *ziel )
{
	static const char
	wf[] =
	{	pfm_fwildc_start, pfm_Fname_fwildc, 0
	},
	wq[] =
	{	pfm_fwildc_start, pfm_CWD_fwildc, '/',
		pfm_fwildc_start, pfm_Fname_fwildc, 0
	};
	char *cmd;
	int  lng;
	/*
	 *	Prepare command-pattern for repeated expansion.
	 */
	lng = pfmVExpandsize( cppat, "-?", quelle, ziel ) + 4;
	cmd = (char*) xalloca(lng);
	pfmVExpandargs( strcpy(cmd,cppat), lng, "-?", wq, ziel );
	pfmVExpandargs( cmd, lng, wf, quelle, "" );
	/*
	 *	the user may want to change parameters to "cp -R"
	 */
	if ( !pfmTermGetstrl("",lng,cmd,cmd,(char*)0) )
	{	DIR	*dir;

		if ( !(dir = opendir(quelle)) )
			pfmTermErrormessage( strerror(errno), quelle, (char*)0 );
		else
		{	struct dirent	*ent;

			cppat = cmd;
			lng = pfmVExpandsize(cppat,wf,"-?","-?") + MAXNAMLEN + 1;
			cmd = (char*) xalloca(lng);

			pfmPreExecshell(0);

			while ( (ent = readdir(dir)) )
				if (strcmp(ent->d_name,".") && strcmp(ent->d_name,".."))
				{	strcpy( cmd, cppat );
					pfmVExpandargs( cmd, lng, ent->d_name, "-?", "-?" );
					if ( pfmSystem(cmd) ) break;
				}
			pfmPostExecshell(EXECSH_WAIT);
			closedir(dir);
		}
	}
	return (-1);
}
#endif /* !HAS_NATIVE_CPDIR_CMD */

/***** eof *********************************************************/
