/***********************************************************************
* Copyright (c) 1995-2003 by peter fuerst
*
*	This program is free software; you can redistribute it and/or modify
*	it - as explained in the accompaning file COPYING - under the terms
*	of the GNU General Public License as published by the Free Software
*	Foundation; either version 2, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	Author: Peter Frst, post@pfrst.de
*
*	@(#)files.c	10/25/95
*
***********************************************//*$preferred tabstop=4*/

#include <stdio.h>
#include <stddef.h>
#include <dirent.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/types.h>
#ifdef __linux__
 #include <utime.h>
#endif
#include "spawn.h"

#include "pfm.h"
#include "pfmstat.h"

/***** define & typedef ********************************************/

#define _s_islnk(data)	S_ISLNK((data).st.lmode)
#define _s_isdir(data)	S_ISDIR((data).st.mode)

#ifndef EISDIR
 #define EISDIR EACCESS
#endif


/***** forward declaration *****************************************/

__DC( static int _pfmFileShowinfo,(x_filedat*) );

/***** global ******************************************************/

/***** local *******************************************************/

#ifdef DEUTSCH
 static const char invalid_att_txt[] = "Ungueltiges attribut: ";
 static const char new_mode_txt[] = "Neues attribut: ";
 static const char modepat_txt[] = "Attributmuster: ";
 static const char overwrite_txt[] = "Ueberschreiben [J/N] ?";
 static const char hit_any_other_key[] = " weiter mit anderer Taste...      ";
#else
 static const char invalid_att_txt[] = "Invalid attribute: ";
 static const char new_mode_txt[] = "New mode: ";
 static const char modepat_txt[] = "mode pattern: ";
 static const char overwrite_txt[] = "Overwrite [Y/N] ?";
 static const char hit_any_other_key[] = " any other key to continue...     ";
#endif
static const char _ccX_empty[1] = "";
static const char *fsize_t_fmt;


/********************************************************************
*	utilities
********************************************************************/

__DF(
const char *pfmFileBasename,( name ), const char* name )
{
	if (name)
	{	char *p = strrchr(name,'/');
		if (p) name = p+1;
	}
	return name;
}
/*
__DF(
unsigned pfmFileDirnamelng,( name ), const char* name )
{
	if (name)
	{	char *p = strrchr(name,'/');
		if (p) return (p != name ? p-name:1);
	}
	return 0;
}
*/

/********************************************************************
*	retreive/handle file modes
********************************************************************/

static const int	_mode_msk[] =
{	0, S_IREAD, S_IWRITE, S_IEXEC,
	S_IRGRP, S_IWGRP, S_IXGRP,
	S_IROTH, S_IWOTH, S_IXOTH
};
static const char	_mode_rwx[] = "-rwxrwxrwx";

__DF(
int pfmFileEncodemode,( mstr, mode, oldmode ),
const char	mstr[12] __j__
int	*mode __j__
int	oldmode )
{
	char	buf[12];
	int 	i;
	unsigned neu = oldmode;

	for (i = 0; i < 10; i++) buf[i] = '?';
	buf[11] = buf[10] = 0;

	for (i = 0; i < 11; i++)
	{	if ( !mstr[i] ) break;
		if ('*' == mstr[i])
		{	if ( mstr[i+1] ) return (-1);
			break;
		}
		buf[i] = mstr[i];
	}
	switch (*buf)
	{	case 'd': i = S_IFDIR; break;
		case 'c': i = S_IFCHR; break;
		case 'b': i = S_IFBLK; break;
		case '-': i = S_IFREG; break;
		case '?': i = 0; break;
		default:  return (-1);
	}
	if (i)
	{	neu &= ~S_IFMT;
		neu |= i;
	}
	if ('-' == buf[3])
		neu &= ~S_ISUID;
	else if ('s' == buf[3])
	{	neu |= S_ISUID;
		buf[3] = 'x';
	}
	if ('-' == buf[6])
		neu &= ~S_ISGID;
	else if ('s' == buf[6])
	{	neu |= S_ISGID;
		buf[6] = 'x';
	}
	switch (buf[10])
	{	case 0: case ' ': case '-':
			neu &= ~S_ISVTX;
			break;
		case 't':
			neu |= S_ISVTX;
		case '?':
			break;
		default:
			return (-1);
	}
	for (i = 1; i < sizeof(_mode_msk)/sizeof(_mode_msk[0]); i++)
		if (buf[i] == _mode_rwx[i])
			neu |= _mode_msk[i];
		else if ('-' == buf[i])
			neu &= ~_mode_msk[i];
		else if (buf[i] != '?')
			return (-1);

	*mode = neu;
	return 0;
}

__DF(
char *pfmFileDecodemode,( mode, buffer ),
int 	mode __j__
char	buffer[12] )
{
	register short	j;

	switch (mode & S_IFMT)
	{	case S_IFDIR: *buffer = 'd'; break;
		case S_IFCHR: *buffer = 'c'; break;
		case S_IFBLK: *buffer = 'b'; break; /*
		case S_IFREG: */
		default:      *buffer = '-';
	}
	for (j = 1; j < sizeof(_mode_msk)/sizeof(_mode_msk[0]); j++)
		buffer[j] = mode & _mode_msk[j] ? _mode_rwx[j]:'-';

	if (mode & S_ISUID) buffer[3] = 's';
	if (mode & S_ISGID) buffer[6] = 's';
	buffer[10] = mode & S_ISVTX ? 't':0;
	buffer[11] = 0;
	return buffer;
}


/*
 *	Retrieve extended 'stat' information
 */
__DF(
static int _pfmFileStat,( fname, sbuf, lbuf, pbuf ),
	const char		*fname __j__
	struct stat 	*sbuf __j__
	struct stat 	*lbuf __j__
	struct _pfmstat	*pbuf )
{
	unsigned lmode = 0;
	struct stat	tmp;
	int	r;

	if (!sbuf) sbuf = &tmp;

	if (lbuf) memset( lbuf, 0, sizeof(*lbuf) );
	memset( sbuf, 0, sizeof(*sbuf) );
	r = stat( fname, sbuf );

	#ifdef HAS_SYMLINK_STAT
		if (pbuf || lbuf)
		{	struct stat	ltmp;
			if (!lbuf) lbuf = &ltmp;
			if ( !lstat(fname,lbuf) )
			{	lmode = lbuf->st_mode;
				/*
				 * take a symlink's own state, if the file
				 * pointed to, could not be 'stat'ed...
				 */
				if (r && S_ISLNK(lbuf->st_mode))
				{	*sbuf = *lbuf;
					r = 0;
				}
			}
		}
	#endif
	if (pbuf)
	{	memset( pbuf, 0, sizeof(*pbuf) );
		pbuf->lmode = lmode;
		if (!r)
		{	pbuf->ino = sbuf->st_ino;
			pbuf->mode = sbuf->st_mode;
			pbuf->size = sbuf->st_size;
			pbuf->atime = sbuf->st_atime;
			pbuf->mtime = sbuf->st_mtime;
			pbuf->ctime = sbuf->st_ctime;
		}
	}
	return r;
}

__DF(
int pfmFileStat,( fname, sbuf, pbuf ),
	const char		*fname __j__
	struct stat 	*sbuf __j__
	struct _pfmstat	*pbuf )
{
#if 1
	return _pfmFileStat( fname, sbuf, (struct stat*)0, pbuf );
#else
	struct stat	tmp;
	int	r;

	if (!sbuf) sbuf = &tmp;

	memset( sbuf, 0, sizeof(*sbuf) );
	r = stat( fname, sbuf );

	if (pbuf)
	{	memset( pbuf, 0, sizeof(*pbuf) );
		#ifdef HAS_SYMLINK_STAT
		{	struct stat	tmp;
			int	l = lstat( fname, &tmp );
			if (!l)
			{	pbuf->lmode = tmp.st_mode;
				/*
				 * take a symlink's own state, if the file
				 * pointed to, could not be 'stat'ed...
				 */
				if (r && S_ISLNK(tmp.st_mode))
				{	*sbuf = tmp;
					r = l;
				}
			}
		}
		#endif
		if (!r)
		{	pbuf->ino = sbuf->st_ino;
			pbuf->mode = sbuf->st_mode;
			pbuf->size = sbuf->st_size;
			pbuf->atime = sbuf->st_atime;
			pbuf->mtime = sbuf->st_mtime;
			pbuf->ctime = sbuf->st_ctime;
		}
	}
	return r;
#endif
}


/*
 *	Query an file-attribute-pattern (for attribute-matching)
 */
__DF(
const char *pfmFileGetModepattern,( prompt, newpat ),
const char *prompt __j__
const char *newpat )
{
	static char lastpat[16] = "?rwxrwxrwx";

	if ( !prompt ) prompt = modepat_txt;
	if ( !newpat ) newpat = lastpat;

	while (1)
	{	char buf[16];
		int  mode;

		if ( pfmTermGetstrl(prompt,12,buf,newpat,(char*)0) )
			break;
		if ( !pfmFileEncodemode(buf,&mode,0) )
			return strcpy(lastpat,buf);

		pfmTermErrormessage( invalid_att_txt, buf, (char*)0 );
	}
	return (char*)0;
}

/*
 *	Query for new file attribute
 */
__DF(
int pfmFileGetnewmode,( oldmode, newmode ), int oldmode __j__ int *newmode )
{
	char	obuf[16];

	for (pfmFileDecodemode(oldmode,obuf); ; )
	{	char nbuf[16];
		int  neu;

		strcpy( nbuf, obuf );

		if ( pfmTermGetstrl(new_mode_txt,11,nbuf+1,obuf+1,(char*)0) ) break;

		if ( !pfmFileEncodemode(nbuf,&neu,oldmode) )
		{	*newmode = neu;
			return 0;
		}
		pfmTermErrormessage( invalid_att_txt, nbuf, (char*)0 );
	}
	return (-1);
}


/*
 *	Change attributes of the file at the cursor position
 */
__DF(
int pfmMnuFileModifyatt,(m), x_menubar *m )
{
	x_filedat  *p = pfmCwdCurrentfile();

	if (p)
	{	struct stat	sbuf;
		if ( stat(p->nm,&sbuf) )
			pfmTermErrormessage( strerror(errno), p->nm, (char*)0 );
		else
		{	int	mode;
			if ( !pfmFileGetnewmode(sbuf.st_mode, &mode) )
				if ( chmod(p->nm,mode & ~S_IFMT) )
					pfmTermErrormessage( strerror(errno), p->nm, (char*)0 );
				else
					_pfmFileShowinfo(p);
		}
	}
	(void)m;
	return 0;
}


/********************************************************************
*	display information about file
********************************************************************/
/*
 *	Expand non-printable characters
 */
__DF(
int pfmIsprintExpandString,( raw, buf, limit ),
const char	*raw __j__
char		*buf __j__
const char	*limit )
{
	int	i;

	for (i = 0; buf < limit; i++)
	{	unsigned  c = (unsigned char)raw[i];
		if (!c) break;
		if (isascii(c) && isprint(c))
			*buf++ = c;
		else
		{	sprintf( buf, "\\%o", c );
			while (*buf) buf++;
		}
	}
	*buf = 0;
	return i;
}

/*
 *	Extended file info
 */
#define WW 38
#define WH 16

__DF(
static void pfmbuildinfolines,( buf, sbuf ),
char	buf[][WW] __j__
const struct stat	*sbuf )
{
	char *p;
	long i;

	if ( !fsize_t_fmt )
		fsize_t_fmt = sizeof(fsize_t) > sizeof(long) ? "%lld":"%ld";

	sprintf( buf[0], "%-11s", "Type:" );
	if ( S_ISREG(sbuf->st_mode) ) strcat( buf[0], "File " );
	if ( S_ISDIR(sbuf->st_mode) ) strcat( buf[0], "Directory " );
	if ( S_ISCHR(sbuf->st_mode) ) strcat( buf[0], "Character-special " );
	if ( S_ISBLK(sbuf->st_mode) ) strcat( buf[0], "Block-special " );
	#if HAS_SYMLINK
		if ( S_ISLNK(sbuf->st_mode) ) strcat( buf[0], "Symbolic link" );
	#endif
	sprintf( buf[1], "%-11s%u", "Inode:", (unsigned)sbuf->st_ino );
	sprintf( buf[2], "%-11s%u", "Links:", (unsigned)sbuf->st_nlink );

	if ( !(i = BLKSIZE(*sbuf)) ) i = 512;
	i = ((i-1)+sbuf->st_size) / i;
	sprintf( p=buf[3], "%-11s", "Size:" );
	sprintf( p = strchr(p,0), fsize_t_fmt, sbuf->st_size );
	sprintf( p = strchr(p,0), "  (%lu blk%s)", i, i > 1 ? "s":"" );

	sprintf( buf[4], "%-11s%u  %s",
		"User Id:", (unsigned)sbuf->st_uid, pfmGetusername(sbuf->st_uid) );
	sprintf( buf[5], "%-11s%u  %s",
		"Group Id:", (unsigned)sbuf->st_gid, pfmGetgroupname(sbuf->st_gid) );

	sprintf( buf[6], "%-11s%u %u (rdev: %u %u)",
		"Device:", major(sbuf->st_dev), minor(sbuf->st_dev),
			major(sbuf->st_rdev), minor(sbuf->st_rdev) );

	pfmFileDecodemode( sbuf->st_mode, buf[8] );
	sprintf( buf[7], "%-11s%c %.3s %.3s %.3s %.1s (%07o)",
		"Attribute:", *buf[8], buf[8]+1, buf[8]+4, buf[8]+7, buf[8]+10,
			(unsigned short)sbuf->st_mode );

	sprintf( buf[ 8], "%-11s%.24s", "Access:", ctime(&sbuf->st_atime) );
	sprintf( buf[ 9], "%-11s%.24s", "Modify:", ctime(&sbuf->st_mtime) );
	sprintf( buf[10], "%-11s%.24s", "Change:", ctime(&sbuf->st_ctime) );
}

__DF(
static int pfmbuildinfonameline,( buf, ww, name, pos ),
char	buf[] __j__
int	ww __j__
const char	*name __j__
int	pos )
{
	char	*bp = buf;
	int		i;

	name += pos;
	*bp = pos > 0 ? '<':'`';
	bp++;
	i = pfmIsprintExpandString( name, bp, bp+(ww-(1+1)-(1+3)) );
	bp = strchr( bp, 0 );
	*bp = '\'';
	if (name[i]) *bp = '>';
	while (++bp < buf+ww) *bp = ' ';
	buf[ww-1] = 0;
	return i;
}

__DF(
static void pfmbuildinfoenterline,( buf, ww, slnk, shft ),
char	buf[] __j__
int	ww __j__
const char	*slnk __j__
int	shft )
{
	int i = 1;
	ww -= 2*2;
	*buf = ' ';
	#if HAS_SYMLINK
		if ( slnk )
		{	buf[1] = 'S';
			buf[2] = ',';
			i += 2;
		}
	#endif
	if (shft)
	{	strcpy( buf+i, "<->," );
		i += 4;
	}
	strncpy( buf+i, i>1 ? hit_any_other_key:pfm_C_hit_any_key, ww+1-i );
	buf[ww-1] = 0;
}

__DF(
static int _pfmFileShowinfo,( inf ), x_filedat *inf )
{
	WINDOW *wip = newwin( WH, WW, (wheight(stdscr)-WH)/2,
							(wwidth(stdscr)-WW)/2 );
	if (wip)
	{
		struct stat	sbuf;
		char	lines[11][WW];
		char	*name = inf->nm;
		char	*symlnkbuf = 0;
		int 	i, pos, teil;
		#if HAS_SYMLINK
			struct stat	lbuf;

			_pfmFileStat( inf->nm, &sbuf, &lbuf, &inf->st );
			if ( _s_islnk(*inf) )
			{	int j = lbuf.st_size & 0x1fff;
				symlnkbuf = (char*) alloca(j+2);
				if ( symlnkbuf )
				{	j = readlink( inf->nm,symlnkbuf, j );
					if (j > 0)
						name = symlnkbuf;
					else
						j = 0;
					symlnkbuf[j] = 0;
				}
			}
		#else
			_pfmFileStat( inf->nm, &sbuf, (struct stat*)0, &inf->st );
		#endif

		if ( pfmFileMultilinked(sbuf.st_mode,sbuf.st_nlink) )
			inf->flag |= X_FILEDAT_MLNK_FLAG;
		else
			inf->flag &= ~X_FILEDAT_MLNK_FLAG;

		pfmbuildinfolines( lines, &sbuf );

		wstandout(wip);
		wmove( wip, 0, 0 );
		for (i = 0; i < WH*WW; i++) /* this let's standout take effect... */
			waddch( wip, ' ' );

		for (i = 2; i < 13; i++)
			mvwaddstr( wip, i, 1, lines[i-2] );

		for (teil = pos = 0; ; )
		{	char	*nxp;

			i = pfmbuildinfonameline( lines[0], WW, name, pos );
			nxp = name+pos+i;
			teil |= *nxp;
			pfmbuildinfoenterline( lines[1], WW, symlnkbuf, teil );

			#if HAS_SYMLINK
				i = _s_islnk(*inf) && name != symlnkbuf ? '~':' ';
				mvwaddch( wip, 2, 11, i );
			#endif
			mvwaddstr( wip, 1, 1, lines[0] );
			wstandend(wip);
			mvwaddstr( wip, WH-2, (WW-strlen(lines[1]))/2, lines[1] );
			wstandout(wip);
			touchwin(wip);
			wrefresh(wip);

			pfmTermClrcon();
			while ((i = pfmTermGetch()) < 1)
				;
			switch (i)
			{	case KEY_RIGHT: case 'l': case '>':
					if (teil)
					{	if (*nxp && pos+8 < strlen(name)) pos += 8;
						continue;
					}
				case KEY_LEFT: case 'h': case '<':
					if (teil)
					{	if ((pos-=8) < 0) pos = 0;
						continue;
					}
				#if HAS_SYMLINK
					break;
				case 's': case 'S': /* toggle SymLink/Target */
					if (symlnkbuf)
					{	name = (name != symlnkbuf ? symlnkbuf : inf->nm);
						pos = 0;
						teil = 0;
						continue;
					}
				#endif
			}
			break;	/* all done !	*/
		}
		delwin(wip);
		pfmCwdWinDrawfileswin(TRUE);
	}
	return 0;
}
#undef WW
#undef WH


__DF(
int pfmMnuFileInfo,(m), x_menubar *m )
{
	x_filedat	*info = pfmCwdCurrentfile();

	if (info) _pfmFileShowinfo(info);

	(void)m;
	return 0;
}


/********************************************************************
*	support for copy, link, remove, edit
********************************************************************/
/*
 *	ask before removing,overwriting,... a file.
 */
__DF(
static int _pfmFileConfirmaction,( question, fname, yes ),
const char *question __j__
const char *fname __j__
bool yes )
{
	char *buf = (char*) alloca(strlen(question)+24);
	if (!buf)
	{	pfmTermErrormessage( pfm_C_out_of_stack, "alloca()", (char*)0 );
		return 0;
	}
	strcpy( buf, pfmStrtail(fname,20) );
	strcat( strcat(buf,": "), question );
	return pfmTermYesNo( buf, yes );
}

/*
 *	If destination path ends with '/' or a wildcard,
 *	extract the filename from the source path.
 */
__DF(
char  *pfmSetCopyWildname,( von, ziel, zlaenge ),
const char	*von __j__
char	ziel[/*zlaenge*/] __j__
int 	zlaenge )
{
	char  *z = strrchr( ziel, '/' );

	if (z++ && (!*z || !strcmp(z,"*")))
	{
		const char	*q = pfmFileBasename(von);
		*z = 0;		/* purge wildcard, if any	*/

		while (z < ziel+zlaenge)
			if ( !(*z++ = *q++) ) break;
		ziel[zlaenge-1] = 0;
	}
	return ziel;
}

/*
 *	Store/retreive destination of last copy/link/edit
 */
__DF(
const char *pfmFileSGetLastdest,( newdest ), const char *newdest )
{
	static const char null = 0;
	static char	*lastdest = (char*)&null;

	if ( newdest )
	{	lastdest = pfmStrdup( newdest, lastdest != &null ? lastdest:(char*)0 );
		if ( !lastdest ) lastdest = (char*)&null;
	}
	return lastdest;
}


/********************************************************************
*	copy operations
********************************************************************/
/*
 *	Copy file
 *
 *	returns 0, if file was copied, != 0, if not, for whatever reason.
 */
__DF(
int pfmFileCopyfiles,( quelle, ziel ),
const char *quelle __j__
const char *ziel )
{
	int	fq, err = 1;

	errno = 0;

	if ((fq = open(quelle,O_RDONLY|O_BINARY)) < 0)
		pfmTermErrormessage( strerror(errno), quelle, (char*)0 );
	else
	{	struct stat	sbuf;
		ino_t 	inode;
		int 	attr;
		dev_t 	dev;
		time_t	am[2]; /* struct utimbuf am[1]; */

		fstat( fq, &sbuf );
		am[0] = sbuf.st_atime;
		am[1] = sbuf.st_mtime;
		inode = sbuf.st_ino;
		attr  = sbuf.st_mode;
		dev = sbuf.st_dev;

		if ( S_ISDIR(sbuf.st_mode) ) /* open() may not have failed */
		{	pfmTermErrormessage( strerror(errno=EISDIR), quelle, (char*)0 );
			goto closesource;
		}
		else
		{	unsigned l = strlen(ziel) + strlen(pfmFileBasename(quelle)) + 2;
			char    *z = (char*) alloca(l);

			if ( !z )
			{	pfmTermErrormessage( pfm_C_out_of_stack, ziel, (char*)0 );
				goto closesource;
			}
			ziel = pfmSetCopyWildname( quelle, strcpy(z,ziel), l );
		}
		/*
		 *	Do not attempt to copy a file onto itself !
		 */
		if (stat(ziel,&sbuf) || inode != sbuf.st_ino || dev != sbuf.st_dev)
			if ( S_ISDIR(sbuf.st_mode) )
				pfmTermErrormessage( strerror(errno=EISDIR), ziel, (char*)0 );
			else
			{	int	fz;

				errno = 0;
				fz = creat( ziel, S_IWRITE|S_IREAD );

				if (fz < 0)
					pfmTermErrormessage( strerror(errno), ziel, (char*)0 );
				else
				{	char	buf[512];
					int		n;
					err = 0;
					while ((n = read(fq,buf,sizeof(buf))) > 0)
						if (write(fz,buf,n) != n)
						{	pfmTermErrormessage( strerror(errno), ziel, (char*)0 );
							err = 1;
							break;
						}
					if (n < 0) /* read-error */
					{	pfmTermErrormessage( strerror(errno), quelle, (char*)0 );
						err = 1;
					}
					close(fz);
					chmod( ziel, attr );
					utime( ziel, (void*)am );
				}
			}
		closesource:
		close(fq);
	}
	return err;
}


/*
 *	Copy the file at the cursor position.
 */
__DF(
int pfmMnuFileCopyfile,(m), x_menubar *m )
{
	x_filedat  *p = pfmCwdCurrentfile();
	if (p)
#if 0
		if ( _s_isdir(*p) )
			pfmTermErrormessage( strerror(EISDIR), p->nm, (char*)0 );
		else
		{	char	dest[CWD_MAX_LNG];
			if ( !pfmTermGetstrl(pfm_C_destination,sizeof(dest),dest,
								pfmFileSGetLastdest((char*)0),(char*)0) )
			{	pfmVExpandargs(dest, sizeof(dest), p->nm, (char*)0, (char*)0);
				if ( !pfmFileCopyfiles(p->nm,dest) )
				{	pfmFileSGetLastdest(dest);
					pfmCwdRereadChanged();
				}
			}
		}
#else
	{	char	dest[CWD_MAX_LNG];
		if ( !pfmTermGetstrl(pfm_C_destination,sizeof(dest),dest,
			                 pfmFileSGetLastdest((char*)0),(char*)0) )
			if ( _s_isdir(*p) )
			{	if ( !pfmFileCopydir(p->nm,dest) )
					pfmFileSGetLastdest(dest);
			}
			else
			{	pfmVExpandargs(dest, sizeof(dest), p->nm, (char*)0, (char*)0);
				if ( !pfmFileCopyfiles(p->nm,dest) )
				{	pfmFileSGetLastdest(dest);
					pfmCwdRereadChanged();
				}
			}
	}
#endif
	(void)m;
	return 0;
}


#if HAS_SYMLINK
/*
 *	Copy a symlink as symlink.
 */
__DF(
static int pfmFileCopysymlink, ( alt, neu ),
const char  *alt __j__
const char  *neu )
{
	struct stat	sbuf;

	errno = 0;

	/* S_ISLNK(sbuf.st_mode) is not redundant ! */
	if (!lstat(alt,&sbuf) && S_ISLNK(sbuf.st_mode))
	{	int  n = sbuf.st_size & 0x3fff;
		char *pfad = (char*) alloca(n+2);
		if (pfad)
		{	n = readlink( alt, pfad, n );
			if (n > 0)
			{	pfad[n] = 0;
				return symlink(pfad,neu);
			}
		}
		return -1;
	}
	return EINVAL;
}
#endif

/*
 *	Copy a special file as special file.
 */
__DF(
static int pfmFileCopynode, ( alt, neu ),
const char  *alt __j__
const char  *neu )
{
	struct stat	sbuf;

	errno = 0;

	if ( stat(alt,&sbuf) ) return -1;

	if (S_ISREG(sbuf.st_mode) || S_ISDIR(sbuf.st_mode)) return EINVAL;

	return mknod( neu, sbuf.st_mode, sbuf.st_dev );
}


/********************************************************************
*	Hard- and Soft-Link operations
********************************************************************/
/*
 *	Link operations on named file
 */
__DF(
static int  _pfmFileRelinkfile,( alt, neu, relink ),
const char *alt __j__
const char *neu __j__
__DC(int  (*relink),(const char*,const char*)) )
{
	unsigned   l = strlen(alt) + strlen(neu) + 1;
	char *ziel = (char*) alloca(l);

	ziel = ziel ? pfmSetCopyWildname( alt, strcpy(ziel,neu), l ):(char*)neu;

	if ( strcmp(alt,ziel) )
	{	if (access(ziel,F_OK) == 0) /* ANSI-rename would overwrite ! */
			if ( !_pfmFileConfirmaction(overwrite_txt,ziel,FALSE) )
				return (-1);
		errno = 0;

		if ( (*relink)(alt,ziel) )
		{	__DC(extern int rename,(const char*,const char*));
			/*
			 *	To rename across filesystems, we need to move the file.
			 *	Logically here is not an especially good place to handle
			 *	physical moving, but there is no better place outside.
			 *	(Would need a flag-parameter and errno anyway to suppress
			 *	the error message, would need to hand over 'errno', ...)
			 */
			if (EXDEV == errno && rename == relink)
			{	int r;
			#if HAS_SYMLINK
				/*	...but avoid dereferencing a symlink, if possible.
				 */
				r = pfmFileCopysymlink( alt, ziel );
				if (EINVAL == r)
			#endif
					/*	...and avoid dumping a special file, if possible.
					 */
					r = pfmFileCopynode( alt, ziel );

				if (r && r != EINVAL)
				{	const char *msg = errno
						? strerror(errno):pfm_C_out_of_stack;
					pfmTermErrormessage(msg, alt, ziel, (char*)0);
					return 1;
				}
				/*	use 'neu' instead of 'ziel' because of wildcard
				 *	expansion inside of pfmFileCopyfiles() !
				 */
				if ( pfmFileCopyfiles(alt,neu) )
					return 1; /* error message already shown */

				if ( unlink(alt) )
				{	pfmTermErrormessage(strerror(errno), alt, (char*)0);
					return 1;
				}
			}
			else
			{	pfmTermErrormessage(strerror(errno), ziel,alt, (char*)0);
				return 1;
			}
		}
	}
	return 0;
}

/*
 *	Rename a named file
 */
__DF(
int  pfmFileRenamefile,( alt, neu ), const char *alt __j__ const char *neu )
{
	__DC(extern int rename,(const char*,const char*)); /* <unistd.h> */
	return _pfmFileRelinkfile( alt, neu, rename );
}

#if HAS_SYMLINK
/*
 *	Create symbolic link to a named file
 */
__DF(
int  pfmFileSymlinkfile,( alt, neu ), const char *alt __j__ const char *neu )
{
	__DC(extern int symlink,(const char*,const char*)); /* <unistd.h> */
	return _pfmFileRelinkfile( alt, neu, symlink );
}
#endif


/*
 *	Link operations on the file at the cursor position
 */
__DF(
static int _pfmMnuFileRelinkfile,( alt, relink ),
const char *alt __j__
__DC(int  (*relink),(const char*,const char*)) )
{
	x_filedat  *p = pfmCwdCurrentfile();

	if (p)
	{	char	dest[CWD_MAX_LNG];
		if ( !pfmTermGetstrl(pfm_C_destination,sizeof(dest),dest,
		                     pfmFileSGetLastdest((char*)0),(char*)0) )
		{	pfmVExpandargs(dest, sizeof(dest), p->nm, (char*)0, (char*)0);
			if ( !_pfmFileRelinkfile(alt ? alt:p->nm, dest, relink) )
			{	pfmFileSGetLastdest(dest);
				pfmCwdRereadChanged();
			}
		}
	}
	return 0;
}

/*
 *	Rename the file at the cursor position
 */
__DF(
int pfmMnuFileRenamefile,(m), x_menubar *m )
{
	__DC(extern int rename,(const char*,const char*)); /* <unistd.h> */
	(void)m;
	return _pfmMnuFileRelinkfile( (char*)0, rename );
}

/*
 *	Create new (hard-)link to the file at the cursor position
 */
__DF(
int pfmMnuFileLinkfile,(m), x_menubar *m )
{	/*
	 *	we only hardlink here, since symlinking just '%F' without
	 *	path may be not too useful in most cases.
	 */
	__DC(extern int link,(const char*,const char*)); /* <unistd.h> */
	(void)m;
	return _pfmMnuFileRelinkfile( (char*)0, link );
}

#if HAS_SYMLINK
/*
 *	Create symbolic link (to the file at the cursor position).
 */
__DF(
int pfmMnuFileSymlink,(m), x_menubar *m )
{
	__DC(extern int symlink,(const char*,const char*)); /* <unistd.h> */
	static const char	pre[] =
	{	pfm_fwildc_start, pfm_CWD_fwildc, '/',
		pfm_fwildc_start, pfm_Fname_fwildc, 0
	};
	char	alt[CWD_MAX_LNG];
	/*
	 *	for symlinks it is crucial to allow for a complete sourcepath,
	 *	not just silently using '%F' !
	 */
	if ( !pfmTermGetstrl(pfm_C_file_name,sizeof(alt),alt,pre,(char*)0) )
	{	pfmExpandargs( alt, sizeof(alt) );
		return _pfmMnuFileRelinkfile( alt, symlink );
	}
	(void)m;
	return 0;
}
#endif


/********************************************************************
*	removing files
********************************************************************/
/*
 *	Remove file from current working directory
 */
__DF(
int pfmFileRemovefile,( data ), x_filedat *data )
{
	const char *cp;
	int	error;
	errno = 0;

	if (_s_isdir(*data) && !_s_islnk(*data))
	{	error = rmdir(data->nm);
		cp = "rmdir()";
	}
	else
	{	error = unlink(data->nm);
		cp = "unlink()";
	}
	if (error)
		pfmTermErrormessage( strerror(errno), data->nm, cp, (char*)0 );

	return error;
}


/*
 *	Remove the file at the cursor position
 */
__DF(
int pfmMnuFileRemovefile,(m), x_menubar *m )
{
	x_filedat  *p = pfmCwdCurrentfile();

	if (p && _pfmFileConfirmaction(pfm_C_confirm,p->nm,FALSE))
	{
		if ( pfmFileRemovefile(p) ) return 1;
		pfmCwdRereadChanged();
	}
	(void)m;
	return 0;
}


/********************************************************************
*	editing, viewing, printing files
********************************************************************/
/*
 *	Call the editor
 */
__DF(
int pfmFileEditfile,( editor, wait, name ),
const char	*editor __j__
int			wait __j__
const char	*name )
{
	if (name)
	{	unsigned size = pfmVExpandsize(editor,name,(char*)0,(char*)0) + 2;
		char	*line = (char*) alloca(size);
		if (line)
		{	strcpy( line, editor );
			pfmVExpandargs( line, size, name, (char*)0, (char*)0 );
			/* Assume, no file besides 'name' will be changed ;-) */
			pfmExecshell( line, (wait ? EXECSH_WAIT:0) | EXECSH_NORRD );
		}
	}
	return 0;
}

/*
 *	Edit the file at the cursor position
 */
__DF(
int pfmMnuFileEditfile,(m), x_menubar *m )
{
	x_filedat	*p = pfmCwdCurrentfile();
	if (p)
	{	struct stat	sbuf;
		time_t mt = p->st.mtime;
		pfmFileEditfile( pfmConfEditorCmd(), 0, p->nm );
		pfmFileStat( p->nm, &sbuf, &p->st );
		if (mt != p->st.mtime)
			pfmTermRedraw();
	}
	(void)m;
	return 0;
}

/*
 *	Edit named file
 */
__DF(
int pfmMnuFileEditdeffile,(m), x_menubar *m )
{
	char	name[CWD_MAX_LNG];

	if ( !pfmTermGetstrl(pfm_C_file_name,sizeof(name),name,
							pfmFileSGetLastdest((char*)0),(char*)0) )
	{	pfmExpandargs( name, sizeof(name) );
		pfmFileEditfile( pfmConfEditorCmd(), 1, name );
		pfmFileSGetLastdest(name);
		pfmCwdRereadChanged();
	}
	(void)m;
	return 0;
}


/*
 *	View the file or chdir to the directory  at the cursor position
 */
__DF(
int pfmMnuFileViewfile,(m), x_menubar *m )
{
	x_filedat	*info = pfmCwdCurrentfile();

	if (info)
	{	if ( _s_isdir(*info) )
			return pfmCwdChangenamedir(info->nm);
		#if HAS_SYMLINK
			/*
			 *	If the symlink does not point to an existing file,
			 *	try to handle it as URL.
			 */
			if (_s_islnk(*info) && access(info->nm,F_OK) != 0)
			{	struct stat	lbuf;
				if ( !lstat(info->nm,&lbuf) )
				{	int 	n = lbuf.st_size & 0x3fff;
					char	*buf = (char*) alloca(n+2);
					if (buf)
					{	n = readlink( info->nm, buf, n );
						if (n > 0)
						{	buf[n] = 0;
							if (pfmFileGotoURL(buf,FALSE) > 0)
								return 0;
						}
					}
				}
				_pfmFileShowinfo(info);
				return (-1);
			}
		#endif
		return pfmViewFileviewer(info->nm);
	}
	(void)m;
	return 0;
}

/*
 *	View the file at the cursor position, regardless of type
 */
__DF(
int pfmMnuFileViewanyfile,(m), x_menubar *m )
{
	x_filedat	*info = pfmCwdCurrentfile();
	(void)m;
	return (info ? pfmViewFileviewer(info->nm) : 0);
}


/*
 *	Same action as by pfmMnuFileViewanyfile(), just for explicit
 *	'name' instead for '%F'
 */
__DF(
int pfmMnuFileViewdeffile,(m), x_menubar *m )
{
	char	name[CWD_MAX_LNG];

	if ( !pfmTermGetstrl(pfm_C_file_name,sizeof(name),name,
							pfmFileSGetLastdest((char*)0),(char*)0) )
	{	pfmExpandargs( name, sizeof(name) );
		pfmViewFileviewer(name);
		pfmFileSGetLastdest(name);
	}
	(void)m;
	return 0;
}


/*
 *	Print the file at the cursor position
 */
__DF(
int pfmMnuFilePrintfile,(m), x_menubar *m )
{
	char	cmd[CWD_MAX_LNG];

	if ( !pfmTermGetstrl("",sizeof(cmd),cmd,
	                     pfmConfPrintCmd(),(char*)0) )
	{	pfmExpandargs( cmd, sizeof(cmd) );
		pfmExecshell( cmd, EXECSH_WAIT|EXECSH_NORRD );
	}
	(void)m;
	return 0;
}

/***** eof *********************************************************/
