/********************************************************************
* Copyright (c) 1995 by p.fuerst	-	All rights reserved
*
* Redistribution and use in source and binary forms are permitted
* provided that the above copyright notice and this paragraph are
* duplicated in all such forms and that any documentation,
* advertising materials, and other materials related to such
* distribution and use acknowledge that the software was developed
* by peter fuerst, hitzlerieder strasse 5, D-87637 seeg.
*
* THIS SOFTWARE IS PROVIDED BY ITS AUTHOR AND CONTRIBUTORS ``AS IS'' AND
* WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
* THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
* THE POSSIBILITY OF SUCH DAMAGE.
*
*	@(#)ckunix.h	10/25/95
*
*	Wed May  8 09:41:07 1996 (831541267)
*	Mon Jul  8 00:29:48 1996 (836778588)
*	Thu Nov 27 18:24:56 1997 (880651496)
*	Wed Dec 24 22:33:40 1997 (882999220)
*	Sat Apr  4 00:28:16 1998 (891642496)
*	Sat Feb 24 23:58:40 2001 (983055520)
*	Sat Jan 21 00:35:37 2006 (1137800137)
*
*	Check for some specific unix(-like) operating system
*	and create appropriate definitions.
*
*	NOTE: If you use a preprocessor, that does not understand
*	'defined(macro)', you'll need to rwrite this header file.
*
********************************************************************/

#ifndef __CKUNIX_H__
#define __CKUNIX_H__ 1

/***** check for a specific unix(-like) operating system ***********/

#if defined(COHERENT) && !defined(__COHERENT__)
# define __COHERENT__	COHERENT
#endif
#ifdef __COHERENT__
# if defined(__I386__) || defined(_I386)
#  define __COHERENT_386__	386
# else
#  define __COHERENT_286__	286
# endif
# define __UNIX_OS__	"Coherent"
#endif

#if defined(linux) && !defined(__linux__)
# define __linux__	linux
#endif
#ifdef __linux__
# define __UNIX_OS__	"Linux"
#endif

#if defined(FreeBSD) && !defined(__FreeBSD__)
# define __FreeBSD__	FreeBSD
# ifndef __BSD__
#  define __BSD__	__FreeBSD__
# endif
#endif
#if defined(NetBSD) && !defined(__NetBSD__)
# define __NetBSD__	NetBSD
# ifndef __BSD__
#  define __BSD__	__NetBSD__
# endif
#endif
#if defined(BSD) && !defined(__BSD__)
# define __BSD__	BSD
#endif
#ifdef __FreeBSD__
# define __UNIX_OS__	"FreeBSD"
#elif defined(__NetBSD__)
# define __UNIX_OS__	"NetBSD"
#elif defined(__BSD__)
# define __UNIX_OS__	"BSD"
#endif

#if defined(solaris) && !defined(__solaris__)
# define __solaris__	solaris
#endif
#if defined(sun) && !defined(__sun__)
# define __sun__	sun
#endif
#if defined(__solaris__)
# define __UNIX_OS__	"Solaris"
#elif defined(__sun__)
# define __UNIX_OS__	"Sun OS"
#endif

#ifdef _AIX /* is this reliable ? */
# ifndef __AIX__
#  define __AIX__	_AIX
# endif
# define __UNIX_OS__	"AIX"
#endif

#if defined(__sgi) && !defined(__sgi__)
# define __sgi__	__sgi
#endif
#if defined(sgi) && !defined(__sgi__)
# define __sgi__	sgi
#endif
#ifdef __sgi__
# ifndef __IRIX__
#  define __IRIX__	__sgi__
# endif
# define __UNIX_OS__	"IRIX"
#endif

#if defined(unix) && !defined(__unix__)
# define __unix__	unix
#endif
/*
 *	Tests for a not yet handled Unix system should be appended here.
 *	A/UX, HPUX, Irix, MacOSX, Minix, NeXT, OSF, SCO, Sinix, True64,
 *	Ultrix, Unicos, Xenix, ...
 */
#if defined(__unix__) && !defined(__UNIX_OS__)
# define __UNIX_OS__	"Generic Unix"
#endif

/***** help handling prototypes ... ********************************/

#ifndef c89_fct_declaration_and_definition

/* NOTE: "ifdef __STDC__" would *not* do !!! */
#if defined(__STDC__) && __STDC__ != 0
# define c89_fct_declaration_and_definition 1
#elif defined(__GNUC__) /* __STDC__ may not be defined, due to '-traditional' */
# define c89_fct_declaration_and_definition 1
#elif defined(__cplusplus)
# define c89_fct_declaration_and_definition 1
/* found the following two tests in some header, don't know if reliable */
#elif defined(__AIX__)
# define c89_fct_declaration_and_definition 1
#elif defined (__mips) && defined (_SYSTYPE_SVR4) /* is this IRIX ? */
# define c89_fct_declaration_and_definition 1
#endif

#endif
/*
 * __DC( typ function,(typ1 arg1, . . . ,...) );
 * __DF( typ function,(arg1, . . .), typ1 arg1 __j__ . . . VARARGS ) {...}
 *
 * (`__j__' is choosen just because of its visual advantages: resemblance
 * with both `,' and `;' and not adding too much clutter to the text)
 */
#ifdef c89_fct_declaration_and_definition /* ANSI/ISO-standard syntax */
# undef c89_fct_declaration_and_definition
# define __DC(func,argnamelist) func argnamelist
# define __DF(func,argnamelist,argdeclist) func (argdeclist)
# define VARARGS , ...
# define __j__ ,
#else /* K & R syntax */
# define __DC(func,argnamelist) func ()
# define __DF(func,argnamelist,argdeclist) func argnamelist argdeclist;
# define VARARGS
# define __j__ ;
#endif

#ifndef __GNUC__
# ifdef __cplusplus
#  define __inline__	inline
# else
#  define __inline__
# endif
#endif

#ifdef __COHERENT__
# ifdef __COHERENT_286__
/* the desireable 'typedef ...' is likely to conflict with stdio.h,... */
#  include <stddef.h>
# else
#  ifndef __GNUC__
#   undef Void
/*  this struct would serve 'Void*' better than '#define Void char' */
    typedef struct{} Void;
#   define Void Void
#  endif
# endif
#endif
#ifndef Void
# define Void void
#endif

#endif /* __CKUNIX_H__ */

/***** eof "ckunix.h" **********************************************/
