/*
 * Copyright (c) 1996 peter fuerst	-	All rights reserved
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by peter fuerst, hitzlerieder strasse 5, D-87637 seeg.
 *
 * THIS SOFTWARE IS PROVIDED BY ITS AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 *	execvpe.c
 *
 *	Provide execvpe() on OSs, where it is not available.
 */
#include <unistd.h> /* exec...() */
#include <stdlib.h>
#include <string.h>
#include <errno.h>
/* identify OS and get a proper definition of the __DC(), __DF() macros */
#include "ckunix.h"
#ifdef __COHERENT_286__
# include <access.h>
# define X_OK AEXEC
#endif

#ifdef __cplusplus
 extern "C" {
#endif

#if !_SYS_HAS_EXECVPE_

#ifdef __COHERENT__
# include <path.h>
#else
# define EXEPATHMAXSZ	(128*sizeof(int)*sizeof(int))

__DF(
static const char *_xpath,( path, buf ),
const char	*file __j__
char		buf[EXEPATHMAXSZ] )
{
	/* Note: execve("file",...) is in effect execve("./file",...) */
	char *pp;
	int lx;

	if (!file || strchr(file,'/'))
		return file;

	if ( !(pp = getenv("PATH")) )
		return file; /* see note ! */

	for (lx = strlen(file); ; )
	{	char *ep = strchr( pp, ':' );
		if ( !ep ) ep = strchr( pp, 0 );
		if (ep > pp)
		{	int ld = ep-pp;
			if (EXEPATHMAXSZ < ld+lx+2)
			{	errno = E2BIG;
				return 0;
			}
			memcpy( buf, pp, ld );
			buf[ld] = '/';
			strcpy( buf+ld+1, file );
			if (0 == access(buf,X_OK))
				return buf;
		}
		if ( !*ep ) break;
		pp = ep + 1;
	}
	errno = EACCES;
	return 0; /* see note ! */
}
#endif	/* !__COHERENT__ */

__DF(
int execvpe,( file, argv, envp ),
const char	*file __j__
char* const	argv[] __j__
char* const	envp[] )
{
	const char *pp;
	#ifdef __COHERENT__
		pp = path( getenv("PATH"), file, X_OK );
	#else
		char	buf[EXEPATHMAXSZ];
		pp = _xpath( file, buf );
	#endif
	return pp ? execve(pp, argv, envp) : -1;
}
#endif /* !_SYS_HAS_EXECVPE_ */

#ifdef __cplusplus
 } // extern "C"
#endif

/* eof */
