/*
 * Copyright (c) 1996 peter fuerst	-	All rights reserved
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by peter fuerst, hitzlerieder strasse 5, D-87637 seeg.
 *
 * THIS SOFTWARE IS PROVIDED BY ITS AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 *	The convenient 'spawn'-functions, completing the set of 'exec'-functions.
 *
 *	spawn.c	-	internal backend and global 'spawn'-functions.
 *
 *	Sat Sep  7 22:14:11 1996 (842127251)
 *	Wed Jul  8 11:00:07 1998 (899888407)
 */
#ifdef __linux__
# define _GNU_SOURCE 1 /* exexcvpe() may be available */
#endif
#include <unistd.h> /* fork(), pipe(), ... */
#include <stdlib.h>	/* exit() */
#include <stdio.h>
#include <errno.h>
#include "spawn.h"	/* includes ckunix.h, sys/wait.h, spawn*-prototypes */
#if !_SYS_HAS_EXECVPE_
 __DC( int execvpe,(const char*, char*const[], char*const[]) );
#endif

typedef __DC( int(*EXECFCT),(const char*,char*const[],char*const[]) );

#ifdef __cplusplus
 extern "C" {
#endif

#define _spawnv _ZL7spawnvPFiPKcPKPcS3_EiS0_S3_S3_Pi
__DC(
int _spawnv,(EXECFCT, int, const char*, char*const[], char*const[], int [2]) );

#ifndef __SPAWN_FRONT__	/* internal backend */

__DF(
int _spawnv,( exec, mode, path, argv, envp, pdes ),
EXECFCT     exec    __j__
int         mode    __j__
const char  *path   __j__
char* const argv[]  __j__
char* const envp[]  __j__
int         pdes[2] )
{
	if ( pdes ) pdes[1] = pdes[0] = -1;

	switch ( mode )
	{	int pid, c2pfd[2], p2cfd[2];
		case P_NOWAIT:
		case P_WAIT:
			if ( pdes )
			{	/*
				 * NOTE: in P_WAIT mode we are in danger of a deadlock !
				 * The child may fill up the pipe (c2pfd) and subsequently
				 * be blocked until a read is done on this pipe, which will
				 * not occur before the child has finished...
				 */
				if (pipe(c2pfd) < 0)
					return -1;
				if (pipe(p2cfd) < 0)
				{	close(c2pfd[0]);
					return -1;
				}
			}

			if ((pid = fork()) < 0)
			{	if ( pdes )
				{	close(p2cfd[0]);
					close(p2cfd[1]);
					close(c2pfd[0]);
					close(c2pfd[1]);
				}
				return -1;
			}
			if ( pid )	/* Parent process got a child */
			{
				if ( pdes )
				{	/* parent --> p2c[1] -> p2c[0] --> child */
					pdes[1] = p2cfd[1];
					/* parent <-- c2p[0] <- c2p[1] <-- child */
					pdes[0] = c2pfd[0];
				}
				while (P_WAIT == mode)
				{	/* waitpid() is preferable over this loop, but
					 * may be unavailable on some non-POSIX-system
					 */
					int status;
					int wid = wait(&status);
					if (wid == pid) return status;
					if (wid <= 0) return -1;
				}
				return pid;
			}
			/* Child process, fall trough to exec */
			if ( pdes )
			{	/* stdin  <-- p2c[0] <- p2c[1] <-- parent */
				dup2( p2cfd[0], fileno(stdin) );
				close(p2cfd[0]);
				/* stdout --> c2p[1] -> c2p[0] --> parent */
				dup2( c2pfd[1], fileno(stdout) );
				close(c2pfd[1]);
			}
		case P_OVERLAY:
			/* NOTE: The following call requires its parameters
			 * to be pushed in standard "C" order, since
			 * 'exec' may need 2 or 3 arguments !!!
			 */
			exit( (*exec)(path,argv,envp) );
		default:
			errno = EINVAL;
			return -1;
	}
}

#else /* global 'spawn'-functions */

#if __SPAWN_FRONT__ == 'v'
/*
 *	'spawn'-functions referencing 'execv'
 */
__DF(
int pspawnv,( mode, path, argv, pdes ),
int			mode __j__
const char	*path __j__
char* const	argv[] __j__
int         pdes[2] )
{
	return _spawnv( (EXECFCT)execv, mode, path, argv, NULL, pdes );
}

__DF(
int spawnv,( mode, path, argv ),
int			mode __j__
const char	*path __j__
char* const	argv[] )
{
	return _spawnv( (EXECFCT)execv, mode, path, argv, NULL, (int*)0 );
}

__DF(
int spawnl,( mode, path, arg0 /*,...,NULL*/),
int			mode __j__
const char	*path __j__
const char	*arg0
VARARGS )
{	/*
	 * NOTE: Taking &arg0 as (char*[]) requires, that standard
	 * "C" parameter-passing conventions are used !!!
	 */
	return _spawnv( (EXECFCT)execv, mode, path, (char**)&arg0, NULL, (int*)0 );
}

#elif __SPAWN_FRONT__ == 'e'
/*
 *	'spawn'-functions referencing 'execve'
 */
__DF(
int pspawnve,( mode, path, argv, envp, pdes ),
int			mode __j__
const char	*path __j__
char* const	argv[] __j__
char* const	envp[] __j__
int         pdes[2] )
{
	return _spawnv( execve, mode, path, argv, envp, pdes );
}

__DF(
int spawnve,( mode, path, argv, envp ),
int			mode __j__
const char	*path __j__
char* const	argv[] __j__
char* const	envp[] )
{
	return _spawnv( execve, mode, path, argv, envp, (int*)0 );
}

__DF(
int spawnle,( mode, path, arg0 /*,...,NULL,envp*/),
int			mode __j__
const char	*path __j__
const char	*arg0
VARARGS )
{	/*
	 * NOTE: Taking &arg0 as (char*[]) requires, that standard
	 * "C" parameter-passing conventions are used !!!
	 */
	const char	**envp = &arg0;
	while (*envp) envp++;
	++envp;
	return _spawnv( execve, mode, path, (char**)&arg0, (char**)*envp, (int*)0 );
}

#elif __SPAWN_FRONT__ == 'V'
/*
 *	'spawn'-functions referencing 'execvp'
 */
__DF(
int pspawnvp,( mode, path, argv, pdes ),
int			mode __j__
const char	*path __j__
char* const	argv[] __j__
int         pdes[2] )
{
	return _spawnv( (EXECFCT)execvp, mode, path, argv, NULL, pdes );
}

__DF(
int spawnvp,( mode, path, argv ),
int			mode __j__
const char	*path __j__
char* const	argv[] )
{
	return _spawnv( (EXECFCT)execvp, mode, path, argv, NULL, (int*)0 );
}

__DF(
int spawnlp,( mode, path, arg0 /*,...,NULL*/),
int			mode __j__
const char	*path __j__
const char	*arg0
VARARGS )
{	/*
	 * NOTE: Taking &arg0 as (char*[]) requires, that standard
	 * "C" parameter-passing conventions are used !!!
	 */
	return _spawnv( (EXECFCT)execvp, mode, path, (char**)&arg0, NULL, (int*)0 );
}

#elif __SPAWN_FRONT__ == 'E'
/*
 *	'spawn'-functions referencing 'execvpe'
 */
__DF(
int pspawnvpe,( mode, path, argv, envp, pdes ),
int			mode __j__
const char	*path __j__
char* const	argv[] __j__
char* const	envp[] __j__
int         pdes[2] )
{
	return _spawnv( execvpe, mode, path, argv, envp, pdes );
}

__DF(
int spawnvpe,( mode, path, argv, envp ),
int			mode __j__
const char	*path __j__
char* const	argv[] __j__
char* const	envp[] )
{
	return _spawnv( execvpe, mode, path, argv, envp, (int*)0 );
}

__DF(
int spawnlpe,( mode, path, arg0 /*,...,NULL,envp*/),
int			mode __j__
const char	*path __j__
const char	*arg0
VARARGS )
{	/*
	 * NOTE: Taking &arg0 as (char*[]) requires, that standard
	 * "C" parameter-passing conventions are used !!!
	 */
	const char	**envp = &arg0;
	while (*envp) envp++;
	++envp;
	return _spawnv( execvpe, mode, path, (char**)&arg0, (char**)*envp, (int*)0 );
}

#else
# error __SPAWN_FRONT__ must be from {'v', 'e', 'V', 'E'}
#endif

#endif /* __SPAWN_FRONT__ defined */

#ifdef __cplusplus
 } // extern "C"
#endif

/* eof */
