/*
 * Copyright (c) 1996-2025 peter fuerst	-	All rights reserved
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by peter fuerst, hitzlerieder strasse 5, D-87637 seeg.
 *
 * THIS SOFTWARE IS PROVIDED BY ITS AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 *	strlib.c
 *
 *	some usefull string functions
 *
 *	Sat Sep  7 22:14:11 1996 (842127251)
 *	Wed Dec 24 20:12:08 1997 (882990728)
 *	Thu Jan  1 00:33:00 1998 (883611180)
 */
/*$preferred_tabstop=4*/
#ifdef COHERENT
# ifndef _I386
#  define __COHERENT_286__
# endif
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/times.h>
#ifndef __COHERENT_286__
# include <unistd.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include "ckunix.h"

#ifdef COHERENT
# define _NEED_STRCASECMP_	1
# define _NEED_STRLWRUPR_	1
# define _NEED_STRDUP_	1 /* declared on 4.2, but not in any library */
#endif
#ifdef linux
# define _NEED_STRLWRUPR_	1
#endif
#ifndef _NEED_STRDUP_
# define _NEED_STRDUP_	0
#endif
#ifndef _NEED_STRCASECMP_
# define _NEED_STRCASECMP_	0
#endif
#ifndef _NEED_STRLWRUPR_
# define _NEED_STRLWRUPR_	0
#endif
#define _NEED_MEMCASECMP_	1


#if _NEED_STRDUP_
/*
 *	System V, BSD 4.3 - Save a string in the arena. Header 'string.h'.
 */
#ifdef	_BUILD_STRDUP_OBJ_

__DF( char *strdup,( str ), const char *str )
{
	if (str)
	{	char *p = (char*)malloc( strlen(str)+1 );
		if (p) return strcpy( p, str );
	}
	return NULL;
}

__DF( char *strndup,( str, n ), const char *str __j__ size_t n )
{
	if (str)
	{	char  *p;
		size_t l = strlen(str);
		if (l < n) n = l;
		if ((p = (char*)malloc(n+1)) != NULL)
		{	for (l = 0; l < n; l++) p[l] = *str++;
			p[n] = 0;
			return p;
		}
	}
	return NULL;
}
#endif
#endif


#if _NEED_STRCASECMP_
/*
 *	BSD 4.3 - string compare in case-insensitive manner.
 */
#ifdef	_BUILD_STRCASECMP_OBJ_

__DF( int strcasecmp,( s1, s2 ), const char *s1 __j__ const char *s2 )
{
	for ( ; *s1 && *s2; s1++, s2++)
	{	int d = tolower(*s1) - tolower(*s2);
		if (d) return d;
	}	
	return 0;
}

#ifdef __GNUC__
# ifdef __mips
   __asm__("stricmp: j strcasecmp; nop");
   __asm__("strcmpi: j strcasecmp; nop");
#  define strcmpi_jmp
# endif
# ifdef __i386__
   __asm__("stricmp: jmp strcasecmp");
   __asm__("strcmpi: jmp strcasecmp");
#  define strcmpi_jmp
# endif
#endif
#ifndef strcmpi_jmp
__DF( int stricmp,( s1, s2 ), const char *s1 __j__ const char *s2 )
{	return strcasecmp(s1,s2);
	}
__DF( int strcmpi,( s1, s2 ), const char *s1 __j__ const char *s2 )
{	return strcasecmp(s1,s2);
	}
#endif

#endif /* __BUILD_... */

#ifdef	_BUILD_STRNCASECMP_OBJ_

__DF(
int strncasecmp,( s1, s2, n ),
const char *s1 __j__
const char *s2 __j__
size_t n )
{
	for ( ; n && *s1 && *s2; s1++, s2++, n--)
	{	int d = tolower(*s1) - tolower(*s2);
		if (d) return d;
	}	
	return 0;
}

#ifdef __GNUC__
# ifdef __mips
   __asm__("strnicmp: j strncasecmp; nop");
#  define strnicmp_jmp
# endif
# ifdef __i386__
   __asm__("strnicmp: jmp strncasecmp");
#  define strnicmp_jmp
# endif
#endif
#ifndef strnicmp_jmp
__DF(
int strnicmp,( s1, s2, n ),
const char *s1 __j__
const char *s2 __j__
size_t n )
{	return strncasecmp(s1,s2,n);
	}
#endif

#endif /* __BUILD_... */
#endif /* __NEED_... */


#if _NEED_STRLWRUPR_
/*
 *	Convert all characters of a string to lower/upper case.
 */
#ifdef	_BUILD_STRLWR_OBJ_

__DF( char *strlwr,( str ), char *str )
{
	char *p;
	for (p = str; *p; p++) *p = tolower(*p);
	return str;
}
#endif /* __BUILD_... */

#ifdef	_BUILD_STRUPR_OBJ_

__DF( char *strupr,( str ), char *str )
{
	char *p;
	for (p = str; *p; p++) *p = toupper(*p);
	return str;
}
#endif /* __BUILD_... */
#endif /* __NEED_... */


#if _NEED_MEMCASECMP_
/*
 *	like memcmp(), but in case-insensitive manner.
 */
#ifdef	_BUILD_MEMCASECMP_OBJ_

__DF(
int memcasecmp,( v1, v2, n ),
const char *v1 __j__
const char *v2 __j__
size_t n )
{
	if (v1) 
	{	if (v2) 
		{	const char *s1, *s2, *pe;
			for (s1=v1, s2=v2, pe=s1+n; s1 < pe; s1++, s2++)
			{	int	n1 = tolower((unsigned char)*s1);
				int	n2 = tolower((unsigned char)*s2);
				if (n1 < n2) return -1;
				if (n1 > n2) return 1;
			}
			return 0;
		}
		return 1;
	}
	return (v2 ? -1:0);
}

#ifdef __GNUC__
# ifdef __mips
   __asm__("memicmp: j memcasecmp; nop");
#  define memicmp_jmp
# endif
# ifdef __i386__
   __asm__("memicmp: jmp memcasecmp");
#  define memicmp_jmp
# endif
#endif
#ifndef memicmp_jmp
__DF(
int memicmp,( v1, v2, n ),
const char *v1 __j__
const char *v2 __j__
size_t n )
{	return memcasecmp(v1,v2,n);
	}
#endif

#endif
#endif


/* eof */
