/***********************************************************************
* Copyright (c) 1995-2025 by peter fuerst
*
*	This program is free software; you can redistribute it and/or modify
*	it - as explained in the accompaning file COPYING - under the terms
*	of the GNU General Public License as published by the Free Software
*	Foundation; either version 2, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	Author: Peter Frst, post@pfrst.de
*
*	@(#)pfm.c	10/25/95
*
***********************************************//*$preferred tabstop=4*/

#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <ctype.h>
#include <dirent.h>
#include <time.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <grp.h>
#include <pwd.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>
#include "termuxx.h"
#include "spawn.h"


/***** define & typedef ********************************************/

#include "pfm.h"
#include "pfmstat.h"

#ifndef const_cast
 #define const_cast(typ,val)	((typ)val)
#endif


/***** forward declaration *****************************************/

__DC( static int  pfmMnuHelp,(x_menubar*) );
__DC( static int  pfmMnuRefresh,(x_menubar*) );
__DC( static void pfmSignalHandler,(int) );
__DC( static int  pfmSystemEx,(const char*,bool) );
static void pfmReinitscreen ();


/***** global ******************************************************/

#define BOLD "\007" /* use BELl to emphasize the next character */
static const char  mainfnkey[] =
		"F " BOLD"1-up " BOLD"2-dn " BOLD"3-top " BOLD"4-bot "
		BOLD"5-reread " BOLD"6-sort " BOLD"7-swap " BOLD"8-in/exclude "
		BOLD"9-back " BOLD"1"BOLD"0-multiple";

static const x_menubar	fnkeymenu[] =
{
	{ '1', "", (__DC(int(*),(x_menubar*))) pfmCwdWin1lineUp },
	{ '2', "", (__DC(int(*),(x_menubar*))) pfmCwdWin1lineDown },
	{ '3', "", (__DC(int(*),(x_menubar*))) pfmCwdWinTopscroll },
	{ '4', "", (__DC(int(*),(x_menubar*))) pfmCwdWinEndscroll },
	{ '5', "", (__DC(int(*),(x_menubar*))) pfmCwdReread /*pfmMnuUpdatemenu*/ },
	{ '6', "", (__DC(int(*),(x_menubar*))) pfmCwdSortdir },
	{ '7', "", (__DC(int(*),(x_menubar*))) pfmCwdChangedir },
	{ '8', "", (__DC(int(*),(x_menubar*))) pfmMnuMarkcurrent },
	{ '9', "", (__DC(int(*),(x_menubar*))) pfmCwdPreviousdir },
	{ '0', "", (__DC(int(*),(x_menubar*))) pfmFileMultiplefile },
	{ 0, "", pfmMnuNop }
};

const char	*pfm_Domainname = "";
const char	*pfm_Hostname = "";
const char	*pfm_Grpname = "";
const char	*pfm_eGrpname = "";
const char	*pfm_Usrname = "";
const char	*pfm_eUsrname = "";

const x_menubar	*pfm_Currentmenu = fnkeymenu;
const char *pfm_Currentfnkeys = mainfnkey;
const char *pfm_argv0 = "pfm";


/***** local *******************************************************/

static const char pfmversion[] =
	"\rpfm version 0.9, copyright (c) 1995-2003 by peter fuerst.\n"
	"\rThis is free software, that comes with ABSOLUTELY NO WARRANTY.\n"
	"\rYou are welcome to redistribute it under the conditions of the\n"
	"\rGNU General Public License, as described in the accompaning\n"
	"\rfile COPYING.\n\n";

static bool childprocess_is_active = 0;
static const jmp_buf	*pfm_jmpbuf = (const jmp_buf*)0;
static char	*pfmhome = ".";	/* $HOME */
static char tmpfilename[L_tmpnam] = "";

#ifdef DEUTSCH
 static const char quit_txt[] = "pfm wirklich beenden ?";
#else
 static const char quit_txt[] = "Really quit pfm ?";
#endif


/********************************************************************
*	pfm startup, closure and fundamental functions
********************************************************************/
/*
 *	dummy function
 */
__DF(
int pfmMnuNop,(m), x_menubar *m ) { (void)m; return 0; }

/*
 *	tells 'pfmMnuRunmenu' to return
 */
__DF(
int pfmMnuCancelmenu,(m), x_menubar *m ) { (void)m; return -1; }

/*
 *	get user name by real user id
 */
__DF(
char  *pfmGetusername,( uid ), int uid )
{
	struct passwd	*p = getpwuid(uid);

	return (p ? p->pw_name:"");
}

/*
 *	get group name by group id
 */
__DF(
char  *pfmGetgroupname,( gid ), int gid )
{
	struct group	*g = getgrgid(gid);

	return (g ? g->gr_name:"");
}

/*
 *	save a string - in place, if possible
 */
__DF(
char  *pfmStrdup,( newstr, oldstr ), const char *newstr __j__ char *oldstr )
{
	if ( newstr )
	{	char *p = (char*) realloc( oldstr, strlen(newstr)+1 );
		if (p) return strcpy( p, newstr );
	}
	return oldstr;
}

__DF(
const char  *pfmStrtail,( str, len ), const char *str __j__ size_t len )
{
	if ( str )
	{	size_t n = strlen(str);
		if (n > len) str += n-len;
	}
	return str;
}

__DF(
int pfmStrmatch,( str, smatch ), const char *str __j__ const char *smatch )
{
	if ( !smatch ) smatch = "";
	if ( !str ) str = "";
	return !strncmp( str, smatch, strlen(smatch) );
}

__DF(
int pfmStrcasematch,( str, smatch ), const char *str __j__ const char *smatch )
{
	if ( !smatch ) smatch = "";
	if ( !str ) str = "";
	return !strncasecmp( str, smatch, strlen(smatch) );
}

__DF(
int pfmMatchname,( pattern, name ), const char *pattern __j__ const char *name )
{
#if defined(__linux__) ||\
	defined(__FreeBSD__) || defined(__NetBSD__) ||\
	defined(__AIX__) || defined(__IRIX__)
	return !fnmatch( pattern, name, FNM_PERIOD );
#endif
#ifdef __COHERENT__
	return (*name != '.' || '.' == *pattern ? pnmatch(name,pattern,0):0);
#endif
	return 0; /* never match on unknown systems ;-) */
}


/*
 *	functions for expanding %F, %N, %X, %D, %A, %H
 */

/*
 *	expand a pattern, which might contain the wildcards
 */
__DF(
char  *pfmVExpandargs,( cmd, size, fname, curdir, altdir ),
char	cmd[/*size*/] __j__
int 	size __j__
const char	*fname __j__
const char	*curdir __j__
const char	*altdir )
{
	if (cmd && size > 0)
	{
		int len = pfmVExpandsize( cmd, fname, curdir, altdir );

		if (len > strlen(cmd) || strchr(cmd,'\\'))
		{	const char *sfx;
			char *buf;
			if (!curdir) curdir = pfmCwdCWD();
			if (!altdir) altdir = pfmCwdAltDir();
			if ( !fname )
			{	static const char noexpand[] =
				{	pfm_fwildc_start, pfm_Fname_fwildc, 0
				};
				x_filedat *x = pfmCwdCurrentfile();
				fname = x ? x->nm:noexpand;
			}
			sfx = strrchr( fname, '.' );
			if (!sfx) sfx = strchr( fname, 0 );

			if ( (buf = (char*)alloca(len+2)) )
			{	const char *v;
				char *p, *q;
				for (p = buf, q = cmd; (*p = *q); q++)
					switch ( *q )
					{	case '\\':
							if ( (*p = q[1]) )
							{	q++;
								p++;
							}
							continue;
						case '~':
							p = strchr( strcpy(p,pfmhome), 0 );
							continue;
						case pfm_fwildc_start:
							switch ( q[1] )
							{	case pfm_Fname_fwildc:
									p = strchr( strcpy(p,fname), 0 );
									q++;
									continue;
								case pfm_Nosuff_fwildc:
									for (v = fname; v != sfx; ++p,++v) *p=*v;
									*p = 0;
									q++;
									continue;
								case pfm_Suffix_fwildc:
									p = strchr( strcpy(p,sfx), 0 );
									q++;
									continue;
								case pfm_CWD_fwildc:
									p = strchr( strcpy(p,curdir), 0 );
									q++;
									continue;
								case pfm_AltD_fwildc:
									p = strchr( strcpy(p,altdir), 0 );
									q++;
									continue;
								case pfm_HOME_fwildc:
									p = strchr( strcpy(p,pfmhome), 0 );
									q++;
									continue;
							}
						default:
							p++;
					}
				strncpy( cmd, buf, size );
				cmd[size-1] = 0;
			}
		}
	}
	return cmd;
}

__DF(
char  *pfmExpandargs,( cmd, size ),
char	cmd[/*size*/] __j__
int 	size )
{
	return pfmVExpandargs( cmd, size, (char*)0, (char*)0, (char*)0 );
}


/*
 *	expand a array of patterns, which might contain the wildcards
 */
__DF(
void  pfmExpandargv,( argv, buf, size ),
char	*argv[] __j__
char	buf[/*size*/] __j__
int		size )
{
	int i;
	for (i = 0; size > 0 && argv[i]; i++)
	{	strncpy( buf, argv[i], size )[size-1] = 0;
		argv[i] = pfmExpandargs( buf, size );
		buf = strchr(buf,0) + 1;
		size -= (buf - argv[i]);
	}
}

/*
 *	NB: pfmExpandsize() gives *not* the size of the expanded pattern,
 *	but an upper limit for it !
 *	To be precise, it returns the size of the original pattern plus
 *	the size of all wildcard-expansions. Thus it allows to test for
 *	the presence of (unescaped) wildcards in the pattern.
 */
__DF(
int  pfmVExpandsize,( cmd, fname, curdir, altdir ),
const char	*cmd __j__
const char	*fname __j__
const char	*curdir __j__
const char	*altdir )
{
	if ( cmd )
	{	short	acnt=0, dcnt=0, fcnt=0, hcnt=0, ncnt=0, xcnt=0;
		int 	lng = strlen(cmd);
		char	esc = 0;
		const char	*p;

		for (p = cmd; *p; p++)
			if ('\\' ==  esc)
				/* the explanation, why this statement is crucial,
				 * is left as an exercise to the reader ;-) */
				esc = 0;
			else
			{	/* unescaped current char may escape the next */
				esc = *p;
				if (pfm_fwildc_start == *p)
					switch ( p[1] )
					{	case pfm_Fname_fwildc: fcnt++; break;
						case pfm_Nosuff_fwildc: ncnt++; break;
						case pfm_Suffix_fwildc: xcnt++; break;
						case pfm_AltD_fwildc: acnt++; break;
						case pfm_CWD_fwildc: dcnt++; break;
						case pfm_HOME_fwildc: hcnt++;
					}
				else if ('~' == *p)
					hcnt++;
			}
		if (acnt) lng += acnt*strlen(altdir ? altdir:pfmCwdAltDir());
		if (dcnt) lng += dcnt*strlen(curdir ? curdir:pfmCwdCWD());
		if (hcnt) lng += hcnt*strlen(pfmhome);

		if ( !fname )
		{	x_filedat *x = pfmCwdCurrentfile();
			if (x) fname = x->nm;
		}
		if ( fname )
		{	const char *x = strrchr( fname, '.' );
			int  flg = strlen(fname);
			if (!x) x = fname + flg;
			if (fcnt) lng += fcnt*flg;
			if (ncnt) lng += ncnt*(x - fname);
			if (xcnt) lng += xcnt*(fname + flg - x);
		}
		return lng;
	}
	return 0;
}

__DF(
int  pfmExpandsize,( cmd ), const char *cmd )
{
	return pfmVExpandsize( cmd, (char*)0, (char*)0, (char*)0 );
}


/*
 *	executing (shell) commands
 */
__DF(
static int pfmSystemEx,( command, chldstat ),
const char *command __j__
bool chldstat )
{	/*
	 *	unfortunately system() is hardwired to [/usr]/bin/sh ...
	 */
	static char	*shell = 0;
	char	*argv[4];
	int		r;

	if (!shell)
	{	if ( (shell = getenv("SHELL")) )
			shell = *shell ? strdup(shell):0;
		if (!shell) shell = "/bin/sh";
	}
	if (!command) return (access(shell,X_OK) == 0);

	argv[0] = shell;
	argv[1] = "-c";
	argv[2] = const_cast( char*, command );
	argv[3] = 0;

	childprocess_is_active = 1;
	r = spawnv(P_WAIT,argv[0],argv);
	childprocess_is_active = 0;
	#ifdef WIFEXITED
		/* ? we return the child's exitcode
		 * : the child's normal termination always means success */
		if ( WIFEXITED(r) )
			return chldstat ? WEXITSTATUS(r) : 0;
		if ( WIFSIGNALED(r) ) return WTERMSIG(r);
		if ( WIFSTOPPED(r) ) return WSTOPSIG(r);
		return r; /* should not be reached */
	#else
		return (r && 0xff);
	#endif
}

__DF(
int pfmSystem,( command ), const char *command )
{
	return pfmSystemEx( command, 0 );
}


__DF(
void pfmPreExecshell,( mode ), int mode )
{
	pfmTermClearscreen();
	pfmTermResetty();
	(void)mode;
}

__DF(
void pfmPostExecshell,( mode ), int mode )
{
	pfmTermFixmode();
	pfmTermClrcon();
	if (mode & EXECSH_WAIT)
	{	fputs( pfm_C_hit_any_key, stdout );
		fflush(stdout);
		while (pfmTermGetch() < 1);
	}
	pfmTermRefresh(TRUE);
	if ( !(mode & EXECSH_NORRD) )
		pfmCwdRereadChanged();
}


__DF(
int pfmExecshell,( cmd, mode ), const char *cmd __j__ int mode )
{
	while (isspace(*cmd)) cmd++;

	if (*cmd)
	{	int r;
		pfmPreExecshell(mode);
		r = pfmSystemEx( cmd, 0 );
		pfmPostExecshell(mode);
		return r;
	}
	return 0;
}

__DF(
int pfmExecommand,( cmd, mode ), const char *cmd __j__ int mode )
{
	while (isspace(*cmd)) cmd++;

	if (*cmd)
	{	int r;
		pfmPreExecshell(mode);
		r = pfmSystemEx( cmd, 1 );
		pfmPostExecshell(mode);
		return r;
	}
	return 0;
}

__DF(
int pfmExecargv,( argv, mode ), char *const argv[] __j__ int mode )
{
	if (argv)
	{	int r;
		pfmPreExecshell(mode);
		childprocess_is_active = 1;
		r = spawnvp(P_WAIT,argv[0],argv);
		childprocess_is_active = 0;
		pfmPostExecshell(mode);
		#ifdef WIFEXITED
			/* here we return the child's exitcode */
			if ( WIFEXITED(r) ) r = WEXITSTATUS(r);
			if ( WIFSIGNALED(r) ) r = WTERMSIG(r);
			if ( WIFSTOPPED(r) ) r = WSTOPSIG(r);
		#else
			r &= 0xff;
		#endif
		return r;
	}
	return 0;
}

__DF(
const char *pfmSGetLastcommand,( newcmd ), const char *newcmd )
{
	static const char null = 0;
	static char	*lastcmd = (char*)&null;

	if ( newcmd )
	{	lastcmd = pfmStrdup( newcmd, lastcmd != &null ? lastcmd:(char*)0 );
		if ( !lastcmd ) lastcmd = (char*)&null;
	}
	return lastcmd;
}


__DF(
int pfmMnuCommand,(m), x_menubar *m )
{
	char	cmd[2*CWD_MAX_LNG];

	if ( !pfmTermGetstrl(pfm_Shellprompt,sizeof(cmd),cmd,
							pfmSGetLastcommand((char*)0),(char*)0) )
	{	pfmSGetLastcommand(cmd); /* before expansion ! */
		pfmExpandargs( cmd, sizeof(cmd) );
		return pfmExecshell(cmd,EXECSH_WAIT);
	}
	(void)m;
	return 0;
}



/*
 *	Rotate the *visible* parts of a menu, especially usefull, when
 *	the menu is too large to fit on screen.
 *	It is really awkward to shift the whole contents of the array, a
 *	linked list could be handled much more elegant :-(  But here it
 *	is not a performance issue and a list would complicate things in
 *	many other places.
 */
__DF(
void pfmLShiftMenu,( menu ), x_menubar *menu )
{
	x_menubar *p1;

	for (p1 = menu; p1->key && p1->txt; p1++)
		if (p1->key != A_ESC && *p1->txt) break;

	if (p1->key && p1->txt)
	{	x_menubar m0, *p2;
		for (m0 = *p1, p2 = p1+1; p2->key && p2->txt; )
			*p1++ = *p2++;
		*p1 = m0;
	}
}

__DF(
void pfmRShiftMenu,( menu ), x_menubar *menu )
{
	x_menubar m0, *p1, *p2;

	/* start at last non-empty label */
	for (p2 = p1 = menu; p1->key && p1->txt; p1++)
		if ( *p1->txt ) p2 = p1;

	while (menu < p2 && (A_ESC == menu->key || !*menu->txt))
		menu++; /* see pfmLShiftMenu */

	for (m0 = *p2; p2 > menu; p2--)
		*p2 = p2[-1];
	*menu = m0;
}


/*
 *	Display a menu and wait for input
 */
__DF(
int pfmMnuRunmenu,( menu, mfnkeys ),
x_menubar  *menu __j__
const char *mfnkeys )
{
	const x_menubar	*pmnu = pfm_Currentmenu;
	const char *pfnk = pfm_Currentfnkeys;

	if ( mfnkeys ) pfm_Currentfnkeys = mfnkeys;
	if ( menu ) pfm_Currentmenu = menu;
	pfmTermRedraw();

	while ( menu )
	{
		short	key = pfmTermGetch();
		short	upkey = key & 0xff80 ? key:toupper(key);
		const x_menubar	*pm;

		switch ( key )
		{	case A_ETX:			/* ^C (End of TeXt) will	*/
			/*case KEY_EXIT:*/	/* always invoke Quit ...	*/
				pfmMnuQuit(menu);
				continue;
			case KEY_BREAK:		/* The keys	*/
			case A_EOT:			/* ^D (End Of Transmission)	*/
			case A_CAN:			/* ^X (CANcel)	*/
			case A_ESC:			/* Escape		*/
				goto endwhile;	/* always stop the menu ...	*/
		}
		/*
		 *	Otherwise let the menu-defined keys take precedence and
		 *	potentially override function-key default-actions ...
		 */
		for (pm = menu; pm->key && pm->fct; pm++)
			if (key == pm->key || upkey == pm->key)
			{	key = 0;
				if ((*pm->fct)(menu) == -1)
					goto endwhile;
				break;
			}

		switch ( key )
		{	case KEY_HELP: case KEY_SHELP:
			case '?':
				pfmMnuHelp(menu);
				continue;

			case KEY_REFRESH:	/* ^L (Form Feed) is commonly */
			case '\f':			/* used for screen refresh... */
			#if '\f' != 'L'-'@'
			case 'L'-'@':
			#endif
				pfmMnuRefresh(menu);
				continue;

			case KEY_RESTART:
			case A_SYN:			/* ^V (SYNchronize)	*/
				pfmReinitscreen();
				pfmTermRaiseEsc(3);
				goto endwhile;	/* always stop the menu ...	*/

			case KEY_DOWN:		/* The four arrow keys ... */
			case 'j': case '+': case '\r': case '\n':
				pfmCwdWin1lineDown();
				continue;
			case KEY_UP:
			case 'k': case '-': case '^':
				pfmCwdWin1lineUp();
				continue;

			case_HOME_KEYS:		/* Home key and equivalents */
			case 'h':
				pfmCwdWinTopscroll();
				continue;
			case_END_KEYS:		/* end key and equivalents */
			case 'l':
				pfmCwdWinEndscroll();
				continue;

			case KEY_NPAGE:		/* Next page */
			case KEY_C3:		/* lower right of keypad */
			case A_ACK:			/* ^F */
				pfmCwdWinPageDown();
				continue;
			case KEY_PPAGE:		/* Previous page */
			case KEY_A3:		/* upper right of keypad */
			case A_STX:			/* ^B */
				pfmCwdWinPageUp();
				continue;

			case KEY_LEFT: case '<':
				pfmRShiftMenu(menu);
				pfmTermRedrawheader();
				continue;
			case KEY_RIGHT: case '>':
				pfmLShiftMenu(menu);
				pfmTermRedrawheader();
				continue;

			default:
				if (KEY_F(0) <= key && key < KEY_F(11))
				{	key -= KEY_F(1);
					if (key < 0) key = 9;
					(*fnkeymenu[key].fct)(0);
					continue;
				}
				if ('0' <= key && key <= '9')
				{	key -= '1';
					if (key < 0) key = 9;
					(*fnkeymenu[key].fct)(0);
					continue;
				}

		}
	}
	endwhile:
	pfm_Currentfnkeys = pfnk;
	pfm_Currentmenu = pmnu;
	pfmTermRedrawheader();
	return 0;

}


/*
 *	Setting up pfm
 */
__DF(
static void pfmInit,(conf), const char *conf )
{
	char	buf[80], *p;
	int	i;

	if ( (p = getenv("HOME")) )
	{	p = *p ? strdup(p):0;
		if (p) pfmhome = p;
	}
	pfm_Shellprompt = getuid() ? "$":"#";

	p = getdomainname(buf,sizeof(buf)-1) ? 0:strdup(buf);
	pfm_Domainname = (p ? p:"");

	p = gethostname(buf,sizeof(buf)-1) ? 0:strdup(buf);
	pfm_Hostname = (p ? p:"");

	p = strdup( pfmGetgroupname(getgid()) );
	pfm_Grpname = (p ? p:"");

	p = strdup( pfmGetgroupname(getegid()) );
	pfm_eGrpname = (p ? p:"");

	p = strdup( pfmGetusername(getuid()) );
	pfm_Usrname = (p ? p:"");

	p = strdup( pfmGetusername(geteuid()) );
	pfm_eUsrname = (p ? p:"");
	/*
	if ( !access("/usr/games/fortune",AEXEC) )
		pfmExecshell( "/usr/games/fortune", EXECSH_WAIT );
	*/
	pfmSystemEx( (char*)0, 0 ); /* find shell early */
	/*
	 *	try to read configuration and write the current configuration
	 */
	if ( pfmConfRead(conf) )
	{	fputs( pfm_C_hit_enter, stderr );
		getchar();
	}
	pfmConfWrite(conf);
	/*
	 *	 this is the right place to prepare for exceptions etc.
	 */
	#ifdef SIGABRT
		signal( SIGABRT, pfmSignalHandler );
	#endif
	signal( SIGALRM, pfmSignalHandler );
	#ifdef SIGFPE
		signal( SIGFPE, pfmSignalHandler );
	#endif
	signal( SIGHUP, pfmSignalHandler );
	#ifdef SIGILL
		signal( SIGILL, pfmSignalHandler );
	#endif
	signal( SIGINT, pfmSignalHandler );
	signal( SIGQUIT, pfmSignalHandler );
	signal( SIGSEGV, pfmSignalHandler );
	signal( SIGTERM, pfmSignalHandler );
	#ifdef SIGWINCH
		signal( SIGWINCH, pfmSignalHandler );
	#endif
	signal( SIGPIPE, pfmSignalHandler );
	/*
	 *	now we are prepared to start curses.
	 */
	if ( (i=pfmTermInitScr()) )
	{	fprintf( stderr, "%s\n%d, ", strerror(errno), i );
		pfmExit( 1, "Cannot init screen !", (char*)0 );
	}
	pfmCwdChangenamedir("././");	/* "." would be ignored	*/
}


__DF(
static void pfmSignalHandler,( sig ), int sig )
{
	switch ( sig )
	{	char msg[40];
		case SIGPIPE:
			signal( SIGPIPE, pfmSignalHandler );
			break;
		#ifdef SIGWINCH
		case SIGWINCH:
			signal( SIGWINCH, pfmSignalHandler );
			pfmTermRaiseSyn();
			break;
		#endif
		/*
		 * always ask to confirm the following signals
		 */
		case SIGALRM:
		case SIGHUP:
		case SIGINT:
		case SIGQUIT:
			signal( sig, pfmSignalHandler );
			if ( childprocess_is_active ) break;
			sprintf( msg, "Signal %d received, exit [%c/%c] ?",
						sig, *pfm_C_no_nein_non, *pfm_C_yes_ja_oui );
			if ( !pfmTermYesNo(msg,FALSE) ) break;
		/*
		 * catch the following signals to enable cleanup
		 */
		#ifdef SIGABRT
		case SIGABRT:
		#endif
		#ifdef SIGFPE
		case SIGFPE:
		#endif
		#ifdef SIGILL
		case SIGILL:
		#endif
		case SIGSEGV:
		case SIGTERM:
			sprintf( msg, "pfm: exit on signal %d", sig );
			pfmExit( 128+sig, msg, (char*)0 );
	}
}

/*
 *	Cleanup and exit
 */
__DF(
void pfmExit,( excode, msg /* , ..., (const char*)0 */ ),
int        excode __j__
const char *msg
VARARGS )
{
	va_list	args;
	pfmTermEndWin();
	if (*tmpfilename) unlink(tmpfilename);
	va_start( args, msg );
	while ( msg )
	{	fputs( msg, stderr );
		msg = va_arg(args,const char*);
	}
	fputc( '\n', stderr );
	va_end(args);
	exit(excode);
}

__DF(
static void pfmMnuExit,(m), x_menubar *m )
{
	pfmExit( 0, (char*)0 );
	(void)m;
}


/********************************************************************
*	main menu
********************************************************************/
/*
 *	Help for main menu
 */
__DF(
static int pfmMnuHelp,(m), x_menubar *m )
{
	return pfmHelp((char*)0);
}


/*
 *	Extra-, Options-, Quit- menue
 */
__DF(
int pfmMnuExtra,(m), x_menubar *m )
{
	static x_menubar	extramenu[] =
	{
		{ 'A', "c"BOLD"Ancel ", pfmMnuCancelmenu },
		{ 'B', ""/*"Back"*/, pfmMnuCancelmenu },
		{ 'H', BOLD"Help ", pfmMnuHelp },
		{ '?', "", pfmMnuHelp },
		{ 'C', BOLD"Config ",  pfmMnuConfigure },
		{ 'P', BOLD"Print ", pfmMnuFilePrintfile },
		{ 'E', BOLD"Edit ", pfmMnuFileEditdeffile },
		{ 'S', BOLD"Show ", pfmMnuFileViewdeffile },
		{ 'M', BOLD"Mkdir ", pfmMnuFileMkdir },
		{ 'R', BOLD"Rmdir ", pfmMnuFileRmdir },
		{ 'T', BOLD"Tree ", pfmMnuFileTree },
		/* we should expect '\n' == ^J, '\r' == ^M
		{ 'J'-'@',   "", pfmMnuEnterkey },
		{ 'M'-'@',   "", pfmMnuEnterkey },
		*/
		{ '\n',      "", pfmMnuEnterkey },
		{ '\r',      "", pfmMnuEnterkey },
		#ifdef KEY_ENTER
		{ KEY_ENTER, "", pfmMnuEnterkey },
		#endif
		{ 0, "", pfmMnuNop }
	};
	static char	reentry = 0;

	if ( !reentry )
	{	reentry = 1;
		pfmMnuRunmenu( extramenu, (char*)0 );
		reentry = 0;
	}
	(void)m;
	return 0;
}

__DF(
int pfmMnuQuit,(m), x_menubar *m )
{
#if 1
	static char yes[] = BOLD;
	static x_menubar	quitmenu[] =
	{
		{ 'Q',
		#ifdef DEUTSCH
			"pfm wirklich beenden ? ["
		#else
			"Really quit pfm ? ["
		#endif
			BOLD, pfmMnuNop },
		{  -1, pfm_C_yes_ja_oui, pfmMnuNop },
		{ 'Q', "/"BOLD"Quit/", pfmMnuNop },
		{ 'X', "e"BOLD"Xit]", pfmMnuNop },
		{ 0, "", pfmMnuNop }
	};
	int key;

	quitmenu[0].key = *pfm_C_yes_ja_oui;
	__pfmRedrawHeader(quitmenu);
	refresh();
	switch ( (key = pfmTermGetch()) )
	{	default:
			if (toupper(key) != toupper(*pfm_C_yes_ja_oui))
				break;
		case 'q': case 'Q':
		case 'x': case 'X':
			pfmExit( 0, (char*)0 );
	}
	pfmTermRedrawheader();
#else
	if ( pfmTermYesNo(quit_txt,FALSE) )
		pfmExit( 0, (char*)0 );
#endif
	(void)m;
	return 0;
}

/*
 *	Refresh pfm-screen
 */
__DF(
static int pfmMnuRefresh,(m), x_menubar *m )
{
	pfmTermClearscreen();
	pfmTermRefresh(TRUE);
	(void)m;
	return -1;
}

static void pfmReinitscreen ()
{
	pfmTermEndWin();
	pfmTermInitScr();
	pfmTermRedraw();
}

#if 0 /* obsolete ? */
__DF(
static int __pfmMnuClearscreen,(m), x_menubar *m )
{
	pfmTermClearscreen();
	(void)m;
	return 0;
}

__DF(
static int pfmMnuReinitscreen,(m), x_menubar *m )
{
	pfmReinitscreen();
	(void)m;
	return -1;
}

__DF(
static int pfmMnuReread,(m), x_menubar *m )
{
	pfmCwdReread();
	(void)m;
	return -1;
}

__DF(
int pfmMnuUpdatemenu,(m), x_menubar *m )
{
	static x_menubar	updatemenu[] =
	{
		{ 'A', "c"BOLD"Ancel, ", pfmMnuCancelmenu },
		/*
		{ 'E', BOLD"Erase, ", __pfmMnuClearscreen },
		*/
		{ 'F', "re"BOLD"Fresh, ", pfmMnuRefresh },
		{ 'I', "re"BOLD"Init, ", pfmMnuReinitscreen },
		{ 'R', "re"BOLD"Read directory, ", pfmMnuReread },
		{ 0, "", pfmMnuNop }
	};
	static char	reentry = 0;

	if ( !reentry )
	{	reentry = 1;
		pfmMnuRunmenu( updatemenu, (char*)0 );
		reentry = 0;
	}
	(void)m;
	return 0;
}
#endif /* obsolete ? */


/*
 *	mark the file at the cursor position
 */
__DF(
int pfmMnuMarkcurrent,(m), x_menubar *m )
{
	pfmCwdMarkcurrentfile();
	pfmCwdWin1lineDown();
	(void)m;
	return 0;
}


/********************************************************************
*	entry
********************************************************************/

__DF(
int  main,( argc, argv ), int argc __j__ char *argv[] )
{
	static x_menubar  mainmenu[] =
	{
		{ 'H', BOLD"Help ", pfmMnuHelp },
		{ '?', "", pfmMnuHelp },
		{ 'C', BOLD"Copy ", pfmMnuFileCopyfile },
		{ 'L', BOLD"Link ", pfmMnuFileLinkfile },
		#if HAS_SYMLINK
		{ 'Y', "s"BOLD"Ymlink ", pfmMnuFileSymlink },
		#endif
		{ 'I', BOLD"Info ", pfmMnuFileInfo },
		{ '=', "", pfmMnuFileInfo },
		{ 'M', BOLD"Modify ", pfmMnuFileModifyatt },
		{ 'D', BOLD"Delete ", pfmMnuFileRemovefile },
		{ 'R', BOLD"Rename ", pfmMnuFileRenamefile },
		{ 'E', BOLD"Edit ", pfmMnuFileEditfile },
		{ 'S', BOLD"Show ", pfmMnuFileViewfile },
		{ 'S'-'@', "", pfmMnuFileViewanyfile },
		{ 'O', "c"BOLD"Ommand ", pfmMnuCommand },
		{ 'Q', BOLD"Quit ", pfmMnuQuit },
		{ 'X', "e"BOLD"Xtra ", pfmMnuExtra },
		{ 'U', BOLD"User ", pfmMnuNop },
		{ '^', "", (__DC(int(*),(x_menubar*)))pfmCwdWinTopscroll },
		{ 'B', "", (__DC(int(*),(x_menubar*)))pfmCwdWinEndscroll },
		/* we should expect '\n' == ^J, '\r' == ^M
		{ 'J'-'@',   "", pfmMnuEnterkey },
		{ 'M'-'@',   "", pfmMnuEnterkey },
		*/
		{ '\n',      "", pfmMnuEnterkey },
		{ '\r',      "", pfmMnuEnterkey },
		#ifdef KEY_ENTER
		{ KEY_ENTER, "", pfmMnuEnterkey },
		#endif
		{ 0, "", pfmMnuNop }
	};
	char	*conf = 0;
	int 	i;
	#if defined(__COHERENT__) && DEBUG
		extern int	end[];
		printf("main, entry:\ndata end = 0x%04x (%u)\n"
				"stack adr = 0x%04x (%u) <enter>\n",
				end, end, (int(**)())&argc-1, (int(**)())&argc-1 );
		while (getchar() != '\n');
	#endif
	pfm_argv0 = argv[0];
	fputs( pfmversion, stdout );
	/*
	 *	look for commandline-defined initilization
	 */
	for (i = 1; i < argc; ++i)
	{	if ( !strcmp(argv[i],"-t") )
		{	pfmTermTestkeypad(pfmversion);
			exit(0);
		}
		if (!strcmp(argv[i],"-C") && argv[++i])
			conf = argv[i];
		else
		{	fprintf( stderr,
				"usage: %s [-C configfile]\n"
				"\t" "start pfm, optional with alternate configfile\n"
				"usage: %s -t\n"
				"\t" "test keypad, termcap/terminfo translations.\n",
				pfm_argv0, pfm_argv0 );
			exit(0);
		}
	}
	pfmInit(conf);

	while (1)
	{
		pfmMnuRunmenu( mainmenu, mainfnkey );
		/* pfmCwdPreviousdir((char*)0); *//* besser nur mit F9 */
	}
}

/***** eof *********************************************************/
