/***********************************************************************
* Copyright (c) 1995-2003 by peter fuerst
*
*	This program is free software; you can redistribute it and/or modify
*	it - as explained in the accompaning file COPYING - under the terms
*	of the GNU General Public License as published by the Free Software
*	Foundation; either version 2, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	Author: Peter Frst, post@pfrst.de
*
*	@(#)pfm.h	10/25/95
*
*	Currently running on
*	Coherent (3.2.1, 4.2.05), Linux, FreeBSD, AIX, IRIX 6.2
*
***********************************************//*$preferred tabstop=4*/

#ifndef __PFM_H__
#define __PFM_H__ 1

#include <dirent.h>
#include <unistd.h>
#include <ar.h>
#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>	/* WIFEXITED,... */

/***** OS-specific preliminaries ***********************************/

#include "ckunix.h"

#ifdef __linux__
 typedef typeof(off_t) fsize_t;
# include <sys/sysmacros.h>	/* major, minor, makedev */
# include <fnmatch.h>
# define __MAY_HAVE_FSIZE_64__	1
#endif
#if defined(__FreeBSD__) || defined(__NetBSD__)
 typedef typeof(off_t) fsize_t;
# include <fnmatch.h>
#endif
#ifdef __AIX__
 typedef long fsize_t;	/* off_t ? in sys/types.h ? */
# include <fnmatch.h>
#endif
#ifdef __COHERENT__
# include <sys/filsys.h>
#endif
#ifdef __IRIX__
# include <cmplrs/host.h>	/* fsize_t */
# include <sys/sysmacros.h>	/* major, minor, makedev */
# include <fnmatch.h>
#endif

#ifndef __COHERENT__
# define shellsort qsort /* hmmm, hope we don't run out of stack ;-) */
#endif

#ifdef __UNIX_OS__
# define O_BINARY 0
#endif
#define tell(fildes) lseek(fildes,(off_t)0,SEEK_CUR)

#ifdef __COHERENT_286__
 typedef long off_t;
 long lseek();
# ifndef F_OK
#  define F_OK	AEXISTS 
#  define R_OK	AREAD
#  define W_OK	AWRITE
#  define X_OK	AEXEC
# endif
#endif
#ifndef F_OK
/* to do... */
#endif

#ifndef va_copy
# ifdef __WATCOMC__
#  define va_copy(dest,src) (*(dest) = *(src))
# else
#  define va_copy(dest,src) ((dest) = (src))
# endif
#endif

#ifndef __COHERENT_286__
# define SUPPORT_INET_URL
#endif


/***** find curses version *****************************************/

#include "ckcurses.h"

#ifdef NCURSES_MOUSE_VERSION
# ifdef IGNORE_MOUSE_EVENTS
#  undef NCURSES_MOUSE_VERSION
# endif
#endif
#if defined(__linux__) || defined(__XSI_CURSES__)
/* In some improved NCURSES implementations (at least those from
 * Linux-distributions with 2.0.x kernels) there is a new braindamaged
 * feature:
 * 'wrefresh(curscr)' does indeed clear the physical screen, but
 * immediately afterwards updates it with the contents of 'stdscr' :-(
 * (seems to happen also on AIX with _XOPEN_CURSES)
 */
# define BRAINDEAD_WCLEAR_CURSCR	1
# ifndef __AIX__ /* need tgetstr() */
#  ifdef __linux__
#   define TGETSTR_PROTOTYPE_H <termcap.h>
#  else
#   define TGETSTR_PROTOTYPE_H <term.h>
#  endif
# endif
/*
 * In some NCURSES implementations we must use stty to enable
 * translation of linefed to carriage_return-linefeed.
 */
#endif

/*
 * KEY_HOME: Home key (upward+left arrow)
 * KEY_A1:   upper left of keypad
 * KEY_LL:   home down or bottom (lower left)
 * KEY_END:  end key
 * KEY_C1:   lower left of keypad
 * we take one more chance by allowing shifted home- & end-keys...
 */
#ifdef __linux__
/* we should not ignore, that rxvt & the linux-console usually send the same
 * sequences for home-/end-key, that xterm sends for find-/select-key...
 */
# define case_HOME_KEYS case KEY_HOME: case KEY_A1: case KEY_SHOME: case KEY_FIND
# define case_END_KEYS case KEY_END:case KEY_LL:case KEY_C1:case KEY_SEND:case KEY_SELECT
#else
# define case_HOME_KEYS	case KEY_HOME: case KEY_A1: case KEY_SHOME
# define case_END_KEYS	case KEY_END: case KEY_LL: case KEY_C1: case KEY_SEND
#endif


/***** now the pfm-specific stuff... *******************************/

#ifdef __linux__
# define BLKSIZE(stat)		((stat).st_blksize)
#else
#endif

#ifndef BLKSIZE
# if defined(BSIZE)
#  define BLKSIZE(stat)	BSIZE
# elif defined(DEV_BSIZE)
#  define BLKSIZE(stat)	DEV_BSIZE
# elif defined(S_BLKSIZE)
#  define BLKSIZE(stat)	S_BLKSIZE
# else
#  define BLKSIZE(stat)	512U
# endif
#endif


typedef struct menubar	x_menubar;
typedef struct filedat	x_filedat;

struct menubar
{
	short	key;
	const char	*txt;
__DC(int		(*fct),(x_menubar*) );
};

struct _pfmstat	/* subset of 'struct stat' for display/sorting	*/
{
	ino_t			ino;
	unsigned short	mode;	/* file-mode, obtained by [f]stat()	*/
	unsigned short	lmode;	/* link-mode, obtained by lstat()	*/
	fsize_t			size;
	time_t			atime;
	time_t			mtime;
	time_t			ctime;
};

struct filedat
{
	struct _pfmstat	st;
	short			flag;
	char			nm[1];	/* the trailing zero ...	*/
};
#define	X_FILEDAT_MARK_FLAG	1	/* is marked ? */
#define	X_FILEDAT_RMND_FLAG	2	/* "old mark" ? (see multifiles) */
#define	X_FILEDAT_MLNK_FLAG	4	/* st_nlink > 1 ? */
#if defined(__COHERENT_286__) || defined(__I386__)
# define x_filedat_align(n)	(n)
#else
# define x_filedat_align(n)	(((n)+sizeof(long)-1) & ~(sizeof(long)-1))
#endif


#define	pfm_fwildc_start	'%'
#define	pfm_Fname_fwildc	'F'
#define	pfm_Nosuff_fwildc	'N'
#define	pfm_Suffix_fwildc	'X'
#define	pfm_CWD_fwildc		'D'
#define	pfm_AltD_fwildc	'A'
#define	pfm_HOME_fwildc	'H'

#define	arraysize(x)	(sizeof(x)/sizeof(x[0]))

#define	pfmTermLastline()	(pfm_lines - 1)
#define	pfmTermGetch()		pfmTermWgetch(stdscr)
/*
 * The current working directory's path's length may be up to
 * 512 on 16bit systems, up to 1024 on 32bit systems.
 */
#define CWD_MAX_LNG		(sizeof(int*)*sizeof(short)*128)

#define EXECSH_WAIT  	1 /* postExecshell: wait for keystroke after cmd */
#define EXECSH_NORRD	2 /* postExecshell: don't re-read after cmd */

#ifdef __cplusplus
 extern "C" {
#endif

/***** prototypes **************************************************/

__DC( int pfmMnuNop,(x_menubar*) );
__DC( int pfmMnuCancelmenu,(x_menubar*) );
__DC( int pfmMnuQuit,(x_menubar*) );
__DC( int pfmMnuEnterkey,(x_menubar*) );
__DC( const char *pfmSGetLastcommand,(const char*) );
__DC( void pfmPreExecshell,(int) );
__DC( void pfmPostExecshell,(int) );
__DC( int pfmExecshell,(const char*,int) );
__DC( int pfmExecommand,(const char*,int) );
__DC( int pfmExecargv,(char* const[],int) );
__DC( int pfmSystem,(const char*) );
__DC( int pfmMnuRunmenu,(x_menubar*,const char*) );
__DC( const char *pfmStrtail,(const char*,size_t) );
__DC( int pfmStrcasematch,(const char*,const char*) );
__DC( int pfmStrmatch,(const char*,const char*) );
__DC( char *pfmStrdup,(const char*,char*) );
__DC( char *pfmGetusername,(int uid) );
__DC( char *pfmGetgroupname,(int gid) );
__DC( int  pfmVExpandsize,(const char*,const char*,const char*,const char*) );
__DC( int  pfmExpandsize,(const char*) );
__DC( char *pfmVExpandargs,(char[/*size*/],int size,
                            const char*,const char*,const char*) );
__DC( char *pfmExpandargs,(char[/*size*/],int size) );
__DC( void pfmExpandargv,(char*[],char[/*size*/],int size) );
__DC( int pfmMatchname,(const char*,const char*) );
__DC( int pfmHelp,(const char*) );

__DC( int  pfmTermInitScr,(void) );
__DC( void pfmTermEndWin,(void) );
__DC( WINDOW *pfmTermAdjParentCursor,(WINDOW *sub) );
__DC( int pfmTermSubwinRefresh,(WINDOW *sub,bool touch) );
__DC( void pfmTermRedrawheader,(void) );
__DC( void __pfmRedrawHeader,(const x_menubar*) );
__DC( int  pfmTermRedraw,(void) );
__DC( void pfmTermClearscreen,(void) );
__DC( void pfmTermRefresh,(bool clear) );
__DC( int  pfmTermFixmode,(void) );
__DC( int  pfmTermResetty,(void) );
__DC( int  pfmTermGetstrv,(const char*,int,char[],va_list) );
/* (const char*,int,char[],const char*,...,(const char*)0) */
__DC( int  pfmTermGetstrl,(const char*,int,char[],...) );
/* (const char*,const char*,...,(const char*)0) */
__DC( void pfmTermErrormessage,(const char*,...) );
__DC( int  pfmTermYesNo,(const char*,bool) );
__DC( void __pfmRedrawFnkeyline,(const char*) );
__DC( int  pfmTermWgetch,(WINDOW*) );
__DC( int  pfmTermClrcon,(void) );
__DC( void pfmTermRaiseEsc,(int) );
__DC( void pfmTermRaiseSyn,(void) );
__DC( void pfmTermTestkeypad,(const char*) );

__DC( int  pfmMnuFileInfo,(x_menubar*) );
__DC( int  pfmMnuFileEditfile,(x_menubar*) );
__DC( int  pfmMnuFileEditdeffile,(x_menubar*) );
__DC( int  pfmMnuFileViewfile,(x_menubar*) );
__DC( int  pfmMnuFileViewanyfile,(x_menubar*) );
__DC( int  pfmMnuFileViewdeffile,(x_menubar*) );
__DC( int  pfmMnuFileCopyfile,(x_menubar*) );
__DC( int  pfmMnuFileRenamefile,(x_menubar*) );
__DC( int  pfmMnuFileLinkfile,(x_menubar*) );
__DC( int  pfmMnuFileSymlink,(x_menubar*) );
__DC( int  pfmMnuFileRemovefile,(x_menubar*) );
__DC( int  pfmMnuFileModifyatt,(x_menubar*) );
__DC( int  pfmMnuFileMkdir,(x_menubar*) );
__DC( int  pfmMnuFileRmdir,(x_menubar*) );
__DC( int  pfmMnuFileTree,(x_menubar*) );
__DC( int  pfmMnuFilePrintfile,(x_menubar*) );
__DC( char *pfmFileDecodemode,(int,char[12]) );
__DC( int  pfmFileEncodemode,(const char[12],int*,int) );
__DC( const char *pfmFileGetModepattern,(const char*, const char*) );
__DC( const char *pfmFileGetWildnamepattern,(const char*, const char*) );
struct tm; /* forward declaration to suppress warnings */
__DC( time_t    pfmFileGetDateandtime,(const char*, struct tm*) );
__DC( int  pfmScandateandtime,(char*,struct tm*,time_t*) );
__DC( int  pfmFileGetnewmode,(int,int*) );
struct stat; /* forward declaration to suppress warnings */
__DC( int  pfmFileStat,(const char*, struct stat*, struct _pfmstat*) );
__DC( int  pfmFileCopyfiles,(const char*,const char*) );
__DC( int  pfmFileRenamefile ,(const char*,const char*) );
__DC( int  pfmFileRemovefile,(x_filedat*) );
__DC( int  pfmFileMkdir,(const char*) );
__DC( int  pfmFileMultiplefile,(void) );
__DC( const char *pfmFileSGetLastdest,(const char*) );
__DC( int  pfmFileEditfile,(const char*,int,const char*) );
__DC( char *pfmSetCopyWildname,(const char*,char[],int) );
__DC( const char *pfmFileBasename,(const char*) );
__DC( int pfmFileCopydir,(const char*,const char*) );

__DC( const char *pfmCwdAltDir,(void) );
__DC( const char *pfmCwdCWD,(void) );
__DC( int  pfmCwdChangedir,(void) );
__DC( int  pfmCwdChangenamedir,(const char*) );
__DC( int  pfmCwdSortdir,(void) );
__DC( int  pfmCwdPreviousdir,(void) );
__DC( int  pfmCwdReread,(void) );
__DC( int  pfmCwdRereadChanged,(void) );
__DC( int  pfmCwdWinDrawfileswin,(bool) );
__DC( int  pfmCwdWinDrawnoteswin,(bool) );
__DC( int  pfmCwdWin1lineUp,(void) );
__DC( int  pfmCwdWin1lineDown,(void) );
__DC( int  pfmCwdWinPageUp,(void) );
__DC( int  pfmCwdWinPageDown,(void) );
__DC( int  pfmCwdWinTopscroll,(void) );
__DC( int  pfmCwdWinEndscroll,(void) );
__DC( void pfmCwdWinCursorhome,(void) );
__DC( x_filedat *pfmCwdCurrentfile,(void) );
__DC( x_filedat *pfmCwdMarkcurrentfile,(void) );
__DC( x_filedat *pfmCwdDirtblentry,(unsigned) );
#define   pfmCwdGetMarks(dp,bits) ((dp)->flag & (short)(bits))
#define   pfmCwdIsMarked(dp,bits) (pfmCwdGetMarks(dp,bits) == (short)(bits))
__DC( int  pfmCwdMarkon,(x_filedat*,unsigned) );
__DC( int  pfmCwdMarkoff,(x_filedat*,unsigned) );
__DC( int  pfmMnuMarkcurrent,(x_menubar*) );
__DC( void pfmCwdWinJumptoline,(int,int) );
__DC( int  pfmTermRegfctmousemap,(WINDOW*,void(*)(),int,int,int,int) );
__DC( int  pfmTermRegkeymousemap,(WINDOW*,int,int,int,int,int) );
__DC( void pfmTermUnregmousemaps,(WINDOW*) );

__DC( int  pfmFileGotoURL,(const char*,bool) );
__DC( int  pfmViewFileviewer,(const char*) );

__DC( void pfmExit,(int,const char*,.../*,(const char*)0*/) );

__DC( bool pfmConfShowdotfiles,(void) );
__DC( const char *pfmConfDocdir,(void) );
__DC( const char *pfmConfPagerPrg,(void) );
__DC( const char *pfmConfPagerCmd,(void) );
__DC( const char *pfmConfEditorCmd,(void) );
__DC( const char *pfmConfViewerCmd,(void) );
__DC( const char *pfmConfMailerCmd,(void) );
__DC( const char *pfmConfPrintCmd,(void) );
__DC( const char *pfmConfCpdirCmd,(void) );
#ifdef SUPPORT_INET_URL
__DC( const char *pfmConfUrlHdlrCmd,(void) );
__DC( const char *pfmConfWwwCmd,(void) );
__DC( const char *pfmConfFtpCmd,(void) );
__DC( const char *pfmConfTelnetCmd,(void) );
#endif
__DC( bool pfmConfMouseinput,(void) );
__DC( int  pfmConfRead,(const char*) );
__DC( int  pfmConfWrite,(const char*) );
__DC( int  pfmMnuConfigure,(x_menubar*) );
#define BUILTIN_CMD_PFX "*BUILTIN*"

/***** global ******************************************************/

extern const char	pfm_C_confirm[];
extern const char	pfm_C_hit_any_key[];
extern const char   pfm_C_hit_enter[];
extern const char	pfm_C_dir_pathname[];
extern const char	pfm_C_pathname[];
extern const char	pfm_C_file_name[];
extern const char	pfm_C_wild_name[];
extern const char	pfm_C_new_name[];
extern const char	pfm_C_destination[];
extern const char	pfm_C_date_and_time[];
extern const char	pfm_C_no_nein_non[];
extern const char	pfm_C_yes_ja_oui[];
extern const char	pfm_C_out_of_stack[];
extern const char	*pfm_Shellprompt;
extern const char	*pfm_Domainname;
extern const char	*pfm_Hostname;
extern const char	*pfm_Grpname;
extern const char	*pfm_eGrpname;
extern const char	*pfm_Usrname;
extern const char	*pfm_eUsrname;

extern WINDOW	*pfm_CwdirWindow;
extern WINDOW	*pfm_FilesWindow;
extern WINDOW	*pfm_NotesWindow;
extern short    pfm_lines;


#ifdef __cplusplus
 } // extern "C"
#endif

#ifndef NDEBUG

#undef assert
#define assert(test)\
	if (!(test)){\
		char num[10];\
		sprintf( num, ": %d:", __LINE__);\
		pfmExit( 1, __FILE__, num, " assert(", #test, ") failed !", (char*)0 );\
	}
#endif

#endif /* __PFM_H__ */

/***** eof "pfm.h" *************************************************/
