/***********************************************************************
* Copyright (c) 1995-2003 by peter fuerst
*
*	This program is free software; you can redistribute it and/or modify
*	it - as explained in the accompaning file COPYING - under the terms
*	of the GNU General Public License as published by the Free Software
*	Foundation; either version 2, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	Author: Peter Frst, post@pfrst.de
*
*	termio.c	10/25/95
*
***********************************************//*$preferred tabstop=4*/

#include <stdio.h>
#include <stddef.h>
#include <dirent.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/stat.h>


/***** define & typedef ********************************************/

#include "pfm.h"
#ifdef BRAINDEAD_WCLEAR_CURSCR
# ifdef TGETSTR_PROTOTYPE_H
#  include TGETSTR_PROTOTYPE_H
# endif
# include "spawn.h"
#endif

#undef CLR_GETSTR_ON_1ST_DEL /* compile time option for pfmTermGetstr() */


/***** global ******************************************************/

WINDOW	*pfm_CwdirWindow = 0;
WINDOW	*pfm_FilesWindow = 0;
WINDOW	*pfm_NotesWindow = 0;

/***** local *******************************************************/

#define EMPHS "\007" /* use BELl to emphasize the next character */
#define EMPHC 7
static char  filefnkey[] =
		"F " EMPHS"1-no_suffix " EMPHS"2-full_fname " EMPHS"3-suffix "
		EMPHS"4-alternate_dir " EMPHS"5-cwd " EMPHS"6-home";

static char  curses_is_open = 0;
short pfm_lines = 24;


/********************************************************************
*	Initalising and closing terminal IO ...
********************************************************************/
/*
 *	Setting up the terminal for pfm
 */
#ifdef NCURSES_MOUSE_VERSION

int pfmTermFixmouse ()
{
	static int mouse_is_masked = 0;

	pfmTermUnregmousemaps(pfm_FilesWindow);
	pfmTermUnregmousemaps(stdscr);

	if ( pfmConfMouseinput() )
	{	mousemask(
			BUTTON1_CLICKED | BUTTON1_DOUBLE_CLICKED  |
			BUTTON2_CLICKED | BUTTON2_DOUBLE_CLICKED  |
			BUTTON3_CLICKED | BUTTON3_DOUBLE_CLICKED  |
			BUTTON_SHIFT | BUTTON_CTRL | BUTTON_ALT |
			REPORT_MOUSE_POSITION,
			(mmask_t*)0 );
		mouse_is_masked |= (1|2); /* enabled | mousemask called */

		pfmTermRegfctmousemap( pfm_FilesWindow, pfmCwdWinJumptoline,
		                       0, 0, wwidth(pfm_FilesWindow)-1,
		                       wheight(pfm_FilesWindow)-1);
	}
	/*
	 *	Avoid to touch the mouse interface, if we didn't explicitely
	 *	enable it before ! (anyway it may already be disabled by
	 *	resetty(), but we can't find out)
	 */
	else if (mouse_is_masked & 1)
	{	mousemask( 0, (mmask_t*)0 );
		mouse_is_masked = 2; /* 0 | mousemask called */
	}
	return 0;
}
#endif

int pfmTermFixmode ()
{
	extern void pfmTermSetnodelay();
	#ifdef __XSI_CURSES__
		reset_prog_mode(); /* fix, if we had to use 'reset_shell_mode()' */
	#endif
	pfmTermSetnodelay();
	noecho();
	cbreak();
	raw();
	#ifdef NCURSES_MOUSE_VERSION
		pfmTermFixmouse();
	#endif
	curses_is_open = 1;
	return 0;
}

/*
 *	Resetting the terminal
 */
int pfmTermResetty ()
{
	noraw();
	echo();
	resetty();
	#ifdef __XSI_CURSES__
		reset_shell_mode(); /* sometimes 'resetty()' will not do... */
	#endif
	#ifdef RESETTY_NEEDS_STTY_SANE
	{	static char *stty[3] = {"/bin/stty","sane",0};
		spawnv(P_WAIT,stty[0],stty);
	}
	#endif
	curses_is_open = 0;
	return 0;
}

/*
 *	Setting up the terminal
 *
 *	If ye use ncurses, make sure, that the terminfo description
 *	for 'xterm' (used by xterm & rxvt) does *NOT* hold entries
 *	for row_address (vpa) and column_adress (hpa). Force ncurses
 *	to use cursor_adress (cup) only !
 */
int  pfmTermInitScr ()
{
	extern void pfmTermBuildkeyent();
	int	x, ww;

	errno = 0;
	initscr();
	if ( !stdscr ) return -1;
	savetty();
	clear();

	pfm_lines = wheight(stdscr);
	#if __CORNELL_CURSES__	/* libcurses bug	*/
		pfm_lines--;
	#endif
	ww = wwidth(stdscr);
	x = ww < 80 ? ww/4:20;
	pfm_CwdirWindow = subwin( stdscr, 1, ww-1, 1, 0 );
	if ( !pfm_CwdirWindow ) return -2;
	pfm_FilesWindow = subwin( stdscr, pfm_lines-3, ww-x, 2, 0 );
	if ( !pfm_FilesWindow ) return -3;
	pfm_NotesWindow = subwin( stdscr, pfm_lines-3, x-1,  2, ww-x );
	if ( !pfm_NotesWindow ) return -4;

	leaveok( curscr, FALSE );
	leaveok( stdscr, FALSE );
	leaveok( pfm_CwdirWindow, FALSE );
	leaveok( pfm_FilesWindow, FALSE );
	leaveok( pfm_NotesWindow, FALSE );
	pfmTermBuildkeyent();
	pfmTermFixmode();
	move( 0, 0 );
	return 0;
}

/*
 *	Cleanup and exit
 */
void  pfmTermEndWin ()
{
	if ( stdscr )	/* this should also imply curscr!=0	*/
	{
		#ifdef NCURSES_MOUSE_VERSION
			pfmTermUnregmousemaps(pfm_FilesWindow);
			pfmTermUnregmousemaps(stdscr);
		#endif
		wclear(stdscr);		/* 'curscr' would not do...	*/
		wrefresh(stdscr);	/* see pfmTermClearScreen	*/
		pfmTermResetty();
		endwin();
	}
}


/********************************************************************
*	Handle subwindows ...
********************************************************************/

__DF(
WINDOW *pfmTermParentWindow,(win), WINDOW *win )
{
	return parentwindow(win);
}

__DF(
WINDOW *pfmTermAdjParentCursor,(win), WINDOW *win )
{
	register WINDOW *pw = parentwindow(win);
	int	x, y;
	getyx( win, y, x );
	wmove( pw, y + getbegy(win)-getbegy(pw), x + getbegx(win)-getbegx(pw) );
	return pw;
}

/********************************************************************
*	Clear and refresh ...
********************************************************************/

__DF(
int pfmTermSubwinRefresh,( win, touch ), WINDOW *win __j__ bool touch )
{
	win = parentwindow(win);
	if (touch) touchwin(win);
	leaveok( win, FALSE );
	return wrefresh(win);
}

/*
 *	physically clear the screen without clearing stdscr
 */
void  pfmTermClearscreen ()
{
	#ifdef BRAINDEAD_WCLEAR_CURSCR
		char	buf[40], *p = buf;
		if ( !(p = tgetstr("cl",&p)) ) /* if all else fails...	*/
			spawnl(P_WAIT,"/usr/bin/clear","clear",NULL);
		else
		{	putp(p);
			fflush(stdout);
		}
	#else
		wclear(curscr);		/* erase 'curscr' ...	*/
		wrefresh(curscr);	/* ... and write it out	*/
	#endif
}

/*
 *	unconditionally refresh the whole screen
 */
__DF(
void pfmTermRefresh,( clear ), bool clear )
{
	#if 0
		if (clear) pfmTermClearscreen();
	#else
		if (clear) clearok( stdscr, TRUE );
	#endif
	touchwin(stdscr);	/* .. redraw the (whole) screen	*/
	pfmCwdWinCursorhome();
	refresh();
}

/*
 *	redraw the screen
 */
__DF(
void __pfmRedrawHeader,(m), const x_menubar *m )
{
	int	w = wwidth(stdscr)-1; /* don't use the last column */
	int	bold = 0;
	int	col, i;

	#ifdef NCURSES_MOUSE_VERSION
		pfmTermUnregmousemaps(stdscr);
	#endif
	move( 0, 0 );

	for (i=0, col=0; m[i].key && m[i].txt; i++)
	{
		const char	*p = m[i].txt;

		while ( *p )
		{	if ((unsigned char)*p != EMPHC)
			{	if (++col < w)
					addch(*p);
				else
				{	addch('>');
					goto end;
				}
				if ( bold )
				{	wattroff( stdscr, A_BOLD );
					#ifdef NCURSES_MOUSE_VERSION
						/* register this cell for mouse-events */
						pfmTermRegkeymousemap( stdscr, m[i].key,
						                       col-1, 0, col-1, 0 );
					#endif
				}
				bold = 0;
			}
			else
			{	wattron( stdscr, A_BOLD );
				bold = 1;
			}
			p++;
		}
	}
	end: wattroff( stdscr, A_BOLD );
	clrtoeol();
}

void  pfmTermRedrawheader ()
{
	extern const x_menubar	*pfm_Currentmenu;

	__pfmRedrawHeader(pfm_Currentmenu);
	pfmCwdWinCursorhome();
	refresh();
}

__DF(
void  __pfmRedrawFnkeyline,( kstr ), const char *kstr )
{
	int	w = wwidth(stdscr)-1; /* don't use the last column */
	int	bold = 0, col = 0;

	move( pfmTermLastline(), 0 );

	while ( *kstr )
	{	if ((unsigned char)*kstr != EMPHC)
		{	addch(*kstr);
			if (w <= ++col) goto end;
			if ( bold ) wattroff( stdscr, A_BOLD );
			bold = 0;
		}
		else
		{	wattron( stdscr, A_BOLD );
			bold = 1;
		}
		kstr++;
	}
	end: wattroff( stdscr, A_BOLD );
	clrtoeol();
	pfmCwdWinCursorhome();
}

__DF(
static void __pfmTermDrawcwdline,( rfrsh ), bool rfrsh )
{
	const char *cwd = pfmCwdCWD();

	mvwaddstr( pfm_CwdirWindow, 0, 0,
				pfmStrtail(cwd,wwidth(pfm_CwdirWindow)-2) );
	if ( cwd[1] ) waddch( pfm_CwdirWindow, '/' );
	wclrtoeol(pfm_CwdirWindow);
#	if 0 && __CORNELL_CURSES__	/* libcurses bug ...	*/
		if ( !cwd[1] )
			mvwaddch( pfm_CwdirWindow, 0, wwidth(pfm_CwdirWindow)-1, '.' );
#	endif
	if (rfrsh) wrefresh(pfm_CwdirWindow);
}

int  pfmTermRedraw ()
{
	extern const x_menubar	*pfm_Currentmenu;
	extern const char	*pfm_Currentfnkeys;

	__pfmRedrawHeader(pfm_Currentmenu);
	__pfmTermDrawcwdline(FALSE);
	__pfmRedrawFnkeyline(pfm_Currentfnkeys);
	pfmCwdWinDrawnoteswin(FALSE);
	pfmCwdWinDrawfileswin(FALSE);
	pfmTermRefresh(FALSE);
	return 0;
}


/********************************************************************
*	Reading string input ...
********************************************************************/
/*
 *	get a string from the console
 */
#define	GETSTR_ON_HEADER_LINE	1

typedef struct
{
	unsigned	offset;
	WINDOW  	*window;
}	x_getstr;


__DF(
static void	pfmTermGetstrwrite,( dlg, str, len, pos ),
const x_getstr *dlg __j__
const char *str __j__
int	 len __j__
int	 pos )
{
	if ( dlg->window )
	{	int	i, iww = wwidth(dlg->window) - 1; /* > 6 */

		assert(0 <= pos && pos <= len);
		if ((i = iww - dlg->offset + 1) > 8) i = 8;

		while (dlg->offset+pos > iww)
		{	pos -= i;
			len -= i;
			str += i;
		}
		/*	We don't use 'pfmIsprintExpandString()' here, since it would be
		 *	hard to keep track of 'pos' in case of non-printable characters.
		 */
		wmove( dlg->window, 0, dlg->offset );
		while (len-- > 0 && dlg->offset <= iww--)
		{	unsigned c = (unsigned char)*str++;
			unsigned k = toascii(c);
			if (isprint(k))
				k = c;
			else if (k < ' ')
				k += '@';
			else
				k = '?';

			if (c != k) wstandout(dlg->window);
			waddch( dlg->window, k );
			if (c != k) wstandend(dlg->window);
		}
		wclrtoeol(dlg->window);
		wmove( dlg->window, 0, dlg->offset+pos );
		touchwin(dlg->window);
		wrefresh(dlg->window);
	}
}


__DF(
static WINDOW *pfmTermGetstropen,( dlg, msg, size ),
x_getstr   *dlg __j__
const char *msg __j__
unsigned   size )
{
	const char *np;
#	if GETSTR_ON_HEADER_LINE
	if ( !(np = strchr(msg,'\n')) )
#	endif
		np = strchr( msg, 0 );

	dlg->window =
#		if GETSTR_ON_HEADER_LINE
			newwin( *np ? 2:1, wwidth(stdscr)-1, 0, 0 );
#		else
			newwin( 1, wwidth(stdscr)-1, pfmTermLastline(), 0 );
#		endif

	if ( dlg->window )
	{
		int	i, ww = wwidth(dlg->window); /* > 7, so... */

		if (np == msg)
			i = dlg->offset = 0;
		else
		{	/* ...we can safely restrict 2 <= offset <= width-2, i.e.
			 * 1..ww-3 char msg, 1 space, 2..16 char input-window.
			 */
			i = size < 16 ? (size < 2 ? 2:size) : 16;
			dlg->offset = 1 + (np-msg);
			if (dlg->offset+i > ww) dlg->offset = i < ww-1 ? ww-i:2;
			i = dlg->offset - 1;
		}
		__pfmRedrawFnkeyline(filefnkey);
		refresh();

		werase(dlg->window);
#		if 0 && __CORNELL_CURSES__	/* libcurses bug ...	*/
			mvwaddch(dlg->window, 0, wwidth(dlg->window)-1, '.' );
#		endif
		wmove( dlg->window, 0, 0 );
		while (i-- > 0 && msg < np)
			waddch(dlg->window, *msg++ );

#		if GETSTR_ON_HEADER_LINE
		if (wheight(dlg->window) > 1 && *np && *++np)
		{	wmove( dlg->window, 1, 0 );
			for (i = ww, msg = np; i-- > 0 && *msg; )
				waddch(dlg->window, *msg++ );
		}
#		endif
		/* touchwin(dlg->window); */
		leaveok( dlg->window, FALSE );
		wmove( dlg->window, 0, 0 );
		wrefresh(dlg->window);
	}
	return dlg->window;
}


__DF(
static void pfmTermGetstrclose,( dlg ), x_getstr *dlg )
{
	extern const char *pfm_Currentfnkeys;

	if ( dlg->window )
	{	bool zwei = wheight(dlg->window) > 1;
		delwin(dlg->window);
		/*
		if ( zwei ) __pfmTermDrawcwdline(FALSE);
		*/
	}
	dlg->window = 0;
	pfmTermClrcon();
	__pfmRedrawFnkeyline(pfm_Currentfnkeys);
	touchwin(stdscr);
	refresh();
}


/*
 *	Shift string right by 'col' columns
 */
__DF(
static void pfmTermGetstrshiftr,( start, lim, col, fill ),
char	*start __j__
char	*lim __j__
int 	col __j__
int 	fill )
{
	if (col > 0)
	{	if (lim > start+col)
			memmove( start+col, start, lim-start-col );
		memset( start, fill, col );
	}
	*lim = 0;
}

#undef	GETSTR_ON_HEADER_LINE


__DF(
static int __pfmTermGetstrv,( showonly, msg, size, buffer, vdef ),
int		showonly __j__
const char	*msg __j__
int		size __j__
char	buffer[] __j__
va_list	vdef )
{
	static char	insert = 1;
	x_getstr	dlg;
	va_list	args;
	char	*def, *firstdef;
	int		ibuf = 0, lgbuf = 0, loop;
	/*
	 *	Make sure we don't change 'errno' here with some library call !
	 */
	int	errnobackup = errno;

	va_copy( args, vdef );

	if (size < 1) return -1;

	if ( !(firstdef = va_arg(args,char*)) ) /* empty defaults-list,... */
		va_copy( args, vdef ); /* ...stay at the start */

	if ( !(def = firstdef) )
		*buffer = 0;
	else
	{	strncpy( buffer, def, size );
		buffer[size-1] = 0;
		lgbuf = strlen(buffer);
		if (showonly) ibuf = lgbuf;
	}
	pfmTermGetstropen( &dlg, msg, size );
	pfmTermGetstrwrite( &dlg, buffer, lgbuf, ibuf );

	for (loop = 0; ; loop++)
	{	int	ch = pfmTermGetch();
		switch ( ch )
		{
			case A_ETX: case A_EOT: case A_CAN: case A_ESC:
				pfmTermGetstrclose(&dlg);
				errno = errnobackup;
				return -1;
			case '\n': case '\r':
			#ifdef KEY_ENTER
			case KEY_ENTER:
			#endif
				pfmTermGetstrclose(&dlg);
				errno = errnobackup;
				return 0;
			case KEY_LEFT: case KEY_DOWN:
				if (showonly) goto scrolldefs;
				if (ibuf > 0)
				{	ibuf--;
					break;
				}
				continue;
			case KEY_F(9):
				if (!ibuf) goto jumptoend;
			case_HOME_KEYS:
			#ifdef KEY_BTAB
			case KEY_BTAB:
			#endif
				if (!showonly) ibuf = 0;
				break;
			case KEY_RIGHT: case KEY_UP:
				if (showonly) goto scrolldefs;
				if (ibuf < lgbuf)
				{	ibuf++;
					break;
				}
				continue;
			case_END_KEYS:
			jumptoend:
				ibuf = lgbuf;
				break;
			case '\177':
				if (ibuf < lgbuf) goto deletechar;
			case KEY_BACKSPACE: case '\b':
				if (showonly) goto scrolldefs;
				if (ibuf > 0)
				{	ibuf--;
					goto deletechar;
				}
				continue;
			case KEY_DC:
			deletechar:
				#if CLR_GETSTR_ON_1ST_DEL
				if (loop)
				#endif
				{	if (showonly) goto scrolldefs;
					if (ibuf < lgbuf)
					{	int	i;
						for (i = ibuf; i < lgbuf; i++)
							buffer[i] = buffer[i+1];
						buffer[lgbuf--] = 0;
						break;
					}
					continue;
				} /* else fall through to cancel */
			case KEY_REFRESH: case '\f': /* we use ^L to clear the input */
			#if '\f' != 'L'-'@'
			case 'L'-'@':
			#endif
			case KEY_CANCEL:
				if (showonly) goto scrolldefs;
				buffer[0] = 0;
				ibuf = lgbuf = 0;
				break;
			case KEY_F(1): case KEY_F(2): case KEY_F(3):
			case KEY_F(4): case KEY_F(5): case KEY_F(6):
				if (showonly) goto scrolldefs;
				if (lgbuf+2 < size)
				{	lgbuf += 2;
					pfmTermGetstrshiftr( buffer+ibuf, buffer+lgbuf,
											2, pfm_fwildc_start );
					ibuf++;
					switch ( ch )
					{	case KEY_F(1):
							buffer[ibuf] = pfm_Nosuff_fwildc; break;
						case KEY_F(2):
							buffer[ibuf] = pfm_Fname_fwildc; break;
						case KEY_F(3):
							buffer[ibuf] = pfm_Suffix_fwildc; break;
						case KEY_F(4):
							buffer[ibuf] = pfm_AltD_fwildc; break;
						case KEY_F(5):
							buffer[ibuf] = pfm_CWD_fwildc; break;
						case KEY_F(6):
							buffer[ibuf] = pfm_HOME_fwildc;
					}
					ibuf++;
					break;
				}
				continue;
			case KEY_EIC:	/* Exit insert char mode	*/
			case KEY_IC:	/* Insert char or enter insert mode	*/
				insert = (insert + 1) & 1;
				continue;
			/* Next page | lower right of keypad | ^F	*/
			case KEY_NPAGE: case KEY_C3: case A_ACK:
			/* Previous page | upper right of keypad | ^B	*/
			case KEY_PPAGE: case KEY_A3: case A_STX:
			case KEY_F(7):
			scrolldefs:
				if ( firstdef ) /* one default string at least	*/
				{	if ( !(def = va_arg(args,char*)) )
					{	/* list exhausted, reset to the start */
						va_copy( args, vdef );
						def = va_arg( args, char* );
					}
					strncpy( buffer, def, size );
					buffer[size-1] = 0;
					ibuf = lgbuf = strlen(buffer);
				}
				break;
			default:
				if (showonly) goto scrolldefs;
				if (ch & 0xff00) continue; /* no change !	*/
				if (iscntrl(ch) && !isspace(ch)) continue; /* no change !	*/

				if (!insert && ibuf < lgbuf) /* really overwriting */
				{	buffer[ibuf++] = ch;
					break;
				}
				if (ibuf < lgbuf) /* really inserting */
				{	if (lgbuf+1 < size) lgbuf++;
					pfmTermGetstrshiftr( buffer+ibuf, buffer+lgbuf, 1, ch );
					ibuf++;
					break;
				}
				if (lgbuf+1 < size) /* appending */
				{	buffer[ibuf] = ch;
					buffer[++ibuf] = 0;
					lgbuf++;
					break;
				}
		}
		pfmTermGetstrwrite( &dlg, buffer, lgbuf, ibuf );
	}

}


__DF(
int pfmTermGetstrv,( msg, size, buffer, vdef ),
const char	*msg __j__
int		size __j__
char	buffer[] __j__
va_list	vdef )
{
	return __pfmTermGetstrv( 0, msg, size, buffer, vdef );
}


__DF(
int pfmTermGetstrl,( msg, size, buffer ),
const char	*msg __j__
int		size __j__
char	buffer[]
VARARGS /*, const char *def, ..., (const char*)0 */ )
{
	va_list	args;
	int		i;
	va_start( args, buffer );
	i = __pfmTermGetstrv( 0, msg, size, buffer, args );
	va_end(args);
	return i;
}

__DF(
static int _pfmTermGetstrl,( showonly, msg, size, buffer ),
int		showonly __j__
const char	*msg __j__
int		size __j__
char	buffer[]
VARARGS /*, const char *def, ..., (const char*)0 */ )
{
	va_list	args;
	int		i;
	va_start( args, buffer );
	i = __pfmTermGetstrv( showonly, msg, size, buffer, args );
	va_end(args);
	return i;
}


__DF(
void pfmTermErrormessage,( prefix ),
const char *prefix
VARARGS /*, const char *msg1, ..., (const char*)0 */ )
{
	char   	buffer[80];
	va_list	args;

	va_start( args, prefix );

	if ( curses_is_open )
		__pfmTermGetstrv( TRUE, prefix, 80, buffer, args );
	else
	{	char	*p;
		int 	n = 0;
		int 	errnobackup = errno;
		fprintf( stderr, "%s\n", prefix );
		while ( (p = va_arg(args,char*)) )
			fprintf( stderr, "%4d: %s\n", ++n, p );
		errno = errnobackup;
	}
	va_end(args);
}


#if 0
__DF(
int pfmTermYesNo,( question, yes ), const char *question __j__ bool yes )
{
	char answ[4];
	int  okay;
	const char *def = yes ? pfm_C_yes_ja_oui:pfm_C_no_nein_non;
	const char *alt = yes ? pfm_C_no_nein_non:pfm_C_yes_ja_oui;

	if ( curses_is_open )
		okay = !_pfmTermGetstrl( 0, question, 4, answ, def, alt, (char*)0 );
	else
	{	int	errnobackup = errno;
		fprintf( stderr, "%s\n[%s/%s] ",
				question, pfm_C_no_nein_non, pfm_C_yes_ja_oui );
		okay = fgets(answ,4,stdin) ? 1:0;
		errno = errnobackup;
	}
	if (okay && toupper(*answ) == toupper(*pfm_C_yes_ja_oui))
		return 1;
	return 0;
}
#else

__DF(
static int pfmTermStdioYesNo,( question ), const char *question )
{
	char	answ[2];
	int 	errnobackup = errno;
	int 	okay;
	fprintf( stderr, "%s\n[%s/%s] ",
			question, pfm_C_no_nein_non, pfm_C_yes_ja_oui );
	okay = fgets(answ,2,stdin) ? 1:0;
	errno = errnobackup;
	return (okay && toupper(*answ) == toupper(*pfm_C_yes_ja_oui));
}

__DF(
int pfmTermYesNo,( question, yes ), const char *question __j__ bool yes)
{
	if ( curses_is_open )
	{	x_getstr	dlg;

		for (pfmTermGetstropen(&dlg,question,4); ; )
		{	const char *aw = yes ? pfm_C_yes_ja_oui:pfm_C_no_nein_non;
			int	ch;
			pfmTermGetstrwrite( &dlg, aw, strlen(aw), 0 );
			ch = pfmTermGetch();
			switch ( ch )
			{	case KEY_LEFT: case KEY_DOWN: case KEY_RIGHT: case KEY_UP:
				case KEY_NPAGE: case KEY_C3: case A_ACK:
				case KEY_PPAGE: case KEY_A3: case A_STX:
					/* behave similar to pfmTermGetstr?() */
					yes = !yes;
					break;
				case '\n': case '\r':
				#ifdef KEY_ENTER
				case KEY_ENTER:
				#endif
					pfmTermGetstrclose(&dlg);
					return yes;
				default:
					pfmTermGetstrclose(&dlg);
					return (toupper(ch) == toupper(*pfm_C_yes_ja_oui));
			}
		}
	}
	/*
	 *	pfmTermYesNo() must work, even without curses !
	 */
	return pfmTermStdioYesNo(question);
}
#endif

/***** eof *********************************************************/

int dumpwindows ()
{
#if 0

	static FILE  *dp = 0;
	WINDOW	*wp[5];
	char  *wn[5];
	int	i, j, k;

	if (!dp) dp = fopen("/tmp/windows.dump","wb");
	wp[0] = curscr; wn[0] = "curscr";
	wp[1] = stdscr; wn[1] = "stdscr";
	wp[2] = pfm_CwdirWindow; wn[2] = "pfm_CwdirWindow";
	wp[3] = pfm_FilesWindow; wn[3] = "pfm_FilesWindow";
	wp[4] = pfm_NotesWindow; wn[4] = "pfm_NotesWindow";

	for (k = 0; k < 5; k++)
	{	WINDOW	*w = wp[k];
		fprintf( dp, "%s [%d x %d]:\n\n", wn[k], wheight(w), wwidth(w) );
		fprintf( dp, "_beg=(%d,%d), _max=(%d,%d), _cur=(%d,%d)\n",
				w->_begy,w->_begx, w->_maxy,w->_maxx, w->_cury,w->_curx );
		fprintf( dp, "            " );
		for (j = 0; j < wwidth(w); j++) fputc( '-', dp );
		fputc( '\n', dp );
		for (i = 0; i < wheight(w); i++)
		{	fprintf( dp, "(%2d,%2d)(%2d) ",
					w->_firstchar[i], w->_lastchar[i], w->_numchngd[i] );
			for (j = 0; j < wwidth(w); j++)
			{	wmove( w, i, j );
				fprintf( dp, "%c", winch(w) );
			}
			fputc( '\n', dp );
		}
		fputc( '\n', dp );
	}
	fprintf( dp, "\f\n" );
#endif
	return 0;
}
