/***********************************************************************
* Copyright (c) 1995-2003 by peter fuerst
*
*	This program is free software; you can redistribute it and/or modify
*	it - as explained in the accompaning file COPYING - under the terms
*	of the GNU General Public License as published by the Free Software
*	Foundation; either version 2, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	Author: Peter Frst, post@pfrst.de
*
*	@(#)url.c	10/25/95
*
***********************************************//*$preferred tabstop=4*/

#include <string.h>
#include <ctype.h>
#include <sys/stat.h>

#include "pfm.h"

/***** define & typedef ********************************************/

#ifdef SUPPORT_INET_URL
#define pfm_urlspec_fwildc 'U'
#endif

/***** forward declaration *****************************************/

/***** global ******************************************************/

/***** local *******************************************************/

#ifdef SUPPORT_INET_URL

static const char *www_url_signatur[] =
{	"http://", "https://", "file://", 0
	};
static const char *ftp_url_signatur[] =
{	"ftp://", 0
	};
static const char *telnet_url_signatur[] =
{	"telnet://", "tn3270://", "rlogin://", 0
	};
static const char *gopher_url_signatur[] =
{	"gopher://", "cso://", 0
	};
static const char *wais_url_signatur[] =
{	"wais://", 0
	};
static const char *news_url_signatur[] =
{	"news:", "snews:", "newspost:", "newsreply:",
	"snewspost:", "snewsreply:", 0
	};
static const char *finger_url_signatur[] =
{	"finger://", 0
	};
static const char *mail_url_signatur[] =
{	"mailto:", 0
	};

/*
 *	Check for a URL's type
 */
__DF(
static const char *_pfmFileGetURLobject,( anytype, url, handler ),
bool anytype __j__
const char *url __j__
const char **handler )
{
	/*
	 *	Here we consider the separator "//", if any, as part
	 *	of the specifier, not part of the object name.
	 */
	static struct
	{	const char *(*hdlr)();
		const char **spec;
	}
	_S_url_tbl[] =
	{	{pfmConfWwwCmd,     www_url_signatur},
		{pfmConfFtpCmd,     ftp_url_signatur},
		{pfmConfTelnetCmd,  telnet_url_signatur},
		{pfmConfUrlHdlrCmd, gopher_url_signatur},
		{pfmConfUrlHdlrCmd, wais_url_signatur},
		{pfmConfUrlHdlrCmd, news_url_signatur},
		{pfmConfUrlHdlrCmd, finger_url_signatur},
		{pfmConfMailerCmd,  mail_url_signatur},
	};
	/*
	 *	test for well known URL-types first...
	 */
	int	i;

	for (i = 0; i < sizeof(_S_url_tbl)/sizeof(_S_url_tbl[0]); i++)
	{	const char *s;
		int	j;
		for (j = 0; (s = _S_url_tbl[i].spec[j]); j++)
			if ( pfmStrcasematch(url,s) )
			{	*handler = (*_S_url_tbl[i].hdlr)();
				return (url + strlen(s));
			}
	}
	if (anytype && isalpha(*url))
	{	const char *s = url+1;
		while ( isalnum(*s) ) s++;
		/* We want at least 3 alphanumeric chars. for the specifier */
		if (':' == *s && s > url+2)
		{	if ('/' == s[1] && '/' == s[2]) s += 2;
			*handler = pfmConfUrlHdlrCmd();
			return ++s;
		}
	}
	*handler = pfmConfUrlHdlrCmd();
	return url;
}
#endif


/*
 *	Open a URL and eventually wait for completion
 */
__DF(
int pfmFileGotoURL,( url, anytype ), const char *url __j__ bool anytype )
{
#ifdef SUPPORT_INET_URL
	const char *handler, *objname;

	if ((objname = _pfmFileGetURLobject(anytype,url,&handler)) > url)
	{
		char browsecmd[CWD_MAX_LNG/2];
		char *uspec = (char*) alloca(objname-url+1);
		char *s;
		/*
		 *	Here "%U" stands for "specifier:[//]" and
		 *	"%F" for the object name "domain/path".
		 */
		strncpy( uspec, url, objname-url )[objname-url] = 0;
		strcpy( browsecmd, handler );
		pfmVExpandargs( browsecmd, sizeof(browsecmd),
		                objname, (char*)0, (char*)0 );
		for (s = browsecmd; *s; s++)
			if (pfm_fwildc_start == *s && pfm_urlspec_fwildc == s[1])
				*++s = pfm_Fname_fwildc;
		pfmVExpandargs( browsecmd, sizeof(browsecmd),
		                uspec, (char*)0, (char*)0 );
		return pfmExecommand(browsecmd,EXECSH_WAIT) ? -1:1;
	}
#else
	(void)url; (void)anytype;
#endif
	return 0;
}

/***** eof *********************************************************/
