/***********************************************************************
* Copyright (c) 1995-2003 by peter fuerst
*
*	This program is free software; you can redistribute it and/or modify
*	it - as explained in the accompaning file COPYING - under the terms
*	of the GNU General Public License as published by the Free Software
*	Foundation; either version 2, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	Author: Peter Frst, post@pfrst.de
*
*	@(#)pfm/viewer.c	10/25/95
*
***********************************************//*$preferred tabstop=4*/

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <ctype.h>
#include <sgtty.h>
#include <string.h>
#include <errno.h>
#include "spawn.h"

/***** define & typedef ********************************************/

#include "pfm.h"

#define FILEVIEWONLY "--file-view-only"

#define reallytouch(win) (clearok((win),TRUE),touchwin(win))

/***** forward declaration *****************************************/

__DC( static int  pfmViewBuiltinViewer,(const char*) );
__DC( static int  pfmViewBuiltinPager,(const char*) );
__DC( static int  pfmViewshowfill,(void) );
__DC( static int  pfmViewHelp,(x_menubar*) );
__DC( static int  pfmView1LineUp,(void) );
__DC( static int  pfmView1LineDown,(void) );
__DC( static void pfmviewredrawall,(void) );
__DC( static void pfmdrawviewline,(int) );
__DC( static void pfmviewinitbuf,(long,size_t) );
__DC( static void _pfmdrawfposition,(bool,int,int) );

/***** global ******************************************************/

/***** local *******************************************************/

struct fline
{
	long	pos;	/* line starts at position 'pos' in file	*/
	size_t	lng;	/* line contains 'lng' char's	*/
	char	*buf;	/* line is read into 'buf'	*/
};

static struct
{
	int		fd;		/* input file descriptor */
	long	lng;	/* file length	*/
	char	*fbuf;	/* holds char's to be displayed	*/
	size_t	off;	/* line's 1st column to be displayed	*/
	struct fline	*lbuf;	/* holds an array of 'struct fline's	*/
	struct fline	**lptr;	/* indices into this array	*/
}
fbuffer = { -1, };

static bool	fixedline = 1;
static bool	showhbit  = 1;

static WINDOW	*showwindow = 0;


/********************************************************************
*	Move in the displayed file.
********************************************************************/

/*
 *	start viewer window at file begin.
 */
static int pfmViewTopscroll ()
{
	if (fbuffer.lptr[0]->pos > 0)
		if ( fixedline )
		{	pfmviewinitbuf( 0L, 0 );
			wmove( showwindow, 0, 0 );
			pfmviewredrawall();
		}
		else
		{
		}
	return 0;
}

/*
 *	start viewer window at the file's tail.
 */
static int pfmViewEndscroll ()
{
	if ( fixedline )
	{	int		cols = wwidth(showwindow);
		int		lins = wheight(showwindow);
		ldiv_t	lm;
		long	lpos;

		lm = ldiv( fbuffer.lng, (long)cols );
		lpos = lm.quot - lins;
		if ( lm.rem ) lpos++;
		if (lpos < 0) lpos = 0;
		if (lpos != fbuffer.lptr[0]->pos)
		{	pfmviewinitbuf( lpos*cols, 0 );
			wmove( showwindow, lins-1, 0 );
			pfmviewredrawall();
		}
	}
	else
	{
	}
	return 0;
}

/*
 *	shift cursor or viewer-window one line up
 */
static int pfmView1LineUp ()
{
	if (fbuffer.lptr[0]->pos > 0)
		if ( fixedline )
		{	int	cols = wwidth(showwindow);
			if (fbuffer.lptr[0]->pos < cols)
				pfmViewTopscroll();
			else
			{	int	x, y;
				pfmviewinitbuf( fbuffer.lptr[0]->pos-cols, 0 );
				getyx( showwindow, y, x );
				wmove( showwindow, 0, x );
				pfmviewredrawall();
			}
		}
		else
		{
		}
	return 0;
}

static int pfmViewCursorUp ()
{
	int	x, y;

	getyx( showwindow, y, x );

	if (y < 1) return pfmView1LineUp();

	if (fbuffer.lptr[--y]->lng-fbuffer.off <= x)
		if ((x = fbuffer.lptr[y]->lng-fbuffer.off) > 0) --x;
	_pfmdrawfposition( TRUE, y, x );
	wmove( showwindow, y, x );
	leaveok( showwindow, FALSE );
	wrefresh(showwindow);
	return 0;
}

/*
 *	shift cursor or viewer-window one line down
 */
static int pfmView1LineDown ()
{
	int	cols = wwidth(showwindow);
	int	ll = wheight(showwindow) - 1;

	if (fbuffer.lptr[ll]->pos+cols < fbuffer.lng)
		if ( fixedline )
		{	int	x, y;
			pfmviewinitbuf( fbuffer.lptr[0]->pos+cols, 0 );
			getyx( showwindow, y, x );
			wmove( showwindow, ll, x );
			pfmviewredrawall();
		}
		else
		{
		}
	return 0;
}

static int pfmViewCursorDown ()
{
	int	x, y;

	getyx( showwindow, y, x );

	if (y > wheight(showwindow)-2) return pfmView1LineDown();

	if (fbuffer.lptr[++y]->pos < fbuffer.lng)
	{	if (fbuffer.lptr[y]->lng-fbuffer.off <= x)
			if ((x = fbuffer.lptr[y]->lng-fbuffer.off) > 0) --x;
		_pfmdrawfposition( TRUE, y, x );
		wmove( showwindow, y, x );
		leaveok( showwindow, FALSE );
		wrefresh(showwindow);
	}
	return 0;
}

/*
 *	move cursor left/right or shift viewer-window accordingly
 */
static void _pfmviewcrstoSOL ()
{
	int	x, y;

	getyx( showwindow, y, x );
	_pfmdrawfposition( TRUE, y, 0 );
	wmove( showwindow, y, 0 );
}

static int pfmViewCursorSOL ()
{
	_pfmviewcrstoSOL();
	wrefresh(showwindow);
	return 0;
}

static int pfmViewCursorEnter ()
{
	_pfmviewcrstoSOL();
	return pfmViewCursorDown();
}

static int pfmViewCursorLeft ()
{
	int	x, y;

	getyx( showwindow, y, x );

	if (x > 0)
	{	--x;
		#if 1
			_pfmdrawfposition( TRUE, y, x );
		#endif
		wmove( showwindow, y, x );
		leaveok( showwindow, FALSE );
		wrefresh(showwindow);
	}
	else /* in die vorhergehende Zeile	*/
		if (fbuffer.lptr[y]->pos > 0)
		{	wmove( showwindow, y, wwidth(showwindow)-1 );
			return pfmViewCursorUp();
		}
	return 0;
}

static void _pfmviewcrstoEOL ()
{
	int	x, y;

	getyx( showwindow, y, x );
	x = fbuffer.lptr[y]->lng - 1;
	_pfmdrawfposition( TRUE, y, x );
	wmove( showwindow, y, x > 0 ? x:0 );
}

static int pfmViewCursorEOL ()
{
	_pfmviewcrstoEOL();
	wrefresh(showwindow);
	return 0;
}

static int pfmViewCursorRight ()
{
	int	x, y;

	getyx( showwindow, y, x );

	if (x < wwidth(showwindow)-1
		&& x+fbuffer.off < fbuffer.lptr[y]->lng-1)
	{
		wmove( showwindow, y, ++x );
		#if 1
			_pfmdrawfposition( TRUE, y, x );
		#endif
		leaveok( showwindow, FALSE );
		wrefresh(showwindow);
	}
	else /* in die naechste Zeile	*/
	{
		wmove( showwindow, y, 0 );
		return pfmViewCursorDown();
	}
	return 0;
}

/*
 *	shift viewer-window one page up
 */
static int pfmViewPageUp ()
{
	if ( fixedline )
	{	int	cols = wwidth(showwindow);
		int	lins = wheight(showwindow);

		if (fbuffer.lptr[0]->pos > 0)
			if (fbuffer.lptr[0]->pos < lins*cols)
				pfmViewTopscroll();
			else
			{	int	x, y;
				pfmviewinitbuf( fbuffer.lptr[0]->pos-lins*cols, 0 );
				getyx( showwindow, y, x );
				wmove( showwindow, lins-1, x );
				pfmviewredrawall();
			}
	}
	else
	{
	}
	return 0;
}

/*
 *	shift viewer-window one page down
 */
static int pfmViewPageDown ()
{
	if ( fixedline )
	{	int	cols = wwidth(showwindow);
		int	lins = wheight(showwindow);

		if (fbuffer.lptr[lins-1]->pos+cols < fbuffer.lng)
		{	cols *= lins;
			if (fbuffer.lng <= fbuffer.lptr[lins-1]->pos+cols)
				pfmViewEndscroll();
			else
			{	int	x, y;
				pfmviewinitbuf( fbuffer.lptr[0]->pos+cols, 0 );
				getyx( showwindow, y, x );
				wmove( showwindow, 0, x );
				pfmviewredrawall();
			}
		}
	}
	else
	{
	}
	return 0;
}


static int pfmViewshowfill ()
{
	showhbit = !showhbit;
	pfmviewredrawall();
	return 0;
}


__DF(
static int pfmViewHelp,(m), x_menubar *m )
{
	pfmHelp("viewer");
	erase();
	pfmviewredrawall();
	(void)m;
	return 0;
}


/********************************************************************
*	Draw the file-display window.
********************************************************************/
#define BOLD "\007" /* use BELl to emphasize the next character */

static char	pfm_C_viewfnkey[] =
		"F " BOLD "1-up " BOLD "2-down " BOLD "3-left " BOLD "4-right "
		BOLD "5-Restore_left " BOLD "6-Search " BOLD "7-Again";

const x_menubar	pfm_S_viewfnkeys[] =
{
	{ '1', "", (__DC(int(*),(x_menubar*)))pfmView1LineUp },	/* up	*/
	{ '2', "", (__DC(int(*),(x_menubar*)))pfmView1LineDown },	/* down	*/
	{ '3', "", pfmMnuNop },	/* left	*/
	{ '4', "", pfmMnuNop },	/* right	*/
	{ '5', "", pfmMnuNop },	/* Restore left	*/
	{ '6', "", pfmMnuNop },	/* search	*/
	{ '7', "", pfmMnuNop },	/* Again	*/
	{ 0, "", pfmMnuNop }
};

static const x_menubar	sX_viewmenu[] =
{
	{ KEY_HELP, "", pfmViewHelp },
	{ KEY_SHELP, "", pfmViewHelp },
	{ KEY_HOME, "", pfmViewTopscroll },	/* Home key (upward+left arrow) */
	{ KEY_A1, "", pfmViewTopscroll },	/* upper left of keypad */
	{ KEY_LL, "", pfmViewEndscroll },	/* home down or bottom (lower left) */
	{ KEY_END, "", pfmViewEndscroll },	/* end key */
	{ KEY_C1, "", pfmViewEndscroll },	/* lower left of keypad */
	{ KEY_DOWN, "", pfmViewCursorDown },/* The four arrow keys ... */
	{ KEY_UP, "", pfmViewCursorUp },
	{ KEY_RIGHT, "", pfmViewCursorRight },
	{ KEY_LEFT, "", pfmViewCursorLeft },
	{ KEY_NPAGE, "", pfmViewPageDown },	/* Next page */
	{ KEY_C3, "", pfmViewPageDown },	/* lower right of keypad */
	{ A_ACK, "", pfmViewPageDown },		/* ^F */
	{ KEY_PPAGE, "", pfmViewPageUp },	/* Previous page */
	{ KEY_A3, "", pfmViewPageUp },		/* upper right of keypad */
	{ A_STX, "", pfmViewPageUp },		/* ^B */
	{ 'H', BOLD"Help, ", pfmViewHelp },
	{ '?', "", pfmViewHelp },
	{ 'B', BOLD"Bottom, ", pfmViewEndscroll },
	{ 'T', BOLD"Top, ", pfmViewTopscroll },
	{ '+', BOLD"+n or "BOLD"-n lines, ", pfmMnuNop },
	{ '-', ""/* -n */, pfmMnuNop },
	{ 'F', BOLD"Fixed or "BOLD"Variable lines, ", pfmMnuNop },
	{ 'V', ""/* Variable lines */, pfmMnuNop },
	{ 'S', BOLD"Showfill, ", (__DC(int(*),(x_menubar*)))pfmViewshowfill },
	{ 'Q', BOLD"Quit, ", pfmMnuCancelmenu },
	/*{ 'h', "", pfmViewCursorLeft },*/
	{ '<', "", pfmViewCursorLeft },
	{ '>', "", pfmViewCursorRight },
	{ 'l', "", pfmViewCursorRight },
	{ ' ', "", pfmViewCursorRight },
	{ '\t', "", pfmViewCursorRight },
	{ '\r', "", pfmViewCursorEnter },
	{ '\n', "", pfmViewCursorEnter },
	#ifdef KEY_ENTER
	{ KEY_ENTER, "", pfmViewCursorEnter },
	#endif
	{ 'j', "", pfmViewCursorDown },
	{ 'k', "", pfmViewCursorUp },
	{ '^', "", pfmViewCursorSOL },
	{ '|', "", pfmViewCursorSOL },
	{ '$', "", pfmViewCursorEOL },
	{ 'E', "", pfmViewCursorEOL },
	{ 0, "", pfmMnuNop }
};


__DF(
static void _pfmdrawfposition,( rfrsh, row, col ),
bool rfrsh __j__ int row __j__ int col )
{
	char	buf[20];

	sprintf( buf, "%ld", fbuffer.lptr[row]->pos+fbuffer.off+col );
	move( pfmTermLastline(), sizeof(pfm_C_viewfnkey) );
	clrtoeol();
	move( pfmTermLastline(), wwidth(stdscr)-2-strlen(buf) );
	addstr(buf);
	if (rfrsh) refresh();
}

__DF(
static void pfmdrawviewline,( line ), int line )
{
	struct fline *lp;
	bool	so = 0;
	int		i;

	if (line < 0 || wheight(showwindow) <= line) return;

	lp = fbuffer.lptr[line];
	wmove( showwindow, line, 0 );
	wclrtoeol(showwindow);

	for (i = 0; i < lp->lng; i++)
	{	unsigned c = (unsigned char)lp->buf[i];
		unsigned k = toascii(c);
		if ( iscntrl(k) )	/* Play it save, even on 7bit-terminals... */
		{	if (showhbit || c != k)
				k += '@';
			else
				k = '^';
			if (!so) wstandout(showwindow);
			so = TRUE;
		}
		else
		{	if ( showhbit ) k = c;
			if (so) wstandend(showwindow);
			so = FALSE;
		}
		waddch( showwindow, k );
	}
	wstandend(showwindow);
}

static void pfmviewredrawall ()
{
	int	lins = wheight(showwindow);
	int	i, x, y;

	getyx( showwindow, y, x );

	leaveok( stdscr, TRUE );
	__pfmRedrawFnkeyline(pfm_C_viewfnkey);
	__pfmRedrawHeader(sX_viewmenu);
	_pfmdrawfposition( FALSE, y, x );
	refresh();
	leaveok( stdscr, FALSE );
	leaveok( showwindow, TRUE );

	for (i = 0; i < lins; i++) pfmdrawviewline(i);

	wmove( showwindow, y, x );
	leaveok( showwindow, FALSE );
	wrefresh(showwindow);
}

/********************************************************************
*	Display the contents of a file.
********************************************************************/

/*
 *	Allocate all buffers and the viewer window, open the file
 *  for reading and display the beginning of the file.
 */
__DF(
static void pfmviewinitbuf,( start, offs ), long start __j__ size_t offs )
{
	if (fbuffer.fd > 0)
	{	int	lins = wheight(showwindow);
		int	cols = wwidth(showwindow);

		if ( fixedline )
		{	int	i, rlg;
			lseek( fbuffer.fd, start, SEEK_SET );
			rlg = read( fbuffer.fd, fbuffer.fbuf, lins*cols );
			lseek( fbuffer.fd, 0L, SEEK_END );
			fbuffer.lng = tell(fbuffer.fd);
			for (i = 0; i < lins; i++)
			{	fbuffer.lptr[i] = fbuffer.lbuf+i;
				fbuffer.lbuf[i].buf = fbuffer.fbuf + i*cols;
				fbuffer.lbuf[i].pos = start;
				fbuffer.lbuf[i].lng = 0;
				if (0 < rlg) fbuffer.lbuf[i].lng = (cols < rlg ? cols:rlg);
				start += cols;
				rlg -= cols;
			}
			fbuffer.off = 0;
			(void)offs;	/* not used with fixed lines	*/
		}
		else
		{
		}
	}
}

/*
 *  Close the file and the viewer-window, release all buffers.
 */
static void pfmViewclose ()
{
	if (fbuffer.fd > 0) close(fbuffer.fd);

	if (fbuffer.lptr) free(fbuffer.lptr);
	if (fbuffer.lbuf) free(fbuffer.lbuf);
	if (fbuffer.fbuf) free(fbuffer.fbuf);
	memset( &fbuffer, 0, sizeof(fbuffer) );
	fbuffer.fd = -1;

	if (showwindow)
	{	werase(showwindow);
		wrefresh(showwindow);
		delwin(showwindow);
		showwindow = 0;
	}
}

__DF(
static int pfmViewopen,( fname ), const char *fname )
{
	int	lins = pfmTermLastline()-1;
	int	cols = wwidth(stdscr);

	pfmViewclose();

	fbuffer.fd = open( fname, O_RDONLY );

	if (fbuffer.fd > 0) /* '>0' instead of '>=0' since 0 is always 'stdin' */
	{
		fbuffer.lptr = (struct fline**) malloc(sizeof(struct fline*)*lins);
		fbuffer.lbuf = (struct fline*) malloc(sizeof(struct fline)*lins);
		fbuffer.fbuf = (char*) malloc(lins*cols);

		showwindow = subwin( stdscr, lins, cols-2, 1, 1 );

		if (fbuffer.lptr && fbuffer.lbuf && fbuffer.fbuf && showwindow)
		{	pfmviewinitbuf( 0L, 0 );
			pfmTermClearscreen();
			erase();	/* we also need a blank stdscr	*/
			reallytouch(stdscr);
			wmove( showwindow, 0, 0 );
			pfmviewredrawall();
		}
		else
		{
			pfmViewclose();
		}
	}
	return fbuffer.fd;
}


/*
 *	Display the file and the file-viewer menu and wait for input
 */
__DF(
int pfmViewFileviewer,( fname ), const char *fname )
{
	const char *viewer = pfmConfViewerCmd();

	if (viewer && *viewer)
	{	unsigned size = strlen(viewer) + strlen(fname) + 5;
		char	*line = (char*) alloca(size);
		if (line)
		{	strcpy( line, viewer );
			pfmVExpandargs( line, size, fname, (char*)0, (char*)0 );
			return pfmExecshell( line, EXECSH_NORRD );
		}
	}
	return pfmViewBuiltinViewer(fname);
}

__DF(
static int pfmViewBuiltinViewer,( fname ), const char *fname )
{
	static char	reentry = 0;

	if ( !reentry )
	{
		if (pfmViewopen(fname) > 0)
		{	const x_menubar	*menu = sX_viewmenu;
			reentry = 1;
			while ( menu )
			{	short	key = pfmTermGetch();
				short	upkey = key & 0xff80 ? key:toupper(key);

				switch ( key )
				{
					/*case KEY_EXIT:*/
					case A_ETX:		/* ^C (End Of Text)	*/
					case KEY_BREAK:
					case A_EOT:		/* ^D (End Of Transmission)	*/
					case A_CAN:		/* ^X (CANcel)	*/
					case A_ESC:		/* Escape		*/
						menu = 0;
						break;
		
					case KEY_REFRESH:	/* ^L (Form Feed) is commonly */
					case '\f':			/* used for screen refresh... */
						leaveok( stdscr, FALSE );
						reallytouch(stdscr);
						refresh();
						continue;

					case 'V': case 'v':
						/* for now we need to trap 'variable lines' ... */
						fixedline = 0;
						pfmViewBuiltinPager(fname);
						fixedline = 1;
						pfmviewredrawall();
						continue;

					default:
					{	const x_menubar	*pm;
						if (KEY_F(0) < key && key < KEY_F(8))
						{	key -= KEY_F(1);
							(*pfm_S_viewfnkeys[key].fct)(0);
							continue;
						}
						if ('0' < key && key < '8')
						{	key -= '1';
							(*pfm_S_viewfnkeys[key].fct)(0);
							continue;
						}
						for (pm = menu; pm->key && pm->fct; pm++)
							if (key == pm->key || upkey == pm->key)
							{	if ((*pm->fct)((x_menubar*)menu) == -1)
									menu = 0;
								key = 0;
								break;
							}
					}
				}
			}
			pfmViewclose();
		}
		erase();
		pfmTermRedraw();
		reentry = 0;
	}
	return 0;

}


/********************************************************************
*	Display the contents of a file with the systems viewer.
********************************************************************/

/*
 *	For now we only simulate variable-line-length-mode
 *	by displaying the file through 'more' or 'less'...`
 */
__DF(
static int pfmViewPagerputc,( fp, ch ),
FILE *fp __j__
unsigned char ch )
{
	int	k = toascii(ch);
	/* avoid ctl-characters and don't interrupt 'more' with '\f' */
	if (iscntrl(k) && (!isspace(k) || '\f' == k))
	{	if (fputc('^',fp) == EOF) return EOF;
		ch += '@';
	}
	return fputc(ch,fp);
}

__DF(
static int pfmViewBuiltinPager,( fname ), const char *fname )
{
	int fd = open( fname, O_RDONLY );

	if (fd < 0)
		pfmTermErrormessage( strerror(errno), fname, (char*)0 );
	else
	{	FILE *fpipe;
	
		pfmPreExecshell(0);
	
		if ( !(fpipe = popen(pfmConfPagerPrg(),"w")) )
			pfmTermErrormessage( strerror(errno), (char*)0 );
		else
		{	unsigned char c;
			while (read(fd,&c,sizeof(c)) > 0)
				if (pfmViewPagerputc(fpipe,c) == EOF) break;
			fputc('\n',fpipe);
			pclose(fpipe);
		}
		pfmPostExecshell(EXECSH_WAIT|EXECSH_NORRD);
		clearok( stdscr, TRUE );
		close(fd);
	}
	return 0;
}


#if 0
static unsigned int linepos = 0;
__DF(
static void outc,(c), char c )
{
	if (c == '\f') c = ' ';	/* don't interrupt 'more' with '\f' */

	if (c == '\n')
		linepos = 0;
	else
		if (linepos++ == 72)
		{	putchar('\n');
			linepos = 0;
		}
	putchar(c);
}
#endif


/***** eof *********************************************************/

