/***********************************************************************
* Copyright (c) 1995-2003 by peter fuerst
*
*	This program is free software; you can redistribute it and/or modify
*	it - as explained in the accompaning file COPYING - under the terms
*	of the GNU General Public License as published by the Free Software
*	Foundation; either version 2, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	Author: Peter Frst, post@pfrst.de
*
*	termio.c	10/25/95
*	wgetch.c
*
***********************************************//*$preferred tabstop=4*/

#include <stdio.h>
#include <stddef.h>
#include <dirent.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>


/***** define & typedef ********************************************/

#include "pfm.h"

__DC( bool pfmConfMouseinput,(void) );

#undef  NEED_NODELAY_ON_WGETCH      /* avoid, whenever possible... */
#define USE_CURSES_KEYPAD_MAPPING 1 /* use, whenever possible... */

#if defined(__COHERENT_286__) && NEED_NODELAY_ON_WGETCH
/* here keypad does not work with 'nodelay'. */
# undef USE_CURSES_KEYPAD_MAPPING
#endif
#ifdef __BERKELEY_CURSES__
# if defined(__linux__) || defined(__FreeBSD__)
#  undef USE_CURSES_KEYPAD_MAPPING	/* here is no 'keypad' */
# endif
#endif

#if defined(__linux__) || defined(__FreeBSD__)\
    || defined(_P1003_1B_VISIBLE_HISTORICALLY)
# define HAS_POSIX_NANOSLEEP 1 /* POSIX.1b */
#endif

#define	pfmKEYBUFSIZE	32

/***** global ******************************************************/

/***** local *******************************************************/

static int escapes_pending = 0, restart_pending = 0;


/********************************************************************
*	Table of key sequences
********************************************************************/

#if	!USE_CURSES_KEYPAD_MAPPING
/*
 *	reduced table of function key sequences
 */
typedef struct keyent
{
	short	key;
	short	len;
	char	*seq;
}
x_keyent;

static x_keyent	keyent_dummy = { 0, 0, "" };

static x_keyent	*keyent_table = &keyent_dummy;

/*
 *	map function key sequence to key value
 */
__DF(
static x_keyent *pfmTermMapkeysequ,( str, lng ),
const char *str __j__
short lng )
{
	x_keyent  *xp = keyent_table;

	while ( xp->key )
	{	if (lng >= xp->len && !strncmp(xp->seq,str,xp->len))
			return xp;
		xp++;
	}
	return 0;
}

#endif


void  pfmTermBuildkeyent ()
{
#if	!USE_CURSES_KEYPAD_MAPPING
	tkeyent_t	*tp, *tbl = (tkeymap ? tkeymap:tkeytab);
	int	i;

	for (i = 0, tp = tbl; tp->id; tp++)
		if (tp->cpp && *tp->cpp && **tp->cpp) i++;

	if (i > 0)
	{	if (keyent_table != &keyent_dummy)
			free(keyent_table);

		keyent_table = (x_keyent*) malloc(sizeof(x_keyent)*(i+1));

		if ( !keyent_table )
			keyent_table = &keyent_dummy;
		else
		{	x_keyent  *xp = keyent_table;
			for (tp = tbl; tp->id; tp++)
				if (tp->cpp && *tp->cpp && **tp->cpp)
				{	xp->key = tp->id;
					xp->len = strlen(*tp->cpp);
					xp->seq = *tp->cpp;
					xp++;
				}
			xp->seq = keyent_dummy.seq;
			xp->key = xp->len = 0;
		}
	}
#endif
}


/********************************************************************
*	Mapping pointing-device to menu-entry
********************************************************************/

#ifdef NCURSES_MOUSE_VERSION
/*
 *	table of character-cell-coordinates and corresponding keys
 */
typedef struct hotspot
{
	WINDOW	*win;	/* registered in window win */
	/* screen-, not window- coordinates */
	unsigned char	x1;	/* left.. */
	unsigned char	y1;	/* ..lower corner */
	unsigned char	x2;	/* right.. */
	unsigned char	y2;	/* ..upper corner */
	char		typ;
	union {
	__DC(void	(*fct),(int,int));
		short	key;
	}	act;
}
x_hotspot;

static struct
{
	int       num;
	x_hotspot *lst;
}
mousemap_table = { 0, 0 };


/*
 *	Unregister all hotspots belonging to window
 */
__DF(
void pfmTermUnregmousemaps,( win ), WINDOW *win )
{
	if (win && mousemap_table.lst)
	{	int i;
		/*
		 *	We do not release the list's memory, just clear the entries.
		 *	The same space is likely to be requested soon again.
		 */
		for (i = 0; i < mousemap_table.num; ++i)
			if (win == mousemap_table.lst[i].win)
			{	/*
				 *	no need to touch .x,.y since by construction there
				 *	is at most one matching entry in the list and no
				 *	active entry can be shadowed by the disabled one.
				 */
				mousemap_table.lst[i].act.fct = 0;
			}
	}
}

/*
 *	Register hotspot for mousevent
 */
__DF(
static int pfmTermRegmousemap,( map ), x_hotspot *map )
{
	if (map->win && pfmConfMouseinput())
	{	x_hotspot *pfree = 0;

		/* we need screen coordinates */
		map->x1 += map->win->_begx;
		map->y1 += map->win->_begy;
		map->x2 += map->win->_begx;
		map->y2 += map->win->_begy;

		if ( mousemap_table.lst )
		{	int i;
			for (i = 0; i < mousemap_table.num; ++i)
			{	if (map->x1 == mousemap_table.lst[i].x1 &&
					map->y1 == mousemap_table.lst[i].y1 &&
					map->x2 == mousemap_table.lst[i].x2 &&
					map->y2 == mousemap_table.lst[i].y2)
				{	/*
					 *	Assign new key to existing hotspot (this
					 *	implies disabling (x,y), when key == 0)
					 */
					mousemap_table.lst[i] = *map;
					return 1;
				}
				/* remember the *1st* free slot */
				if (!mousemap_table.lst[i].act.fct && !pfree)
					pfree = mousemap_table.lst+i;
			}
		}
		/*
		 *	If there is no existing entry for (x,y) and key == 0, we are
		 *	done. Otherwise we enter (win,x,y,key) into the free slot or
		 *	expand the list and insert the new values at its end.
		 */
		if ( map->act.fct )
		{	if ( !pfree )
			{	pfree = (x_hotspot*)
					realloc( mousemap_table.lst,
					         (mousemap_table.num+10)*sizeof(*pfree) );
				if ( pfree )
				{	mousemap_table.lst = pfree;
					pfree += mousemap_table.num;
					mousemap_table.num += 10;
					memset( pfree+1, 0, sizeof(*pfree)*9 );
				}
			}
			if ( pfree )
			{	*pfree = *map;
				return 1;
			}
		}
	}
	return 0;
}

/*
 *	Register hotspot for mousevent, mapped to key
 */
__DF(
int pfmTermRegkeymousemap,( win, key, x1, y1, x2, y2 ),
WINDOW *win __j__
int    key __j__
int    x1 __j__
int    y1 __j__
int    x2 __j__
int    y2 )
{
	x_hotspot map;
	map.win = win;
	map.x1  = x1;
	map.y1  = y1;
	map.x2  = x2;
	map.y2  = y2;
	map.typ = 'k';
	map.act.key = key;
	return pfmTermRegmousemap(&map);
}

/*
 *	Register hotspot for mousevent, mapped to function
 */
__DF(
int pfmTermRegfctmousemap,( win, fct, x1, y1, x2, y2 ),
WINDOW *win __j__
__DC(void	(*fct),(int,int)) __j__
int    x1 __j__
int    y1 __j__
int    x2 __j__
int    y2 )
{
	x_hotspot map;
	map.win = win;
	map.x1  = x1;
	map.y1  = y1;
	map.x2  = x2;
	map.y2  = y2;
	map.typ = 'f';
	map.act.fct = fct;
	return pfmTermRegmousemap(&map);
}


/*
 *	map mousevent to key value
 */
__DF(
static int pfmTermMapmouse,( key ), int key )
{
	if (KEY_MOUSE == key)
	{	MEVENT event;
		if (getmouse(&event) != ERR)
			if (event.bstate & (BUTTON1_CLICKED | BUTTON1_DOUBLE_CLICKED))
			{	int i;
				for (i = 0; i < mousemap_table.num; ++i)
				{	x_hotspot *pp = mousemap_table.lst+i;
					if (pp->win &&
						pp->x1 <= event.x && event.x <= pp->x2 &&
						pp->y1 <= event.y && event.y <= pp->y2)
					{
						if (pp->typ != 'f')
							return pp->act.key;
						if (pp->act.fct)
							/* back from screen to window coordinates */
							(*pp->act.fct)(event.x - pp->win->_begx,
							               event.y - pp->win->_begy);
						break;
					}
				}
			}
		key = 0; /* discard mouse-event */
	}
	return key;
}
#endif


/********************************************************************
*	terminal I/O
********************************************************************/

/*
 *	Hack to support ^V and SIGWINCH:
 *	RaiseEsc allows the signal-handlers to abort any input activities, that
 *	may be in progress. RaiseSyn avoids to call pfmReinitscreen() in the
 *	signal-handler, delaying the call to a "normal" state.
 */
__DF(
void pfmTermRaiseEsc,( num ), int num )
{
	escapes_pending = num;
}

void pfmTermRaiseSyn ()
{
	++restart_pending;
}

/*
 *	This function merely exists to keep all dependencies on the
 *	NEED_NODELAY_ON_WGETCH setting local to this file, and should
 *	be considered private to pfmTermFixmode().
 */
void pfmTermSetnodelay ()
{
	/*
	 * Do not set nodelay(), unless ye really can't avoid it !
	 * It inhibits any screen-refresh on xterm, if we do not
	 * sleep for at least 1ms between calls to wgetch(), and
	 * anyway calling wgetch() without waiting for input slows
	 * down your machine noticably !!!
	 */
	#if NEED_NODELAY_ON_WGETCH
		nodelay( pfm_CwdirWindow, TRUE );
		nodelay( pfm_FilesWindow, TRUE );
		nodelay( pfm_NotesWindow, TRUE );
		nodelay( stdscr, TRUE );
	#endif
}

/*
 *	Here - for some reason - we cannot take advantage of curses'
 *	keypad and wgetch().  Instead we have to gather any key
 *	(-sequence) directly from the terminal and evaluate it "by
 *	hand".
 */
#if	!USE_CURSES_KEYPAD_MAPPING
__DF(
static int  pfmTermWgetseq,( win ), WINDOW *win )
{
	static char  keybuf[pfmKEYBUFSIZE] = {0,};
	static short start = 0;
	static short cnt = 0;
	/*
	 *	fill the buffer with waiting characters:
	 */
	int i;
	int	fd = fileno(stdin);
#	if __COHERENT_286__
		if (fd != _tty_ch)
			pfmExit( 1, "pfmTermWgetch(): TTY is not stdin", (char*)0 );
#	endif

	while (cnt < pfmKEYBUFSIZE)
	{
		int  cc = stdin->_cc;

		if ( !cc ) ioctl( fd, TIOCQUERY, &cc );

		if (cc > 0)
			keybuf[(start+cnt++) % pfmKEYBUFSIZE] = fgetc(stdin);
		else
		{	if (cnt > 0) break;
			if (!wgetdelayed(win)) return -1;
		}
	}
	/*
	 *	look for matching key sequences:
	 */
	if (cnt > 1)
	{	char  tmp[pfmKEYBUFSIZE];
		x_keyent  *p;
		int	j;

		for (i=0, j=start; i < cnt; i++, j++, j %= pfmKEYBUFSIZE)
			tmp[i] = keybuf[j];
		tmp[i] = 0;
		if ( (p = pfmTermMapkeysequ(tmp,cnt)) )
		{	start = (start + p->len) % pfmKEYBUFSIZE;
			cnt -= p->len;
			return p->key;
		}
	}
	i = keybuf[start];
	start = ++start % pfmKEYBUFSIZE;
	cnt--;
	return i;
}
#endif

/*
 *	Return one character from the standard input
 *	after evaluating escape sequences.
 */
__DF(
int  pfmTermWgetch,( win ), WINDOW *win )
{
	int  key, use;

	if (escapes_pending > 0) /* hack to support ^V and SIGWINCH */
	{	--escapes_pending;
		return A_ESC;
	}
	if (restart_pending > 0) /* support SIGWINCH */
	{	restart_pending = 0;
		return A_SYN;
	}
#ifdef __XSI_CURSES__
	if ( !wgetdelayed(win) ) wtimeout( win, 5 );
#endif
#if	USE_CURSES_KEYPAD_MAPPING
	use = wgetkeypad(win);

	keypad( win, 1|2 );
	key = wgetch(win);
	keypad( win, use );
#else
	key = pfmTermWgetseq(win);
#endif
#ifndef __XSI_CURSES__
	/*
	 *	As far as possible, avoid a busy-wait-loop. So we sleep for
	 *	a few milliseconds at least, whenever non-blocking mode is
	 *	on and no input was ready.
	 */
	if (!wgetdelayed(win))
		if (-1 == key
			#ifndef __COHERENT_386__
				|| ERR == key
			#endif
			)
		{
			#ifdef __NCURSES__
				napms(5);
			#elif HAS_POSIX_NANOSLEEP
				struct timespec delay;
				delay.tv_sec = 0;
				delay.tv_nsec = 5000000L;
				nanosleep(&delay,&delay);
			#elif defined(__COHERENT_386__)
				nap(5L); /* xterm needs at least 5ms ! */
			#else
				sleep(1); 
			#endif
		}
#endif
#ifdef NCURSES_MOUSE_VERSION
	key = pfmTermMapmouse(key);
#endif
	return key;
}

#undef	pfmTermGetch
int  pfmTermGetch ()
{
	return pfmTermWgetch(stdscr);
}

/*
 *	clear the input buffer
 */
int pfmTermClrcon ()
{
#ifdef __COHERENT__
	int	n, fd = fileno(stdin);
	while (stdin->_cc) fgetc(stdin);
	while (n = 0, ioctl(fd,TIOCQUERY,&n), n)
		read( fd, &n, 1 );
#else
	/* to do... */
#endif
	return 0;
}


/*
 *	Show escape sequences or key-values, that Curses returns with
 *	different keypad-settings.
 *	This will help to check correctnes of termcap/termino entries !
 */
__DF(
void pfmTermTestkeypad,( msg ), const char *msg )
{
	initscr();
	if (stdscr == 0)
	{	fprintf( stderr, "cannot init screen !\n");
		return;
	}
	cbreak();
	raw();
	refresh(); /* seems, we can't do without this */
	if (msg) fputs( msg, stderr );

	while (1)
	{	int i, kbflag, qflag;
		#ifdef NCURSES_MOUSE_VERSION
			mmask_t mmask = 0;
			qflag = 0;
		#else
			qflag = 2;
		#endif
		echo();

		while (qflag != (1|2))
		{
			#ifdef NCURSES_MOUSE_VERSION
			if ( !(qflag & 2) )
				fputs( "\r[Y/N] report mouse events ?\n", stderr );
			#endif
			if ( !(qflag & 1) )
				for (i = '0'; i <= '3'; ++i)
					fprintf( stderr, "\r[%c]   set keypad flag = %c\n", i, i );
			fputs( "\r[q]   quit\n"
				"\renter your choice...", stderr );

			switch ( (i = wgetch(stdscr)) )
			{	case '0': case '1': case '2': case '3':
					if ( !(qflag & 1) )
					{	kbflag = i - '0';
						qflag |= 1;
					}
					break;
				#ifdef NCURSES_MOUSE_VERSION
				case 'Y': case 'y':
				case 'N': case 'n':
					if ( !(qflag & 2) )
					{	if (i != 'N' && i != 'n')
							mmask = ALL_MOUSE_EVENTS | REPORT_MOUSE_POSITION;
						else
							mmask = 0;
						qflag |= 2;
					}
					break;
				#endif
				default:
					if ( isspace(i) ) break;
					if ( !iscntrl(i) ) break;
				case 'Q': case 'q':
					goto end;
			}
			clear();
			refresh();
		}
		noecho();
		keypad( stdscr, kbflag );
		#ifdef NCURSES_MOUSE_VERSION
			mousemask( mmask, (mmask_t*)0 );
		#endif
		fputs( "\rterminate with <ctrl-c> or <ctrl-d>,\n"
		       "\rnow hit any key ...", stderr );
		do
		{	i = wgetch(stdscr);
			fprintf( stderr, "\rkey 0x%03x 0%03o %3d %c",
			         i, i, i, isprint(i) ? i:' ' );
			#ifdef NCURSES_MOUSE_VERSION
				if (kbflag && KEY_MOUSE == i)
				{	MEVENT event;
					fputs( " KEY_MOUSE: ", stderr );
					if (getmouse(&event) != ERR)
						fprintf( stderr, "0x%04x (%d,%d)",
								event.bstate, event.x, event.y );
					else
						fputs( "failed.", stderr );
				}
			#endif
			fputs( "\n\rhit any key ...", stderr );
			fflush(stderr);
		} while (0 < i && i != 3 && i != 4);
	}
	end:
	noraw();
	echo();
	nocbreak();
	endwin();
}


/***** eof *********************************************************/
